#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <tulip/SuperGraph.h>
#include <tulip/TulipPlugin.h>
#include <tulip/Delaunay.h>

#include "EmbedderEmbedGraph.h"
#include "EmbedderParameters.h"
#include "EmbedderDijkstra.h"

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
   using stdext::hash_map;
#endif

#include <vector>

using namespace std;

/** \addtogroup layout */
/*@{*/
/// AreaEmbedder.cpp - An implementation of the ACE algorithm.
/**
 * This plugin implements ACE:
 * "ACE a fast multiscale eigenvectors computation for drawing huge graphs"
 *
 * Authors: Y. Koren, L. Carmel and D. Harel
 * From the sources of Embedder (reference implementation of the mentionned article)
 *
 * See http://www.wisdom.weizmann.ac.il/~yehuda/ for both publication and source code
 *
 * Integration in Tulip by B.Mathieu
 */
class AreaEmbedder : public Layout {
public:
  AreaEmbedder(const PropertyContext &);
  ~AreaEmbedder();
  bool run();
private:
  vector<node> nodemap;
  hash_map<node, unsigned int> rmap;
};

//returns the square of x
inline double sqr (double x) {
  return x*x;
}//end sqr

AreaEmbedder::AreaEmbedder(const PropertyContext &context) : Layout(context) {
}

AreaEmbedder::~AreaEmbedder() {}

bool AreaEmbedder::run() {
  unsigned int numberOfNodes = (unsigned int) superGraph->numberOfNodes();
  vtx_data *graph = new vtx_data[numberOfNodes];
  int *edges = new int[2 * superGraph->numberOfEdges() + numberOfNodes];
  float *ewgts = new float[2 * superGraph->numberOfEdges() + numberOfNodes];
  const unsigned int proj_dim = 2;
  pivots = NULL;

  nodemap.resize(numberOfNodes);
  rmap.clear();
  rmap.resize(numberOfNodes);

  Iterator<node> *it = superGraph->getNodes();
  unsigned int counter = 0;
  while (it->hasNext()) {
    node n = it->next();
    nodemap[counter] = n;
    rmap[n] = counter++;
  } delete it;

  SizesProxy *sizes = superGraph->getProperty<SizesProxy> ("viewSize");
  for (unsigned int i=0; i < numberOfNodes; ++i) {
    node n = nodemap[i];
    graph[i].nedges = superGraph->deg(n) + 1;
    graph[i].edges = edges++;
    graph[i].ewgts = ewgts++;
    graph[i].edges[0] = i;
    graph[i].ewgts[0] = 1.0;
    Size nSize = sizes->getNodeValue (n);
    float nWeight = static_cast<float> (sqrt(sqr(nSize.getW()/2.0) + sqr(nSize.getH()/2.0)));
    
    Iterator<node> *tmp = superGraph->getInOutNodes(n);
    while (tmp->hasNext()) {
      node opposite = tmp->next();
      *edges = rmap[opposite];
      Size oppSize = sizes->getNodeValue (opposite);
      float oppWeight = static_cast<float>(sqrt(sqr (oppSize.getW()/2.0) + sqr (oppSize.getH()/2.0)));
      *ewgts = max(1.0, max(nWeight, oppWeight));
      ++edges;
      ++ewgts;
    } delete tmp;
  }//end for i

  DistType **coords = NULL;
  double **dcoords = new double *[proj_dim];
  dcoords[0] = new double[numberOfNodes*2];
  dcoords[1] = &(dcoords[0][numberOfNodes]);
  int dims = 50;
  embed_graph(graph, numberOfNodes, dims, coords);
  center_coordinate(coords, numberOfNodes, dims);
  PCA(coords, dims, numberOfNodes, dcoords, 0, 1, true);
  delete [] coords[0];
  delete [] coords;

  for (hash_map<node, unsigned int>::const_iterator it = rmap.begin(); it != rmap.end(); ++it) {
    unsigned int index = it->second;
    layoutProxy->setNodeValue(it->first, 
			      Coord(dcoords[0][index], dcoords[1][index]));
  }//end for
 
  delete [] graph[0].edges;
  delete [] graph[0].ewgts;
  delete [] graph;
  delete [] dcoords[0];
  delete [] dcoords;
  if (pivots != NULL) delete [] pivots;
  return true;
}

LAYOUTPLUGIN(AreaEmbedder, "Area HDE", "Daniel Archambault", "07/11/2005", "Ok", "0", "1");
