#include <string>
#include <queue>
#include <math.h>
#include <time.h> 

#include <tulip/MethodFactory.h>
#include <tulip/LayoutProxy.h>
#include <tulip/SelectionProxy.h>
#include <tulip/PropertyProxy.h>
#include "Spanner/TreeSpanner.h"
#include "GEMABubble.h"
// An implementation of the GEMA layout algorithm, based on
// code by Arne Frick placed in the public domain.  See the
// GEMA.h file for further details.

using namespace std;

#define MMIN(x,y)   ((x < y) ? (x) : (y))
#define MMAX(x,y)   ((x < y) ? (y) : (x))
#define MABS(v)     ((v > 0) ? (v) : (-(v)))

long arad;

LAYOUTPLUGIN(GEMABubble,"GEM (Frick) Bubble Area","David Duke","23/07/2001","Alpha","0","1")

GEMABubble::GEMABubble(const PropertyContext &context) : Layout(context) 
{
    i_maxtemp      = IMAXTEMPDEF;
    a_maxtemp      = AMAXTEMPDEF;
    o_maxtemp      = OMAXTEMPDEF;
    i_starttemp    = ISTARTTEMPDEF;
    a_starttemp    = ASTARTTEMPDEF;
    o_starttemp    = OSTARTTEMPDEF;
    i_finaltemp    = IFINALTEMPDEF;
    a_finaltemp    = AFINALTEMPDEF;
    o_finaltemp    = OFINALTEMPDEF;
    i_maxiter      = IMAXITERDEF;
    a_maxiter      = AMAXITERDEF;
    o_maxiter      = OMAXITERDEF;
    i_gravity      = IGRAVITYDEF;
    i_oscillation  = IOSCILLATIONDEF;
    i_rotation     = IROTATIONDEF;
    i_shake        = ISHAKEDEF;
    a_gravity      = AGRAVITYDEF;
    a_oscillation  = AOSCILLATIONDEF;
    a_rotation     = AROTATIONDEF;
    a_shake        = ASHAKEDEF;
    o_gravity      = OGRAVITYDEF;
    o_oscillation  = OOSCILLATIONDEF;
    o_rotation     = OROTATIONDEF;
    o_shake        = OSHAKEDEF;
    
    Map = NULL;
    Q   = NULL;
}

GEMABubble::~GEMABubble()
{}


int GEMABubble::select()  {
  int    u, n, v;
  if (Iteration == 0)  {
    if (Map) delete [] Map;
    Map = new int[NodeCount];
    for (int i = 0; i < NodeCount; i++)
      Map[i] = i;
  }
  n = NodeCount - (Iteration % NodeCount);
  v = rand () % n;  // was 1 + rand() % n due to numbering in GEM
  if (v == NodeCount) v--;
  if (n == NodeCount) n--;
  u = Map[v]; Map[v] = Map[n]; Map[n] = u;
  return u;
}

void GEMABubble::vertexdata_init(const float starttemp, long int leninit) {
  GEMparam *gemP;

  Temperature = 0;
  Center.x = 0;
  Center.y = 0;

  for (int v = 0; v < NodeCount; v++) {
    gemP = GemProp + v;
    gemP->heat = starttemp * leninit;
    Temperature += (long)(gemP->heat * gemP->heat);
    gemP->imp.x = gemP->imp.y = 0;
    gemP->dir  = 0;
    gemP->mass = 1 + gemP->mass / 3;
    Center.x += gemP->pos.x;
    Center.y += gemP->pos.y;
  }
}


void GEMABubble::displace(int v, Vector imp) {
  long int t, n;
  GEMparam *gemP;

  if (imp.x != 0 || imp.y != 0) {
    n = MMAX( labs(imp.x), labs(imp.y)) / 16384L;
    if (n > 1) {
      imp.x /= n;
      imp.y /= n;
    }
    gemP = GemProp + v;
    t  = (long)(gemP->heat);
    n  = (long)sqrt((long double)(imp.x*imp.x + imp.y*imp.y));
    imp.x = imp.x * t / n;
    imp.y = imp.y * t / n;
    gemP->pos.x += imp.x;
    gemP->pos.y += imp.y;
    Center.x += imp.x;
    Center.y += imp.y;

    n = t * (long)sqrt((long double) (gemP->imp.x*gemP->imp.x + 
				      gemP->imp.y*gemP->imp.y));

    if (n) {
      Temperature -= t * t;
      t += (long)(t * Oscillation * 
		  (imp.x * gemP->imp.x + imp.y * gemP->imp.y) / n);
      t = MMIN(t, Maxtemp);
      gemP->dir += Rotation * 
	(imp.x * gemP->imp.y - imp.y * gemP->imp.x) / n;
      t -= (long)(t * MABS(gemP->dir) / NodeCount);
      t = MMAX(t, 2L);
      Temperature += t * t;
      gemP->heat = t;
    }
    gemP->imp = imp;
  }
}


void GEMABubble::a_round() {
  vector<int>::iterator nodeSet;
  int u, v;
  Vector imp, d, pos;
  long int    n;
  GEMparam *gemP, *gemQ;

  for (int i = 0; i < NodeCount; i ++) {
    v = this->select();
    gemP = GemProp + v;

    //long int pRad = static_cast <long> (ceil (gemP->rad));

    pos = gemP->pos;

    n = (long)ceil(a_shake * ELEN);
    imp.x = rand () % (2 * n + 1) - n;
    imp.y = rand () % (2 * n + 1) - n;
    imp.x += (long)((Center.x / NodeCount - pos.x) * gemP->mass *
		    a_gravity);
    imp.y += (long)((Center.y / NodeCount - pos.y) * gemP->mass *
		    a_gravity);

    for (int j = 0; j < NodeCount; j++) {
      gemQ = GemProp + j;
      long int radSum = static_cast <long int> (ceil (gemP->rad + gemQ->rad)); 
      d.x = pos.x - gemQ->pos.x;
      d.y = pos.y - gemQ->pos.y;
      n  = d.x * d.x + d.y * d.y;
      if (n) {
	long sum = ELEN + radSum; //pRad + qRad;
	imp.x += d.x * (sum*sum/n);
	imp.y += d.y * (sum*sum/n);
      }
    }
    nodeSet = Adjacent[v].begin();
    while (nodeSet < Adjacent[v].end()) {
      u  = *nodeSet++;
      gemQ = GemProp + u;
      //long int qRad = static_cast <long int> (ceil (gemQ->rad));
      long int radSum = static_cast <long int> (ceil (gemP->rad + gemQ->rad));
      d.x = pos.x - gemQ->pos.x;
      d.y = pos.y - gemQ->pos.y;
      n  = (long)((d.x * d.x + d.y * d.y) / gemP->mass);
      n  = MMIN(n, MAXATTRACT);  // 1048576L  (N2)
      long sum = ELEN + radSum; //pRad + qRad;
      imp.x -= d.x * n / (sum*sum);
      imp.y -= d.y * n / (sum*sum);
    }
    this->displace(v, imp);
    Iteration++;
  }
}

void GEMABubble::arrange() {
    long int stop_temperature;
    unsigned long stop_iteration;

    this->vertexdata_init(a_starttemp, ELEN);

    Oscillation      = a_oscillation;
    Rotation         = a_rotation;
    Maxtemp          = (long)ceil(a_maxtemp * ELEN);
    stop_temperature = (long)(a_finaltemp * a_finaltemp * ELEN* NodeCount);
    stop_iteration   = 20/*sqrt(NodeCount)*/; 
    Iteration        = 0;
    unsigned int itr = 0;
    
    while (Temperature > stop_temperature && itr++ < stop_iteration) {
      updateLayout();
      if (pluginProgress->progress(Iteration,stop_iteration/2)!=TLP_CONTINUE) return;
      this->a_round();
    }
}

void GEMABubble::updateLayout(){
  for (int i = 0; i < NodeCount; i++) {
    GEMparam p = GemProp[ i ];
    layoutProxy->setNodeValue(
			      Invmap[i],
			      Coord( p.pos.x, p.pos.y, 0 ) );
  }
}
bool GEMABubble::run() {
  node n;
  GEMparam p;
  Iterator<node> *nodes;
  Iterator<node> *neighbors;

  //FOR REPEATABLE LAYOUTS
  srand (time(0));

  SizesProxy *sizes = superGraph->getProperty <SizesProxy> ("viewSize");

  NodeCount = superGraph->numberOfNodes();
  layoutProxy->setAllEdgeValue(vector<Coord>(0));
  GemProp  = new GEMparam[NodeCount];
  Invmap   = new node[NodeCount];
  Adjacent = new vector<int>[NodeCount];
    
  stdext::hash_map<node, int, stdext::hash<node> > nodeNumbers(NodeCount);

  double sum = 0;
  
  nodes = superGraph->getNodes();
  for (int i = 0; nodes->hasNext(); i++) {
    n = nodes->next();
    GemProp[i] = GEMparam(superGraph->deg(n));
    Size curSize = sizes->getNodeValue (n); 
    GemProp[i].rad = sqrt ((curSize.getW()/2.0)*(curSize.getW()/2.0) +
			   (curSize.getH()/2.0)*(curSize.getH()/2.0));
    if (GemProp[i].rad < 1) GemProp[i].rad = 1;
    sum += GemProp[i].rad;
    Invmap[i]  = n;
    nodeNumbers[n] = i;
  } delete nodes;

  clock_t start = clock();

  //compute the average radus and set ELEN
  ELEN = static_cast <long int> (ceil (sum/static_cast<double>(superGraph->numberOfNodes())));
  INIT_ELEN = ELEN*15;
  ELEN *= 15;

  for (int i = 0; i < NodeCount; i++) {
    neighbors   = superGraph->getInOutNodes(Invmap[i]);
    while (neighbors->hasNext()) {
      n = neighbors->next();
      Adjacent[i].push_back( nodeNumbers[n] );
    } delete neighbors;
  }
     
  string errMsg;
  DataSet data;
  SuperGraph *spanTree = NULL;
  Iterator<SuperGraph *> *subGraphs = superGraph->getSubGraphs();
  while (subGraphs->hasNext()) {
    SuperGraph *curSubGraph = subGraphs->next();
    if (curSubGraph->getAttribute <string> ("name") == "Spanning Tree")
      spanTree = curSubGraph;
  } delete subGraphs; //end while

  if (spanTree == NULL) {
    cerr << "Did not find spanning tree.  Computing it now." << endl;
    SelectionProxy sel (superGraph);
    TreeSpanner spanningTree(superGraph, &sel);
    spanTree = spanningTree.graph;
  }
  
  LayoutProxy localLayout(spanTree);
  spanTree->computeProperty ("Bubble Tree", &localLayout, errMsg,
			     NULL, &data);

  nodes = spanTree->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord pos = localLayout.getNodeValue (curNode);
    GemProp[nodeNumbers[curNode]].pos.x = (long)ceil(pos.getX());
    GemProp[nodeNumbers[curNode]].pos.y = (long)ceil(pos.getY());
  }delete nodes;//end while

  if (pluginProgress->progress(100,100) == TLP_CONTINUE) 
    if (a_finaltemp < a_starttemp)
      this->arrange();
  if (pluginProgress->progress(100,100)==TLP_CONTINUE) 
    updateLayout();
  
  clock_t end = clock();
  double elapsedTime = (double) (end - start) / (double) (CLOCKS_PER_SEC);
  cout << "elapsed time " << elapsedTime << endl;
  delete [] GemProp;
  delete [] Invmap;
  delete [] Adjacent;
  if (Map) delete [] Map;
  if (Q) delete Q;
  
  return true; //pluginProgress->state()!=TLP_CANCEL;
}


bool GEMABubble::check(string &erreurMsg) {
    return(true);
}

void GEMABubble::reset()
{}
