
#include <iostream>
#include <string>
#include <math.h>

using namespace std;

#include "BiconFilter.h"
#include <HierarchyMap.h>
#include <tulip/TlpTools.h>
#include <tulip/StableIterator.h>

METRICPLUGIN(BiconFilter,"Bicon Filter","Daniel Archambault","01/11/2005","Alpha","0","1");

//A pair of ids for the biconnected Bridge
StringProxy *bridge;

BiconFilter::BiconFilter(const PropertyContext &context):Metric(context) {}

BiconFilter::~BiconFilter(){}

inline void BiconFilter::copyNodeProperties (node oldNode, node newNode) {
  Iterator <string> *properties = superGraph->getProperties();
  while (properties->hasNext()) {
    string prop_name = properties->next();
    PProxy *curProp = superGraph->getProperty (prop_name);
    curProp->setNodeStringValue (newNode, 
				 curProp->getNodeStringValue (oldNode));
  } delete properties; 
}//end copyNodeProperties

inline void BiconFilter::copyEdgeProperties (edge oldEdge, edge newEdge) {
  Iterator <string> *properties = superGraph->getProperties();
  while (properties->hasNext()) {
    string prop_name = properties->next();
    PProxy *curProp = superGraph->getProperty (prop_name);
    curProp->setEdgeStringValue (newEdge, 
				 curProp->getEdgeStringValue (oldEdge));
  } delete properties;
}//end copyEdgeProperties

void 
BiconFilter::fixArticulationVertex (node n, vector <int> &non_bridge_labels,
				    __gnu_cxx::hash_map <int, unsigned int> 
				    &label_count,
				    MutableContainer <bool> &leaveEdge,
				    MetricProxy &tempMetric) {

  //a proxy for the original edge attribute
  StringProxy *originalEdges = 
    superGraph->getProperty <StringProxy> (ORIGINAL_EDGE_NAME);

  //a vertex for each non bridge label
  node clone_nodes [non_bridge_labels.size()];

  //hash for index given metric value
  __gnu_cxx::hash_map <int, unsigned int> metric_to_clone;

  //hash for metric value the list of edge indexes as string
  __gnu_cxx::hash_map <int, string> originalEdgeList;
  
  //add the non_bridge_labels.sizeclones and clone their properties
  for (unsigned int i = 0; i < non_bridge_labels.size(); i++) {
    clone_nodes[i] = superGraph->addNode();
    copyNodeProperties (n, clone_nodes[i]);
  }//end for i

  //set the metricProxy value
  MetaGraphProxy *featureClones = 
    superGraph->getProperty <MetaGraphProxy> (FEATURE_NODE_NAME);
  for (unsigned int i = 0; i < non_bridge_labels.size(); i++) {
    tempMetric.setNodeValue (clone_nodes[i], non_bridge_labels[i]);
    featureClones->setNodeValue (clone_nodes[i], NULL);
    metric_to_clone [non_bridge_labels[i]] = i;
  }//end for i

  char charID[10]; //edge ID buffer

  //remove all the edges around n that are not leaf edges
  StableIterator <edge> delEdges (superGraph->getInOutEdges (n));
  while (delEdges.hasNext()) {
    edge curEdge = delEdges.next();

    //get edge label
    int curLabel = 
      static_cast <int> (floor (tempMetric.getEdgeValue (curEdge)));

    //get the other vertex on this edge
    node opposite = superGraph->opposite (curEdge, n);

    //keep all bridges on n
    if (label_count[curLabel] < 2) continue;
   
    if (!leaveEdge.get (curEdge.id)) {
      
      //add the original edge values to this metric value
      if (originalEdgeList.find (curLabel) == originalEdgeList.end())
	originalEdgeList[curLabel] = "";
      if (originalEdges->getEdgeValue (curEdge) != "") {
	if (originalEdgeList[curLabel] != "")
	  originalEdgeList[curLabel] = originalEdgeList[curLabel] 
	    + " " + originalEdges->getEdgeValue (curEdge);
	else
	  originalEdgeList[curLabel] = originalEdges->getEdgeValue (curEdge);
      }//end if
      
      //add the edge between clone and the opposite vertex
      edge added = 
	superGraph->addEdge (opposite, 
			     clone_nodes[metric_to_clone [curLabel]]);
      tempMetric.setEdgeValue (added, curLabel);
      copyEdgeProperties (curEdge, added);
      originalEdges->setEdgeValue (added, "");

      //delete the current edge
      superGraph->delEdge (curEdge);
    }//end if
  }//end while

  //add a new edge to complete the articulation vertex duplication
  for (unsigned int i = 0; i < non_bridge_labels.size(); i++) {
    edge added = superGraph->addEdge (n, clone_nodes[i]);
    sprintf (charID, "%d", added.id);
    originalEdges->setEdgeValue (added, 
				 originalEdgeList[non_bridge_labels[i]]);
    leaveEdge.set (added.id, true);
  }//end for i
}//end fixArticulationVertices

bool BiconFilter::run() {
  //for error messages
  string errMsg;
  if (superGraph->numberOfNodes() == 1) {
    metricProxy->setAllNodeValue (0);
    return true;
  }//end if

  //call the biconnected component metric on tempMetric
  MetricProxy tempMetric (superGraph);

  bool success = 
    superGraph->computeProperty ("Biconnected Componnent", 
				 &tempMetric, errMsg);
  if (!success) {
    cerr << "Error in BiconFilter" << endl;
    cerr << "Biconnected Component returned the following error:\n" 
	 << errMsg << endl;
    return false;
  }//end if

  bridge = 
    superGraph->getLocalProperty<StringProxy> (BRIDGE_PROXY);
  bridge->setAllEdgeValue ("");

  //a vector for strict articulation vertices
  vector <node> strictArtVerts;

  //a mutable container indicating if this edge should be deleted.
  //we must delete all edges incident to an articulation vertex that
  //this plugin did not create.
  MutableContainer <bool> leaveEdges;
  leaveEdges.setAll (false);

  //a int value representing the max label appearing in the graph
  int maxAllLabels = 0;

  //Go through all the nodes and put this node on the larges bicon
  //component in which it participates.
  StableIterator <node> biconnodes (superGraph->getNodes());
  while (biconnodes.hasNext()) {
    node curNode = biconnodes.next();

    //for counting labels on edges incident to the node
    __gnu_cxx::hash_map <int, unsigned int> label_count;

    //a vector of non-bridge labels to this node
    vector <int> non_bridge_labels;

    //a vector of unique labels
    vector <int> unique_labels;

    //get the number of labels around this vertex
    StableIterator <edge> edges (superGraph->getInOutEdges (curNode));
    while (edges.hasNext()) {
      edge curEdge = edges.next();
      int curLabel = 
	static_cast <int> (floor (tempMetric.getEdgeValue (curEdge)));

      //see if it is the max of all labels
      if (curLabel > maxAllLabels) maxAllLabels = curLabel;

      //if the current label does not exist, init the counter record the
      //label
      if (label_count.find (curLabel) == label_count.end()) {
	label_count[curLabel] = 1;
	unique_labels.push_back (curLabel);
      }//end if
      else ++label_count[curLabel];
    } //end while

    //find the most common label
    double max_label;
    unsigned int max_label_count = 0;
    for (unsigned int i = 0; i < unique_labels.size(); i++) {
      //find the max
      if (max_label_count < label_count[unique_labels[i]]) {
	max_label = unique_labels[i];
	max_label_count = label_count[unique_labels[i]];
      }//end if 
      if (label_count[unique_labels[i]] > 1)  
	non_bridge_labels.push_back (unique_labels[i]);
    } //end for i

    //tree vertex, label unique like strict art
    if ((unique_labels.size() > 1) && (max_label_count == 1))
      strictArtVerts.push_back (curNode);

    //if there is more than one non-bridge, this is an articulation
    //vertex that needs to be duplicated
    if (unique_labels.size() > 1) {
      strictArtVerts.push_back (curNode);
      fixArticulationVertex (curNode, non_bridge_labels, label_count,
			     leaveEdges, tempMetric);
    }//end if
    
    //otherwise, set the label of the node
    else tempMetric.setNodeValue (curNode, max_label);
    
    //if there is more than one unique label, it is a bridge or an 
    //articulation vertex.  set the edge labels
    if (unique_labels.size() > 1) {
      char bridge_ends [20];
      Iterator<edge> *bridge_edges = superGraph->getInOutEdges (curNode);
      while (bridge_edges->hasNext()) {
	edge curBridgeEdge = bridge_edges->next();
	sprintf (bridge_ends, "%d %d", superGraph->source (curBridgeEdge).id,
		 superGraph->target (curBridgeEdge).id);
	bridge->setEdgeValue (curBridgeEdge, bridge_ends);
      } delete bridge_edges;
    }//end if
  } //end while 
    
  //load the edges into the metric proxy
  Iterator <edge> *edgeload = superGraph->getEdges ();
  while (edgeload->hasNext()) {
    edge curEdge = edgeload->next();
    metricProxy->setEdgeValue (curEdge, tempMetric.getEdgeValue (curEdge));
  } delete edgeload; //end while

  //load the nodes
  Iterator <node> *nodeload = superGraph->getNodes ();
  while (nodeload->hasNext()) {
    node curNode = nodeload->next();
    metricProxy->setNodeValue (curNode, tempMetric.getNodeValue (curNode));
  } delete nodeload;
  
  //uniquely label the strict articulation vertices
  ++maxAllLabels;
  for (unsigned int i = 0; i < strictArtVerts.size(); i++)
    metricProxy->setNodeValue (strictArtVerts[i], ++maxAllLabels);
  
  //return success
  return true;
}








