
/// BoundingObjects.cpp
/** 
 *
 *  Defines required by Shrink Resolve Overlap and Compact Graph
 *  to compute the bounding objects of nodes
 *
 *  HISTORY:
 *
 *  The code was first implemented in February 2004 by Daniel Archambault.
 *
 *  NOTES:
 *
 *  AUTHOR:
 *
 *  Daniel Archambault, University of British Columbia, CA: 
 *  Email: archam@cs.ubc.ca
 *  Version 0.1: 21 February 2005: Initial version
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */ 

#include <tulip/DrawingTools.h>
#include <tulip/ConvexHull.h>
#include <tulip/MetaGraphProxy.h> //DEBUG
#include <tulip/StableIterator.h> //DEBUG
#include "BoundingObjects.h"

using namespace std;

typedef vector<Coord>::iterator coordIt;
typedef vector<Coord>::const_iterator const_coordIt;
typedef vector<unsigned int>::iterator unsignedIntIt;

//A function to return the magnitude of the radius of a node along
//a direction 'dir' of a boundary with the passed scale and rotation
//the boundary polygon must be zero centred, unscaled, and unrotated.
Coord getRad (const string &boundType, Coord dir, double scale, 
	      double rotation, const Boundary &boundary) {
  if (dir.norm() > 1e-16) dir /= dir.norm();

  //if the bounding type is circle, return the node radius
  if (boundType == BOUNDING_CIRCLE) {
    double half_w = boundary[0].getX() / 2.0;
    double half_h = boundary[0].getY() / 2.0;
    return 
      static_cast <float> (scale*sqrt (half_w*half_w + half_h*half_h))*dir;
  }//end if

  //figure out the "real" direction on an unrotated node and store it in
  //the angle theta
  double rotation_rad = toRad (rotation);
  double theta = unwindAngle (rotation_rad + vecToAngle(dir));
 
  //find the two boundary vertices of the boundary that are to either 
  //side of dir (assumes boundary is zero centered)
  unsigned int index = 0;
  const_coordIt previous = boundary.begin();
  const_coordIt next = ++boundary.begin();
  while (next != boundary.end()) {
    if ((vecToAngle (*next) > theta) && (vecToAngle (*previous)) < theta) 
      break;
    index++;
    previous = next;
    ++next;
  }//end for previous 

  //compute the radius length for the specified dir
  //equations used:
  //(x, y) = t*rot_dir (as boudary is zero-centred)
  //(x, y) = *previous + t_2*(*next - previous)
  unsigned int next_index = (index + 1 == boundary.size()) ? 0 : (index + 1);
  Coord rot_dir = angleToVec (theta);
  Coord v = boundary[next_index] - boundary[index];
  double t;
  if (fabs (v.getX()) < 1e-16) t = boundary[index].getX() / rot_dir.getX();
  else {
    double vyvx = v.getY() / v.getX();
    //this only happens when vectors are parallel! should not hapen
    if (fabs (rot_dir.getY() - vyvx*rot_dir.getX()) < 1e-16) {
      cerr << "rot_dir and v are parallel!  There is a bug in getRad "
	   << "rot_dir " << rot_dir << " v " << v << endl;
      return dir;
    }//end if
    else t = (boundary[index].getY() - vyvx*boundary[index].getX()) /
	   (rot_dir.getY() - vyvx*rot_dir.getX());
  }//end else
  return static_cast <float> (scale*t)*dir;
}//end getRad

//computes a bounding box boundary for the specified list of points.  It
//places the centre of the new bounding box in 'centre'
inline Boundary computeBoundingBox (vector<Coord> &points, Coord &centre) {
  Coord max (-DBL_MAX, -DBL_MAX, 0), min (DBL_MAX, DBL_MAX, 0);
  for (coordIt pointcoord = points.begin(); 
       pointcoord != points.end(); ++pointcoord) {
    max.setX (std::max (pointcoord->getX(), max.getX()));
    max.setY (std::max (pointcoord->getY(), max.getY()));
    min.setX (std::min (pointcoord->getX(), min.getX()));
    min.setY (std::min (pointcoord->getY(), min.getY()));
  }//end for
  centre = (max + min)/2.0;
 
  Boundary boxBoundary;
  boxBoundary.push_back (max);
  boxBoundary.push_back (Coord (min.getX(), max.getY(), 0));
  boxBoundary.push_back (min);
  boxBoundary.push_back (Coord (max.getX(), min.getY(), 0));
  return boxBoundary;
}//end computeBoundingBox

//computes a convex hull boundary for the specified list of points.  It
//places the centre of the new convex hull in 'centre'
inline Boundary computeConvexHull (vector<Coord> &points, Coord &centre) {
  vector<unsigned int> convexHull;
  tlp::convexHull (points, convexHull);
  Boundary hullBoundary;
  for (unsignedIntIt hullVert = convexHull.begin();
       hullVert != convexHull.end(); ++hullVert) {
    centre += points[*hullVert];
    hullBoundary.push_back (points[*hullVert]);
  }//end for
  centre = centre/static_cast<double> (convexHull.size());
  return hullBoundary;
}//end computeConvexHull

//Updates the sizes in the boundary table entry for the specifed node n
//with the passed metagraph using boundingType.
void updateBoundary (BoundaryTable &boundaryTable, node n,
		     SuperGraph *metagraph,
		     std::string &boundingType) {
  //if the metagraph of this node is null the bounding object cannot change
  //as it has no interior nodes affected by compaction or overlaps
  if (metagraph == NULL) return;
 
  //use the boundaries of the nodes in the metagraph to create a new boundary
  //for the node n
  LayoutProxy *metaLayout = 
    metagraph->getLocalProperty<LayoutProxy> ("viewLayout");
  MetricProxy *metaRotation =
    metagraph->getLocalProperty<MetricProxy> ("viewRotation");
  MetricProxy *metaScale =
    metagraph->getLocalProperty<MetricProxy> ("viewScale");
 
  StableIterator<node> nodes (metagraph->getNodes());
  vector<Coord> points;
  while (nodes.hasNext()) {
    node curNode = nodes.next();
    Coord pos = metaLayout->getNodeValue (curNode);
    double rotation = metaRotation->getNodeValue (curNode);
    double scale = metaScale->getNodeValue (curNode);
    
    //rotate the boundary into this coordinate frame and load the
    //verts into the points vector
    Boundary &boundary = boundaryTable[curNode.id];
    for (coordIt verts = boundary.begin(); verts != boundary.end(); ++verts) {
      points.push_back (pos + rotate (rotation, scale*(*verts)));
    }
  } //end while

  //compute the updated bounding object
  Coord centre (0, 0, 0);
  if ((boundingType == BOUNDING_BOX) || (boundingType == BOUNDING_CIRCLE))
    boundaryTable[n.id] = computeBoundingBox (points, centre);
  else if (boundingType == BOUNDING_CONVEX_HULL) 
    boundaryTable[n.id] = computeConvexHull (points, centre);

  //centre the subgraph and bounding object around the new node centre.
  Iterator<node> *centreNodes = metagraph->getNodes();
  while (centreNodes->hasNext()) {
    node curNode = centreNodes->next();
    metaLayout->setNodeValue (curNode,
			      metaLayout->getNodeValue (curNode) - centre);
  } delete centreNodes; //end while
  for (coordIt boundvert = boundaryTable[n.id].begin();
       boundvert != boundaryTable[n.id].end(); boundvert++) 
    *boundvert -= centre;
}//updateBoundary

