#ifndef __BOUNDING_OBJECTS_H
#define __BOUNDING_OBJECTS_H

/// BoundingObjects.h
/** 
 *
 *  Defines required by Shrink Resolve Overlap and Compact Graph
 *  to compute the bounding objects of nodes
 *
 *  HISTORY:
 *
 *  The code was first implemented in February 2004 by Daniel Archambault.
 *
 *  NOTES:
 *
 *  AUTHOR:
 *
 *  Daniel Archambault, University of British Columbia, CA: 
 *  Email: archam@cs.ubc.ca
 *  Version 0.1: 21 February 2005: Initial version
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */
#include <tulip/LayoutProxy.h>
#include <tulip/SizesProxy.h>
#include <tulip/MetricProxy.h>
#include <tulip/SuperGraph.h>
#include <tulip/MutableContainer.h>
#include <vector>
#include <iostream>
#include <math.h>
#include <string>
#if __GNUC__ == 3 && __GNUC_MINOR__ > 0
#include <ext/hash_map>
#else
#include <hash_map>
#endif

//a dataset value that returns a pointer to the dependent table
//for a specific level of the hierarchy
#define DATASET_BOUNDARY_TABLE ("Boundary Table")

//a dataset value that defines the bounding type requested
#define DATASET_BOUNDING_TYPE ("Bounding Type")

//a dataset value to indicate we have resolved an overlap
#define DATASET_SHRINK_RESOLVE_TABLE ("Resolve Overlap Table")

//pass in a list of nodes that require all pairs to be considered
#define DATASET_ALL_PAIRS ("Compact/Shrink All Pairs")

//defines for bounding circle, box etc...
#define BOUNDING_CIRCLE ("Circle")
#define BOUNDING_BOX ("Box")
#define BOUNDING_CONVEX_HULL ("Convex Hull")

//A type for the boundaries of nodes and a table that stores them.  The
//table is indexed by node id.
typedef std::vector<Coord> Boundary;
typedef __gnu_cxx::hash_map<unsigned int, Boundary> BoundaryTable;

//A type for tables holding if we have done a shrink resolve overlap
typedef __gnu_cxx::hash_map<const char *, bool> shrinkResTable;

//an inline operator to multiply a Coord by a scalar
inline Coord operator* (double s, Coord c) {
  return Coord (s*c.getX(), s*c.getY(), s*c.getZ());
}//end operator

//A function to return the angle made between a unit vector and
//the x-axis.
inline double vecToAngle (Coord vec) {
  if (vec.norm() > 1e-6) vec /= vec.norm();
  return (vec.getY() < 0) ? (2*M_PI - acos(vec.getX())) : acos (vec.getX());
}//end vec2angle

//A function to return a unit vector in the direction of the passed
//angle with the x-axis
inline Coord angleToVec (double angle) {
  return Coord (cos(angle), sin(angle), 0);
}//end angleToVec

//A function that takes an angle and reports it in the range [0, 2pi)
inline double unwindAngle (double angle) {
  double factors = floor (fabs(angle) / (2*M_PI));
  if (angle < 0) return angle + 2*M_PI*(factors + 1);
  else if (angle >= 2*M_PI) return angle - 2*M_PI*factors;
  else return angle;
}//end unwindAngle

//a function to convert degrees to radians
inline double toRad (double deg) {
  return deg*(M_PI/180.0);
}//end toRad

//a fundtion to convert radians to degrees
inline double toDeg (double rad) {
  return rad*(180.0/M_PI);
}//end toDeg

//a function to rotate by theta in degreesabout a (0, 0) origin
inline Coord rotate (double theta_deg, const Coord pos) {
  Coord result;
  double theta = toRad (theta_deg);
  result.setX (pos.getX()*cos(theta) - pos.getY()*sin (theta));
  result.setY (pos.getX()*sin(theta) + pos.getY()*cos (theta));
  result.setZ (0);
  return result;
}//end rotate

//A function to return the magnitude of the radius of a node along
//a direction 'dir' of a boundary with the passed scale and rotation
//the boundary polygon must be zero centred, unscaled, and unrotated.
Coord getRad (const std::string &boundType, Coord dir, double scale, 
	      double rotation, const Boundary &boundary);

//A function that returns the scale factor required to resolve an
//overlap between nodes of sizes size1 and size2 and with centres
//c1 and c2.  This is the translation scale factor.
inline double getScaleFactor (const std::string &boundType,
			      Coord c1, Coord c2,
			      double scale1, double scale2,
			      double r1, double r2,
			      const Boundary &boundary1,
			      const Boundary &boundary2) {
  return static_cast<double> 
    ((getRad (boundType, c2 - c1, scale1, r1, boundary1).norm() + 
      getRad (boundType, c1 - c2, scale2, r2, boundary2).norm()) /
    (c1 - c2).norm());
}//end getScaleFactor

//A function taht returns the shrink scale factor to a pair of nodes
//resolving the overlap between them
inline double getShrinkFactor (const std::string &boundType, 
			       Coord c1, Coord c2,
			       double scale1, double scale2,
			       double r1, double r2,
			       const Boundary &boundary1,
			       const Boundary &boundary2) {
  return 1.0/getScaleFactor (boundType, c1, c2, scale1, scale2, r1, r2, 
			     boundary1, boundary2);
}//end getShrinkFactor

//DRAWS THE BOUNDARY OF A NODE.  PRIMARILY FOR DEBUG PURPOSES
inline void drawNodeBoundary (SuperGraph *containingGraph,
			      Boundary &boundary) {
  //draw the bounding object
  SizesProxy *sizes = containingGraph->getProperty<SizesProxy> ("viewSize");
  LayoutProxy *layout = 
    containingGraph->getProperty<LayoutProxy> ("viewLayout");
  node first, curBoundary, prev;
  for (unsigned int i = 0; i < boundary.size(); i++) {
    curBoundary = containingGraph->addNode();
    if (i == 0) first = curBoundary;
    sizes->setNodeValue (curBoundary, Size (0.5,0.5,0.5));
    layout->setNodeValue (curBoundary, boundary[i]);
    if (i != 0) containingGraph->addEdge (prev, curBoundary);
    prev = curBoundary;
  }//end for
  containingGraph->addEdge (curBoundary, first);
}//end drawNodeBoundary

//DRAWS THE RADIUS OF A NODE.  PRIMARILY FOR DEBUG PURPOSES
inline void drawRad (SuperGraph *containingGraph, Coord rad, Coord tail) {
  SizesProxy *sizes = containingGraph->getProperty<SizesProxy> ("viewSize");
  LayoutProxy *layout = 
    containingGraph->getProperty<LayoutProxy> ("viewLayout");
  node headNode = containingGraph->addNode();
  node tailNode = containingGraph->addNode();
  containingGraph->addEdge (tailNode, headNode);
  sizes->setNodeValue (headNode, Size (0.5,0.5,0.5));
  sizes->setNodeValue (tailNode, Size (0.5,0.5,0.5));
  layout->setNodeValue (headNode, tail + rad);
  layout->setNodeValue (tailNode, tail);
}//end drawRad

//Generates a boundary (bounding box) for the node with "size"
inline Boundary leafNodeBoundary (Size nodeSize) {
  Boundary boundary;
  boundary.push_back (Coord (nodeSize.getW()/2.0, nodeSize.getH()/2.0));
  boundary.push_back (Coord (-nodeSize.getW()/2.0, nodeSize.getH()/2.0));
  boundary.push_back (Coord (-nodeSize.getW()/2.0, -nodeSize.getH()/2.0));
  boundary.push_back (Coord (nodeSize.getW()/2.0, -nodeSize.getH()/2.0));
  return boundary;
}//end leafNodeBoundary

//Updates the sizes in the boundary table entry for the specifed node n
//with the passed metagraph using boundingType.
void updateBoundary (BoundaryTable &boundaryTable, node n,
		     SuperGraph *metagraph,
		     std::string &boundingType);
#endif
