#include <cstdlib>
#include <iostream>
#include <cmath>
#include <climits>
#include "XWindow.h"
#include <list>
#include <vector>
#include <tulip/Vector.h>
#include <tulip/MutableContainer.h>
#include <tulip/SuperGraph.h>
#include <tulip/TlpTools.h>
#include <sys/timeb.h>
#include <unistd.h>
#include <stack>
#include <deque>
/**
   Pour compiler :
   tapez make
*/

static const int X_SIZE=800;
static const int Y_SIZE=800;
static const int NB_POINTS=10000;
using namespace std;

//convex hull of a set of point O(nlog(n))
struct ConvexHull2DGraham {
  static void compute(const vector<Coord> &points, vector<Coord> *convexHull) {
    vector<Coord> &cH = *convexHull;
    Coord p0 = findP0(points);
    set<Coord, CmpAngle> orderedSet;
    cH.clear();
    buildOrderedSet(points, orderedSet, p0);
    set<Coord, CmpAngle>::const_iterator it = orderedSet.begin();
    cH.push_back(p0);
    cH.push_back(*it++);
    cH.push_back(*it++);
    for (;it != orderedSet.end(); ++it) {
      while( ! (((*it-cH[cH.size()-1])^(cH[cH.size()-2]-cH[cH.size()-1]))[2]>0) )
	cH.pop_back();
      cH.push_back(*it);
    }
  }
private :  
  struct CmpAngle {
    bool operator()(const Coord &c1, const Coord &c2) {
      return ((c1 - p0)^(c2 - p0))[2] > 0; 
    }
    static Coord p0;
  };
  static void buildOrderedSet(const vector<Coord> &points, set<Coord, CmpAngle> &orderedSet, const Coord &p0) {
    CmpAngle::p0 = p0;
    orderedSet.clear();
    for (vector<Coord>::const_iterator itp = points.begin(); itp!=points.end(); ++itp) {
      if (*itp == p0) continue;
      set<Coord, CmpAngle>::iterator it;
      if ( (it = orderedSet.find(*itp)) != orderedSet.end()) {
	if ( (*itp - p0).norm() > ((*it) - p0).norm() ) {
	  orderedSet.erase(it);
	  orderedSet.insert(*itp);
	}
      }
      else
	orderedSet.insert(*itp);
    }
  }
  static Coord findP0(const vector<Coord> &points) {
    vector<Coord>::const_iterator itp = points.begin();
    Coord p0 = *itp++;
    for (; itp!=points.end(); ++itp) {
      if ( p0[0] < (*itp)[0]) continue;
      if ( p0[0] > (*itp)[0])
	p0 = *itp;
      else 
	if (p0[1] > (*itp)[1])
	  p0 = *itp;
    }
    return p0;
  }
};
Coord ConvexHull2DGraham::CmpAngle::p0 = Coord();

//incremental convex hull building complexity ? O(n^2)
struct ConvexHull2D {
  ConvexHull2D() {
    graph=tlp::newSuperGraph();
  }
  ~ConvexHull2D() {
    delete graph;
  }
  list<Coord> getConvexHull() const {
    list<Coord> tmp;
    if (graph->numberOfNodes()==0) return tmp;
    tmp.push_back(convexHull.get(start.id));
    if (graph->numberOfNodes()==1) return tmp;
    node next=start;
    for (int i=0;i<graph->numberOfNodes();++i){
      next=graph->getOutNode(next,1);
      tmp.push_back(convexHull.get(next.id));
    }
    return tmp;
  }
  void addPoint(float x, float y) {
    Coord p(x,y,0);
    if (graph->numberOfNodes()==0) {
      start=graph->addNode();
      convexHull.set(start.id,p);
      return;
    }
    if (graph->numberOfNodes()==1) {
      node tmp=graph->addNode();
      convexHull.set(tmp.id,p);
      graph->addEdge(start,tmp);
      graph->addEdge(tmp,start);
      return;
    }
    bool modif=false;
    node cur=graph->getOutNode(start,1);
    Coord A=convexHull.get(start.id);
    Coord B=convexHull.get(cur.id);
    node back=start;
    node forw=cur;
    int stopIt=0;
    while(((B-A)^(p-A))[2]<0 && stopIt <graph->numberOfNodes()){
      back=forw;
      forw=graph->getOutNode(forw,1);
      A=B;
      B=convexHull.get(forw.id);
      stopIt ++;
    }
    if (stopIt<graph->numberOfNodes()) {
      modif=true;
      cur=forw;
      start=back;
      //forward
      do {
	forw=cur;
	cur=graph->getOutNode(cur,1);
	A=convexHull.get(forw.id);
	B=convexHull.get(cur.id);
      } while(((B-A)^(p-A))[2]>=0);
      //backward
      cur=start;
      do {
	back=cur;
	cur=graph->getInNode(cur,1);
	B=convexHull.get(back.id);
	A=convexHull.get(cur.id);
      } while(((B-A)^(p-A))[2]>=0);
    }
    if(modif) {
      node next=back;
      stack<node> tmp;
      while((next=graph->getOutNode(next,1))!=forw){
	tmp.push(next);
      }
      if (tmp.empty()) {
	Iterator<edge>*ite=graph->getOutEdges(back);
	edge e=ite->next();
	graph->delEdge(e);
	delete ite;
      } 
      while(!tmp.empty()){
	graph->delNode(tmp.top());
	tmp.pop();
      }
      start = graph->addNode();
      graph->addEdge(back,start);
      graph->addEdge(start,forw);
      convexHull.set(start.id,p);
    }
  }
private:
  SuperGraph *graph;
  MutableContainer<Coord> convexHull;
  node start;
};


void drawConvexHull(const vector<Coord> &convHull, XWindow &win) {
  if (convHull.empty()) return;
  vector<Coord>::const_iterator it=convHull.begin();
  Coord start=*it;
  Coord c1=*it;
  ++it;
  for (;it!=convHull.end();++it){
    Coord c2=*it;
    win.drawLine((uint)c1[0],(uint)c1[1],(uint)c2[0],(uint)c2[1]);
    c1=c2;
  }
  win.drawLine((uint)start[0],(uint)start[1],(uint)c1[0],(uint)c1[1]);
}

int main(void) {
  XWindow win(X_SIZE,Y_SIZE);
  win.setBackground(win.black());//0x444444);
  win.setForeground(win.white());
  win.setForeground(255<<11);
  timeb starttime;
  ftime(&starttime);
  srand(starttime.time);
  vector<Coord> points;
  for (int i=0;i<NB_POINTS;++i) {
    Coord tmp((float)(rand()%X_SIZE)/4.0*(cos(2.0*M_PI*(float)i/(float)NB_POINTS))+(float)X_SIZE/2.0,
	      (float)(rand()%Y_SIZE)/4.0*(sin(2.0*M_PI*(float)i/(float)NB_POINTS))+(float)Y_SIZE/2.0,
	      0);
    points.push_back(tmp);
  }
  //  
  //  ConvexHull2D cv;
  //  ConvexHull2DGraham cv;
  //  cv.addPoint((*itAdd)[0],(*itAdd)[1]);
  //  ++itAdd;
  win.clear();
  win.refresh();
  vector<Coord> tmpPoint;
  vector<Coord>::const_iterator it = points.begin();
  vector<Coord> hull;
  tmpPoint.push_back(*it);
  ++it;
  tmpPoint.push_back(*it);
  ++it;
  tmpPoint.push_back(*it);
  ++it;
  while (it!=points.end()) {
    tmpPoint.push_back(*it);
    ++it;
    win.clear();
    win.setForeground(0x0000FF);
    ConvexHull2DGraham::compute(tmpPoint, &hull);
    drawConvexHull(hull, win);
    win.setForeground(0x00FF00);
    vector<Coord>::iterator itp = points.begin();
    for(;itp != points.end(); ++itp) {
      win.drawPoint((uint)(*itp)[0],(uint)(*itp)[1]);
    }
    win.refresh();
  }
  return EXIT_SUCCESS;
}









