/**
 Author: David Auber
 Email : auber@labri.fr
 Last modification : 20/08/2001
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of Compact Graph to improve information density.  For nodes, 
 * history, author and licence please see the header file Compact.h
 */

#include <math.h>
#include <float.h>
#include "Compact.h"
#include <BoundingObjects.h>
#include <tulip/MetaGraphProxy.h>
#include <HierarchyMap.h>

LAYOUTPLUGIN(Compact,"Compact Graph","David Auber/ Daniel Archambault","25/11/2004","Ok","0","1");

using namespace std;

Compact::Compact(const PropertyContext &context):Layout(context) {
}

Compact::~Compact() {
}

//this function centres a graph around the passed node centre
void Compact::centreGraph (Coord centre) {
  //get the layout proxy and the position of the centre
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  
  //centre nodes around centre
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord position = pos->getNodeValue (curNode);
    pos->setNodeValue (curNode, position - centre);
  }//end while
  delete nodes;
  
  //centre edge bends around centerPos
  Iterator <edge> *edges = superGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next ();
    vector <Coord> bends = pos->getEdgeValue (curEdge);
    for (unsigned int i = 0; i < bends.size (); i++) 
      bends[i] = bends[i] - centre;
    pos->setEdgeValue (curEdge, bends);
  }//end while
  delete edges;
}//end centre

bool Compact::run() {
  //proxy for positions
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  SizesProxy *sizes = superGraph->getProperty <SizesProxy> ("viewSize");
  MetaGraphProxy *hierMetaGraphs = 
    superGraph->getProperty <MetaGraphProxy> (FEATURE_NODE_NAME);

  //A vector of nodes in the graph
  vector <node> v_nodes;

  //put all the nodes in the vector and set all z to zero
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext ()) v_nodes.push_back (nodes->next());
  delete nodes;

  //load all the initial positions into the layout proxy
  for (unsigned int i = 0; i < superGraph->numberOfNodes(); i++)
    layoutProxy->setNodeValue (v_nodes[i], 
			       pos->getNodeValue (v_nodes[i]));
  
  //this is the maximum shrink that can be applied about the centroid of
  //the node such that no two nodes overlap
  double max_allow_shrink = -1;

  //go through every pair of nodes in the vector and determine the
  //shrink factor closest to one but still below it that can be applied
  //to the data to improve information density.
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    for (unsigned int j = i + 1; j < superGraph->numberOfNodes (); j++) {
      //get the positions and sizes of the nodes
      Coord pos_i = pos->getNodeValue (v_nodes[i]);
      Coord pos_j = pos->getNodeValue (v_nodes[j]);
      Size size_i;
      Size size_j;
      size_i = sizes->getNodeValue (v_nodes[i]);
      size_j = sizes->getNodeValue (v_nodes[j]);

      vector <Coord> chull_i, chull_j;
      //convex hull coordinates should be filled in later if this is to
      //be used

      double scale_factor = getScaleFactor (BOUNDING_CIRCLE,
					    size_i, size_j, pos_i, pos_j, 
					    0, 0, chull_i, chull_j);

      //if two nodes overlap, warn that their situation probably won't
      //improve
      if (floor (scale_factor) > 1) {
	cerr << "Warning!  Nodes overlap at the beginning of Compact." << endl 
	     << "They will overlap after Compact is complete.  "
	     << "Ignored in scale calculation." << endl;
	cerr << (pos_i - pos_j).norm() << " " 
	     << (getRad (BOUNDING_CIRCLE, size_i, pos_i, pos_j - pos_i, 
			 0, chull_i) + 
		 getRad (BOUNDING_CIRCLE, size_j, pos_i, pos_i - pos_j, 
			 0, chull_j)) << endl;
	continue;
      }//end if

      //see if the scale factor is the maximum shrink factor seen
      //so far
      if (scale_factor > max_allow_shrink)
	max_allow_shrink = scale_factor;
    }//end for i
  }//end for j

  //if max_allow_shrink < 0, we cannot apply any scaling.  We are done.
  if ((max_allow_shrink < 0) || (max_allow_shrink > 1)) return true;
  
  //find the centre of this graph by coordinate
  Coord max = pos->getMax (superGraph);
  Coord min = pos->getMin (superGraph);
  Coord centre;
  centre.setX ((max.getX() + min.getX())/2.0);
  centre.setY ((max.getY() + min.getY())/2.0);
  centre.setZ (0);
  centreGraph (centre);
  
  //apply the max allowed shrink to all the nodes of the graph radially
  //out from the centre of the graph
  cerr << "compacting graph by " << max_allow_shrink << " about "
       << centre << endl;
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    Coord curPos = pos->getNodeValue (v_nodes[i]);
    curPos *= max_allow_shrink;
    layoutProxy->setNodeValue (v_nodes[i], curPos);
  }//end for i
  return true;
}//end run
