/**
 Author: David Auber
 Email : auber@labri.fr
 Last modification : 20/08/2001
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of Compact Graph to improve information density.  For nodes, 
 * history, author and licence please see the header file Compact.h
 */

#include <math.h>
#include <float.h>
#include <algorithm>
#include "CompactFinalLayout.h"
#include <tulip/TlpTools.h>
#include <tulip/StableIterator.h>

LAYOUTPLUGIN(CompactFinalLayout,"Compact Final Layout","Daniel Archambault","25/11/2004","Ok","0","1");

using namespace std;

typedef vector<node>::iterator nodeIt;

CompactFinalLayout::CompactFinalLayout(const PropertyContext &context):Layout(context) {
}

CompactFinalLayout::~CompactFinalLayout() {
}

//Compacts a particular level of the final hierarchical layout.
bool CompactFinalLayout::compactLevel(SuperGraph *graph, 
				      string &compactType) {  
  string name = graph->getAttribute<string> ("name");
  //if shrink resolve has already been successful on this graph, we 
  //cannot compact it.  Return true
  if ((*shrinkResolveTable) [name.c_str()]) {
    cout << "resolved overlap" << endl;
    return true;
  }

  //decide if atLeastOneHierarchyNode is required
  bool atLeastOneHierarchyNode;
  if (compactAllPairs == NULL) 
    atLeastOneHierarchyNode = false;
  
  else {
    atLeastOneHierarchyNode = true;
    for (unsigned int i = 0; i < compactAllPairs->size(); i++)
      if (name.find ((*compactAllPairs)[i]) == 0)
	atLeastOneHierarchyNode = false;
  }//end else
  
  //this is the maximum shrink that can be applied about the centroid of
  //the node such that no two nodes overlap
  double max_allow_shrink = -1;

  //put all the nodes of the graph inside a vector
  vector<node> nodeList;
  Iterator<node> *nodes = graph->getNodes();
  while (nodes->hasNext()) nodeList.push_back (nodes->next());
  delete nodes;
  
  MetaGraphProxy *metagraphs = 
    graph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  SizesProxy *nodeSize =
    graph->getProperty<SizesProxy> ("viewSize");
  LayoutProxy *nodeLayout = 
    graph->getProperty<LayoutProxy> ("viewLayout");
  MetricProxy *nodeRotation = 
    graph->getProperty<MetricProxy> ("viewRotation");
  MetricProxy *nodeScale = 
    graph->getProperty<MetricProxy> ("viewScale");

  //go through every pair of nodes in the vector or every pair of 
  //hierarchy and one other node
  for (unsigned int i = 0; i < nodeList.size(); i++) {
    //only compute shrink if one node is a hierarchy node 
    //atLeastOneHierarchyNode is true
    node &node_i = nodeList[i];
    if (atLeastOneHierarchyNode && 
	(metagraphs->getNodeValue (node_i) == NULL)) continue;

    for (unsigned int j = i + 1; j < nodeList.size(); j++) {
      node &node_j = nodeList[j];

      //get the positions and sizes of the nodes
      Coord pos_i = nodeLayout->getNodeValue (node_i);
      Coord pos_j = nodeLayout->getNodeValue (node_j);
      double rot_i = nodeRotation->getNodeValue (node_i);
      double rot_j = nodeRotation->getNodeValue (node_j);
      double scale_i = nodeScale->getNodeValue (node_i);
      double scale_j = nodeScale->getNodeValue (node_j);

      //compute the scale factor
      double scale_factor = getScaleFactor (compactType, pos_i, pos_j, 
					    scale_i, scale_j,
					    rot_i, rot_j,
					    (*boundaryTable)[node_i.id], 
					    (*boundaryTable)[node_j.id]);
	

      //if two nodes overlap, warn that their situation probably won't
      //improve
      if (floor (scale_factor) > 1) {
	Coord rad1 = getRad (compactType, pos_i, scale_i, rot_i,
			     (*boundaryTable)[node_i.id]);
	Coord rad2 = getRad (compactType, pos_j, scale_j, rot_j,
			     (*boundaryTable)[node_j.id]);

	cerr << "Warning!  Nodes overlap at the beginning of CompactFinalLayout." << endl 
	     << "They will overlap after CompactFinalLayout is complete.  "
	     << "Ignored in scale calculation." << endl;
	cerr << "distance " << (pos_i - pos_j).norm() << " " 
	     << "sum of radii " 
	     << (rad1 + rad2).norm() << endl;
	continue;
      }//end if

      //see if the scale factor is the maximum shrink factor seen
      //so far
      if (scale_factor > max_allow_shrink)
	max_allow_shrink = scale_factor;
    }//end for i
  }//end for j

  //find the centre of this graph by coordinate
  double maxx = -DBL_MAX, minx = DBL_MAX, maxy = -DBL_MAX, miny = DBL_MAX;
  for (unsigned int i = 0; i < nodeList.size(); i++) {
    //see if they extend the bounding box of the node
    Coord curPos = nodeLayout->getNodeValue (nodeList[i]);
    if (maxx < curPos.getX()) maxx = curPos.getX();
    if (maxy < curPos.getY()) maxy = curPos.getY();
    if (minx > curPos.getX()) minx = curPos.getX();
    if (miny > curPos.getY()) miny = curPos.getY();
  }//end for i
  Coord centre;
  centre.setX ((maxx + minx)/2.0);
  centre.setY ((maxy + miny)/2.0);
  centre.setZ (0);
  
  //if max_allow_shrink (0, 1), apply scaling.
  if ((max_allow_shrink > 0) && (max_allow_shrink < 1)) { 
    //apply the max allowed shrink to all the nodes of the graph radially
    //out from the centre of the graph
    cerr << "final layout section by " << max_allow_shrink << " about "
	 << centre << endl;
    
    //load all the final positions into the layout proxy
    for (nodeIt curNode = nodeList.begin(); curNode != nodeList.end();
	 ++curNode) {
      Coord curPos = nodeLayout->getNodeValue (*curNode);
      curPos -= centre;
      curPos *= max_allow_shrink;
      curPos += centre;
      nodeLayout->setNodeValue (*curNode, curPos);
    }//end for i
  }//end if
  return true;
}//end compactLevel

//Recursively calls CompactFinalLayout to compact all the subgraphs and
//the final graph.
bool CompactFinalLayout::recurseOnSubGraphs (SuperGraph *graph,
					     string &compactType) {
  //for error messages
  string errMsg;
  //success of recursive calls
  bool success = true;

  //reecurively shrink the subgraphs
  MetaGraphProxy *metagraphs = 
    graph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  SizesProxy *sizes =
    graph->getProperty<SizesProxy> ("viewSize");
  Iterator <node> *nodes = graph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    SuperGraph *curGraph = metagraphs->getNodeValue (curNode);

    //the size is either decided by the subgraph or, if there is no
    //subgraph, the node itself
    if (curGraph != NULL) {
      success = recurseOnSubGraphs (curGraph, compactType);
      if (!success) {
	string name = curGraph->getAttribute <string> ("name");
	cerr << "Error recursing on subgraph " << name << endl;
	return false;
      }//end if
      updateBoundary (*boundaryTable, curNode, curGraph, compactType);
    }//end if
    else (*boundaryTable)[curNode.id] = 
	   leafNodeBoundary (sizes->getNodeValue (curNode));
  } delete nodes; //end while

  //resolve overlaps at the level
  string name = graph->getAttribute <string> ("name");
  cerr << "compacting graph " << name << endl;
  bool level_success = compactLevel (graph, compactType);
  if (!level_success) {
    cerr << "Error compacting graph " << name << endl;
    return false;
  }//end if
  
  return true;
}//end recurseOnSubGraphs


bool CompactFinalLayout::run() {

  //proxy for positions
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  //load all the initial positions into the layout proxy
  Iterator <node> *bkpnodes = superGraph->getNodes();
  while (bkpnodes->hasNext()) {
    node curNode = bkpnodes->next();
    layoutProxy->setNodeValue (curNode, pos->getNodeValue (curNode));
  } delete bkpnodes; //end while

  //get the virtual node table
  bool madeBT = !dataSet->get <BoundaryTable *> (DATASET_BOUNDARY_TABLE,
						 boundaryTable);
  if (madeBT) {
    cerr << "Boundary table not specified.  Making one and deleting it at"
	 << "the end of plugin execution " << endl;
    boundaryTable = new BoundaryTable;
    dataSet->set <BoundaryTable *> (DATASET_BOUNDARY_TABLE,
				    boundaryTable);
  }//end if

  //get the shrink resolve table table
  bool madeRT = !dataSet->get <shrinkResTable *> (DATASET_SHRINK_RESOLVE_TABLE,
						  shrinkResolveTable);
  if (madeRT) {
    shrinkResolveTable = new shrinkResTable;
    dataSet->set <shrinkResTable *> (DATASET_SHRINK_RESOLVE_TABLE,
				     shrinkResolveTable);
  }//end if

  //get the type of compaction required
  string compactType;
  bool success = dataSet->get <string> (DATASET_BOUNDING_TYPE, 
					compactType);
  if (!success) {
    cerr << "Bounding type not specified.  Using bounding circle by default."
	 << endl;
    compactType = BOUNDING_CIRCLE;
  }//end if

  //get the list of graphs that require compaction of all pairs
  bool compactAllPairsSuccess = 
    dataSet->get <vector <string> *> (DATASET_ALL_PAIRS, compactAllPairs);
  if (!compactAllPairsSuccess) compactAllPairs = NULL;

   //Recurse on the the other sub graphs
  bool recursive_success = recurseOnSubGraphs (superGraph, compactType);

  //if we made the virtual node table (at root) delete it
  if (madeBT) delete boundaryTable;

  //if we made the shrink reslove table, delete it
  if (madeRT) delete shrinkResolveTable;

  //Compact the level and return the result
  return recursive_success;
}//end run

bool CompactFinalLayout::check(string &erreurMsg){
  erreurMsg="";
  return true;
}

void CompactFinalLayout::reset(){}
