/**
 Author: David Auber
 Email : auber@labri.fr
 Last modification : 20/08/2001
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of Compact Graph to improve information density.  For nodes, 
 * history, author and licence please see the header file CompactTree.h
 */

#include <math.h>
#include <vector>
#include "CompactTree.h"
#include <tulip/TreeTest.h>
#include <tulip/MutableContainer.h>
#include <tulip/SortIterator.h>

LAYOUTPLUGIN(CompactTree,"Compact Tree","David Auber/ Daniel Archambault","25/11/2004","Ok","0","1");

using namespace std;

//A local metric proxy for the theta coordinate of a node.
MetricProxy *thetas;

//mutable containers for nodes on the other ends of branches.
MutableContainer <double> rads;
MutableContainer <Coord> centres;

CompactTree::CompactTree(const PropertyContext &context):Layout(context){}

CompactTree::~CompactTree() {}

//The following function computes the maximum shrinking factor that
//can be applied to one node of a triple of nodes so that the triple
//does not overlap.  The 'root' Coord is the coordinates of the root.  The
//'leaf' Coord is the coordinates of the leaf.  The 'd' Coord is the
//position of the node being moved in.  The doubles are the radius of
//the root, leaf and d nodes respectively.
inline double CompactTree::compactD (Coord root, Coord leaf, Coord d,
				     double root_rad, double leaf_rad, 
				     double d_rad) {
  //compute a vector between root and d and a normalized vector
  Coord d_root, norm_d_root;
  d_root = d - root;
  norm_d_root = d_root;
  //if substantially greater than zero magnitude
  if (d_root.norm() > 1e-6)
    norm_d_root /= d_root.norm();

  //compute a vector between leaf and root
  Coord leaf_root, norm_leaf_root;
  leaf_root = leaf - root;
  norm_leaf_root = leaf_root;
  //is substantially greater than zero magnitude
  if (leaf_root.norm() > 1e-6)
    norm_leaf_root /= leaf_root.norm();

  //compute the cosine and sin of the angle between the two vectors
  double cos_val = norm_leaf_root.getX()*norm_d_root.getX() + 
    norm_leaf_root.getY()*norm_d_root.getY();

  //compute the contraction distance to the first point of contact
  //between node d and the leaf (the + root of the quadratic)
  double sum_rads = leaf_rad + d_rad;
  double discriminant = sum_rads*sum_rads + 
    leaf_root.norm()*leaf_root.norm()*(cos_val*cos_val - 1);

  //if there is a real root, we may need to return it
  if (discriminant > 0) {
    double positiveNum = leaf_root.norm()*cos_val + sqrt (discriminant);
    //if the positive root is greater than zero it is valid
    if (positiveNum > 0) return positiveNum/d_root.norm();
    
    //otherwise, it is invalid because we pass through the centre node
    //return scale to the centre node.
    else return (root_rad + d_rad)/d_root.norm();
  }//end if

  //otherwise, no colision with leaf, return max allow with root
  else 
    return (root_rad + d_rad)/d_root.norm();
}//end compactD

//This small function translates all the dependents of the passed
//node 'n' by 'trans'
void CompactTree::transDependents (node &n, Coord trans) {
  //translate the node
  layoutProxy->setNodeValue (n, layoutProxy->getNodeValue (n) + trans);
  centres.set (n.id, layoutProxy->getNodeValue (n));

  //if this has no out nodes, we are done
  if (superGraph->outdeg (n) == 0) return;

  //if it has an out degree, translate its other nodes too
  Iterator <node> *nodes = superGraph->getOutNodes (n);
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    transDependents (curNode, trans);
  } delete nodes; //end while
}//end transDependents

//computes the new centre and radius for a branch and set them.  
//The branch is rooted at "root" and has nodes "compactedNodes"
void CompactTree::setCentreAndRad (node root, 
				   vector <node> &alreadyCompacted) {

  //proxy for sizes
  SizesProxy *sizes = superGraph->getProperty <SizesProxy> ("viewSize");

  //for all the nodes, find the max
  //go through the layout and find the node of max/min +/- size
  //in x and y
  double maxx, minx, maxy, miny;

  //set the bounds of the node to ones that will be beaten.
  minx = miny = DBL_MAX;
  maxx = maxy = DBL_MIN;

  //go through the nodes and compute the node size
  for (unsigned int i = 0; i < alreadyCompacted.size(); i++) {
    node curNode = alreadyCompacted[i];
    
    //get this node size and position
    Size size = sizes->getNodeValue (curNode);
    Coord pos = layoutProxy->getNodeValue (curNode) - centres.get (root.id);
    
    //see if they extend the bounding box of the node
    if (maxx < pos.getX() + size.getW()/2.0)
      maxx = pos.getX() + size.getW()/2.0;
    if (maxy < pos.getY() + size.getH()/2.0)
      maxy = pos.getY() + size.getH()/2.0;
    if (minx > pos.getX() - size.getW()/2.0)
      minx = pos.getX() - size.getW()/2.0;
    if (miny > pos.getY() - size.getH()/2.0)
      miny = pos.getY() - size.getH()/2.0;
  }//end for

  Coord centre ((maxx + minx)/2.0, (maxy + miny)/2.0, 0);
  centres.set (root.id, centre + centres.get (root.id));
  rads.set (root.id, sqrt ((((maxx - minx)*(maxx - minx))/4.0) + 
			   ((maxy - miny)*(maxy - miny))/4.0)); 
}//end computeCentreAndRad

//This function compacts a specific branch areound a node, updating
//its radius and centre accordingly.
void CompactTree::compactBranch (node &root) {
  //if the out degree is zero, we have reached the base case.
  //return
  if (superGraph->outdeg (root) == 0) return;

  //compact all the branches that are subgraphs of this branch.
  Iterator <node> *nodes = superGraph->getOutNodes(root);
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    compactBranch (curNode);
  } delete nodes; //end while

  //a vector to store the list of nodes that have been compacted
  //already
  vector <node> alreadyCompacted;
  alreadyCompacted.push_back (root);

  //get the centre of the current root
  double rad = rads.get (root.id);
  Coord centre = centres.get (root.id);
  
  //around the root of this branch, compute the thetas
  nodes = superGraph->getOutNodes (root);
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord curPos = layoutProxy->getNodeValue (curNode) - centre;
    thetas->setNodeValue (curNode, ((curPos.getX()) < 0 ? M_PI : 0) + 
			  atan (curPos.getY()/curPos.getX()));
  } delete nodes; //end while

  //go through the children of root and compact them
  SortNodeIterator sortedNodes (superGraph->getOutNodes(root), thetas);
  while (sortedNodes.hasNext()) {
    node curNode = sortedNodes.next();

    //get the positions and sizes of the current node
    Coord curCentre; 
    curCentre = centres.get (curNode.id);
    double curRad = rads.get (curNode.id);
   
    //an array of scale factors for the already compacted nodes
    double scale_factors [alreadyCompacted.size()];
    for (unsigned int i = 0; i < alreadyCompacted.size(); i++)
      scale_factors [i] = -1;

    //for all the nodes that have already been compacted
    //the furthest in we can scale the node wthout overlaps
    for (unsigned int i = 0; i < alreadyCompacted.size(); i++) {
      //get the centre of the current compacted node
      double radCompact = rads.get (alreadyCompacted[i].id);
      Coord centreCompact = centres.get (alreadyCompacted[i].id);

      //if this already compacted node is outside my distance from the
      //centre node, do not count me
      if ((curCentre - centre).norm() < (centreCompact - centre).norm())
	scale_factors[i] = -1;

      //otherwise slide the current node as far in close to the compacted
      //nodes as possible, before overlap
      else
	scale_factors [i] = compactD (centre, centreCompact, curCentre,
				      rad, radCompact, curRad);
    }//end i

    //the final scale factor we would like to take
    double final_scale_factor = DBL_MAX;

    //check the overlap of the various scale factors, in decreasing
    //radial order.  Stop when scale factor increases or overlaps
    //are incurred
    for (int i = alreadyCompacted.size() - 1; i >= 0; i--) {
      //if the scale factor is greater than the current final_scale_factor
      //there is no point in checking it out.  It will not be the
      //optimum scale factor
      if (scale_factors[i] > final_scale_factor) continue;

      //compute the position of the node, if the scale factor
      //was applied
      Coord node_pos;
      node_pos = curCentre - centre;
      node_pos *= scale_factors[i];
      node_pos += centre;
      
      //check for overlaps in already compacted nodes
      bool exist_overlap = false;
      for (int j = 0; j < alreadyCompacted.size(); j++) {
	//get the centre of the current compacted node
	double radCompact = rads.get (alreadyCompacted[j].id);
	Coord centreCompact = centres.get (alreadyCompacted[j].id);
	
	//compute the distance between the nodes
	double dist = (centreCompact - node_pos).norm();
	
	//if there is a possibility for existing overlap, check
	//to see if there is.
	if (radCompact + curRad > dist + 1e-3) {
	  exist_overlap = true;
	  break;
	}//end if
      }//end for j
      
      //if there is no overlap, update the translation vector
      if (!exist_overlap)
	final_scale_factor = scale_factors[i];
    }//end for i

    //if the final_scale_factor is DBL_MAX, we cannot translate the node
    if (fabs (final_scale_factor - DBL_MAX) < 1e-6) continue;

    //compute the required translation vector for the
    //node from the computed scale factor
    Coord trans;
    trans = curCentre - centre;
    trans *= final_scale_factor;
    trans += centre - curCentre;

    //translate all the dependents
    transDependents (curNode, trans);

    //push back the newly compacted node
    alreadyCompacted.push_back (curNode);
  } //end while

  //compute the new centre and radius of this branch
  setCentreAndRad (root, alreadyCompacted);
}//end compactBranch  

//The following function fixes the tree so that the new root is
//the actual root
void CompactTree::fixTree (node cur_root, node came_from) {
  //go through all the edges at the current root
  Iterator <edge> *edges = superGraph->getInOutEdges (cur_root);
  while (edges->hasNext()) {
    edge curEdge = edges->next();

    //if the edge is not pointing out from the cur_root and the
    //edge is not the edge we used to get to the cur_root, we
    //have to flip its direction and recurse on the other
    //end
    if ((superGraph->source (curEdge) != came_from) &&
	(superGraph->target (curEdge) == cur_root)) {
      superGraph->reverse (curEdge);
      fixTree (superGraph->target (curEdge), cur_root);
    }//end if
  } delete edges; //end while
}//end fixTree

//The following function makes the node of highest degree the root
//of the tree
node CompactTree::makeHighestDegreeRoot () {
  //find the highest degree node in the tree
  Iterator <node> *nodes = superGraph->getNodes();
  unsigned int max_degree = 0;
  node new_root;
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    if (superGraph->deg (curNode) > max_degree) {
      new_root = curNode;
      max_degree = superGraph->deg (curNode);
    }//end if
  } delete nodes; //end while

  //now correct the tree so that the new root is actually the root
  fixTree (new_root, new_root);
  return new_root;
}//end makeHighestDegreeRoot
  
bool CompactTree::run() {
  //proxy for positions 
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  SizesProxy *sizes = superGraph->getProperty <SizesProxy> ("viewSize");

  //take out a local property for theta radian coord
  thetas = 
    superGraph->getLocalProperty <MetricProxy> ("Compact Thetas");

  //set the original positions of the node in layoutProxy and
  //initialize early node centres and radiuses
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord curPos = pos->getNodeValue (curNode);
    Size curSize = sizes->getNodeValue (curNode);
    layoutProxy->setNodeValue (curNode, curPos);
    centres.set (curNode.id, curPos);
    rads.set (curNode.id, sqrt (((curSize.getW()*curSize.getW())/4.0 + 
				 (curSize.getH()*curSize.getH())/4.0)));
  } delete nodes; //end while
  
  //We make the highest degree node in the tree the root
  node root = makeHighestDegreeRoot ();

  //compact the branches
  compactBranch (root);

  //delete the local property
  superGraph->delLocalProperty ("Compact Thetas");
  return true;
}//end run

bool CompactTree::check (std::string &errMsg) {
    if (!TreeTest::isTree (superGraph)) {
      errMsg = "The passed graph to Compact Tree must be a tree.\n";
      return false;
    }//end if
    return true;
}//end check

void CompactTree::reset () {
}//end reset
