#include <tulip/TlpTools.h>
#include <tulip/StableIterator.h>
#include <tulip/TulipPlugin.h>
#include <iostream>

using namespace std;


///
class ConnectedAndTreeComponent2:public Metric { 
private: 
  //If set to true, the tree root is the first node that lies on a cycle.
  //Otherwise, the tree root is the first node on the tree.
  bool nonTreeRoots;

public:
  ConnectedAndTreeComponent2(const PropertyContext &context):Metric(context){};

  //=======================================================================
  /**
   * The function computes a depth first search on graph starting from node
   * n.  At an intermediate step, it just travelled the edge from cameFrom
   * to n.  Nodes and edges of the trees in the graph are removed and placed
   * inside the superGraph treesGraph.  The nodes visited thus far in the
   * depth first search are contained in visited.
   */
  bool dfsErase (node n, node cameFrom, SuperGraph *graph, 
		 SuperGraph *treesGraph, MutableContainer<bool> &visited) {
    //if we have already visited the node, return
    if (visited.get(n.id)) return false;
    
    //visit all adjacent nodes to this node in the graph and determine
    //if they are part of a tree
    visited.set(n.id,true);
    StableIterator<edge> it (graph->getInOutEdges(n));
    while(it.hasNext()) {
      edge curEdge = it.next();
      node opposite = graph->opposite (curEdge, n);
      if (opposite == n) continue; //no self-loop visits
      if (cameFrom != opposite) { 
	if (dfsErase (opposite, n, graph, treesGraph, visited)) {
	  if (!treesGraph->isElement(n)) treesGraph->addNode(n);
	  treesGraph->addEdge (curEdge);
	  if (superGraph->target (curEdge) == n)
	    superGraph->reverse (curEdge);
	}//end if
      }//end if
    }//end while
    
    //if we are at a new leaf, delete it from graph and put it into
    //the treesGraph.
    if (graph->deg(n) < 2) {
      treesGraph->addNode(n);
      graph->delNode(n);
      return true;
    }//end if
    else if (!nonTreeRoots && treesGraph->isElement (n)) {
      StableIterator<edge> delit(treesGraph->getInOutEdges(n));
      while (delit.hasNext()) treesGraph->delEdge (delit.next());
      treesGraph->delNode(n);
    }//end else if
    return false;
  }//end dfsErase
  
  //=======================================================================
  /**
   * This function finds the first cycle in graph, if one exists, to start
   * the tree detecting depth first search from a non-tree.  The cycle
   * detection starts from node n and, at an intermediate step, it has just
   * transitioned to n from cameFrom.  If a cycle exists, a node on the cycle
   * is returned in cycleNode.  It returns true if a cycle was found.
   */
   bool findCycle(node n, node cameFrom, SuperGraph *graph, MutableContainer<bool> &visited, node &cycleNode) {
     //if we have visited this node before, we have a cycle
    if (visited.get(n.id)) {
      cycleNode = n;
      return true;
    }

    //otherwise, keep looking for one
    visited.set(n.id, true);
    StableIterator<node> it(graph->getInOutNodes(n));
    while(it.hasNext()) {
      node curNode = it.next();
      if (curNode != cameFrom)
	if (findCycle(curNode, n, graph, visited, cycleNode)) return true;
    }//end while
    //we have not found a cycle here
    return false;
   }//end findCycle

  //=======================================================================
  /**
   * This function is simply the entry point for the plugin.
   */
  bool run() {
    SuperGraph *graph=tlp::newCloneSubGraph(superGraph);
    string errMsg;    
    MutableContainer<bool> visited;
    MutableContainer<bool> cycleVisit;
    nonTreeRoots = false;
    if (dataSet != NULL) dataSet->get ("Non-Tree Roots", nonTreeRoots);

    //subgraph to hold the trees
    SuperGraph *treesGraph = superGraph->addSubGraph();

    //Find all the trees and connected components
    StableIterator<node> it(graph->getNodes());
    visited.setAll(false);
    while(it.hasNext()) { 
      node n = it.next();
      if (!visited.get(n.id)) {
	//find a cycle node if there is one
	node cycleNode;
	cycleVisit.setAll (false);
	bool foundCycle = findCycle(n, n, graph, cycleVisit, cycleNode);
	if (foundCycle) 
	  dfsErase (cycleNode, cycleNode, graph, treesGraph, visited);
	else 
	  dfsErase (n, n, graph, treesGraph, visited);
      }//end if visited
    }//end while stable node it

    MetricProxy connectedcomponent(graph);
    if (!graph->computeProperty("Connected Component", 
				&connectedcomponent, errMsg))
      cerr << "Error : " << errMsg << endl;
    
    MetricProxy connectedcomponent2(treesGraph);
    if (!treesGraph->computeProperty("Connected Component", 
				     &connectedcomponent2, errMsg))
      cerr << "Error : " << errMsg << endl;
    double max = 0;
    Iterator<node> *nodes = graph->getNodes();
    while(nodes->hasNext()){
      node n = nodes->next();
      max >?= connectedcomponent.getNodeValue(n);
      metricProxy->setNodeValue(n, connectedcomponent.getNodeValue(n));
    }delete nodes;
    max += 1.0;
    nodes = treesGraph->getNodes();
    while(nodes->hasNext()){
      node n = nodes->next();
      metricProxy->setNodeValue(n, connectedcomponent2.getNodeValue(n) + max);
    }delete nodes;
    superGraph->delAllSubGraphs(graph);
    superGraph->delAllSubGraphs(treesGraph);
    return true;
  }
  //=======================================================================
};

METRICPLUGIN(ConnectedAndTreeComponent2,"Connected_Tree Component 2", "Daniel Archambault","03/01/2005","Alpha","0","1");
