/**
 Author: Daniel Archambault
 Email : archam@cs.ubc.ca
 Last modification : 08/12/2004
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/
/**
 * implementation of Resolve Overlap.  For nodes, history, 
 * author and licence please see the header file FastOverlapRemoval.h
 */

#include <algorithm>
#include "FastOverlapRemoval.h"

LAYOUTPLUGIN(FastOverlapRemoval,"Fast Overlap Removal","Daniel Archambault","08/12/2004","Ok","0","1");

using namespace std;
using namespace __gnu_cxx;

//a function that returns the overlap term of x
inline double olap (double pu, double pv, double su, double sv) {
  return (su + sv)/2.0 - fabs (pu - pv);
}//end olap

//A comparator for events by y coord
typedef struct compY {
  bool operator() (const event &e1, const event &e2) {
    return e1.nodepos.getY() < e2.nodepos.getY();
  }//end operator()
  static bool scanLess (const event &e1, const event &e2) {
    return e1.nodepos.getY() < e2.nodepos.getY();
  }//end less
  static bool eventSort(const event &e1, const event &e2) {
    if (e1.coord.getY() < e2.coord.getY()) return true;
    if (e1.coord.getY() > e2.coord.getY()) return false;
    return !e1.isOpen;
  }//end operator eventSort
  inline static double getScan (const Coord c) {
    return c.getY();
  }
  inline static void setScan (Coord &c, double value) {
    c.setY (value);
  }
  inline static double getPerp (const Coord c) {
    return c.getX();
  }
  inline static double getSizeScan (const Size s) {
    return s.getH();
  }
  inline static double getSizePerp (const Size s) {
    return s.getW();
  }
}; 

//a comparator for events by x coord
typedef struct compX {
  bool operator() (const event &e1, const event &e2) {
    return e1.nodepos.getX() < e2.nodepos.getX();
  }//end operator()
  static bool scanLess (const event &e1, const event &e2) {
    return e1.nodepos.getX() < e2.nodepos.getX();
  }//end less
  static bool eventSort(const event &e1, const event &e2) {
    if (e1.coord.getX() > e2.coord.getX()) return true;
    if (e1.coord.getX() < e2.coord.getX()) return false;
    return !e1.isOpen;
  }//end operator eventSort
  inline static double getScan (const Coord c) {
    return c.getX();
  }
  inline static void setScan (Coord &c, double value) {
    c.setX (value);
  }
  inline static double getPerp (const Coord c) {
    return c.getY();
  }
  inline static double getSizeScan (const Size s) {
    return s.getW();
  }
  inline static double getSizePerp (const Size s) {
    return s.getH();
  }
} compX;

//gets the left hand neighbours of insertEvent
template <typename cmp>
void FastOverlapRemoval::getLeftNeigbours (set<event, cmp> &leftNeighbours, 
					   typename set<event, cmp>::iterator 
					   left, 
					   typename set<event, cmp>::iterator
					   begin,
					   event &insertEvent) {
  do { 
    --left;
    //cout << "left " << left->n.id << endl;
    double overScan = 
      olap (cmp::getScan (layout->getNodeValue (left->n)), 
	    cmp::getScan (layout->getNodeValue (insertEvent.n)),
	    cmp::getSizeScan (sizes->getNodeValue (left->n)),
	    cmp::getSizeScan (sizes->getNodeValue (insertEvent.n)));
    //cout << "overScan " <<  overScan << endl;
    //cout << "olap perp " 
    // << olap (cmp::getPerp (layout->getNodeValue (left->n)), 
    //	  cmp::getPerp (layout->getNodeValue (insertEvent.n)),
    //	  cmp::getSizePerp (sizes->getNodeValue (left->n)),
    //	  cmp::getSizePerp (sizes->getNodeValue (insertEvent.n))) 
       // << endl;
    //cout << "pass " << pass << endl;
    if (overScan <= 0) {
      leftNeighbours.insert (*left);
      return;
    }//end if
    if (pass == 2) {
      leftNeighbours.insert (*left);
      return;
    }//end else if
    else if ((overScan <= 
	      olap (cmp::getPerp (layout->getNodeValue (left->n)), 
		    cmp::getPerp (layout->getNodeValue (insertEvent.n)),
		    cmp::getSizePerp (sizes->getNodeValue (left->n)),
		    cmp::getSizePerp (sizes->getNodeValue (insertEvent.n)))))
      leftNeighbours.insert (*left);
  } while (left != begin);
}//end getLeftNeigbours

//gets the right hand neighbours of insertEvent
template <typename cmp>
void FastOverlapRemoval::getRightNeigbours (set<event, cmp> &rightNeighbours, 
					    typename set<event, cmp>::iterator 
					    right, 
					    typename set<event, cmp>::iterator
					    end,
					    event &insertEvent) {
  for (++right; right != end; ++right) {
//cout << "right " << right->n.id << endl;
    double overScan = 
      olap (cmp::getScan (layout->getNodeValue (right->n)), 
	    cmp::getScan (layout->getNodeValue (insertEvent.n)),
	    cmp::getSizeScan (sizes->getNodeValue (right->n)),
	    cmp::getSizeScan (sizes->getNodeValue (insertEvent.n)));

    
//  cout << "overScan " <<  overScan << endl;
//  cout << "olap perp " 
// << olap (cmp::getPerp (layout->getNodeValue (right->n)), 
//	  cmp::getPerp (layout->getNodeValue (insertEvent.n)),
//	  cmp::getSizePerp (sizes->getNodeValue (right->n)),
//	  cmp::getSizePerp (sizes->getNodeValue (insertEvent.n))) 
// << endl;

    if (overScan <= 0) {
      rightNeighbours.insert (*right);
      return;
    }//end if
    if (pass == 2) {
      rightNeighbours.insert (*right);
      return;
    }//end else if
    if (overScan <= 
	olap (cmp::getPerp (layout->getNodeValue (right->n)), 
	      cmp::getPerp (layout->getNodeValue (insertEvent.n)),
	      cmp::getSizePerp (sizes->getNodeValue (right->n)),
	      cmp::getSizePerp (sizes->getNodeValue (insertEvent.n))))
      rightNeighbours.insert (*right);
  }//end for right
}//end getRightNeigbours

//removes the passed iterator objects from neighbours.  The iterators
//must be contained inside of neighbours.
template<typename cmp>
void eraseNeighbours (set<event, cmp> &neighbours,
		      vector<typename set<event, cmp>::iterator> &del) {
  typename vector<typename set<event, cmp>::iterator>::iterator 
    delIt = del.begin();
  for (; delIt != del.end(); ++delIt) neighbours.erase (*delIt);
}//end eraseNeighbours

template<typename cmp>
void removeTransitivesForward (set<event, cmp> &scanNeighbours,
			       set<event, cmp> &removeNeighbours) {
  vector<typename set<event, cmp>::iterator> toEraseForward;
  typename set<event, cmp>::iterator scanRightIt = scanNeighbours.begin();
  typename set<event, cmp>::iterator removeIt = removeNeighbours.begin();
  while ((scanRightIt != scanNeighbours.end()) && 
	 (removeIt != removeNeighbours.end())) {
    if (scanRightIt->n.id == removeIt->n.id) {
      toEraseForward.push_back (removeIt);
      ++removeIt;
      ++scanRightIt;
    }//end if
    else if (cmp::scanLess (*scanRightIt, *removeIt)) ++scanRightIt;
    else ++removeIt;
  }//end while
  eraseNeighbours<cmp> (removeNeighbours, toEraseForward);
}//end removeTransitivesForward

template<typename cmp>
void removeTransitivesBackward (set<event, cmp> &scanNeighbours,
				set<event, cmp> &removeNeighbours) {
  vector<typename set<event, cmp>::iterator> toEraseBackward;
  typename set<event, cmp>::reverse_iterator scanLeftIt = 
    scanNeighbours.rbegin();
  typename set<event, cmp>::reverse_iterator removeIt = 
    removeNeighbours.rbegin();
  while ((scanLeftIt != scanNeighbours.rend()) && 
	 (removeIt != removeNeighbours.rend())) {
    if (scanLeftIt->n.id == removeIt->n.id) {
      toEraseBackward.push_back ((++removeIt).base());
      ++scanLeftIt;
    }//end if
    else if (cmp::scanLess (*scanLeftIt, *removeIt)) ++removeIt;
    else ++scanLeftIt;
  }//end while
  eraseNeighbours<cmp> (removeNeighbours, toEraseBackward);
}//end removeTransitivesBackward

template<typename cmp>
void removeTransitives (set<event, cmp> &leftNeighbours,
			set<event, cmp> &rightNeighbours,
			hash_map<unsigned int, set<event, cmp> > &left,
			hash_map<unsigned int, set<event, cmp> > &right) {
  typename set<event, cmp>::iterator leftIt = leftNeighbours.begin();
  for (; leftIt != leftNeighbours.end(); ++leftIt)
    removeTransitivesForward<cmp> (rightNeighbours, right[leftIt->n.id]);
  typename set<event, cmp>::iterator rightIt = rightNeighbours.begin();
  for (; rightIt != rightNeighbours.end(); ++rightIt)
    removeTransitivesBackward<cmp> (leftNeighbours, left[rightIt->n.id]);
}//end removeTransitives

template<typename cmp>
bool isPresentForward (typename set<event, cmp>::iterator scanRightIt,
		       typename set<event, cmp>::iterator end,
		       set<event, cmp> &neighbours) {
  ++scanRightIt;
  typename set<event, cmp>::iterator nnIt = neighbours.begin();
  while ((scanRightIt != end) && (nnIt != neighbours.end())) {
    if (scanRightIt->n.id == nnIt->n.id) return true;
    else if (cmp::scanLess (*scanRightIt, *nnIt)) ++scanRightIt;
    else ++nnIt;
  }//end while
  return false;
}//end isPresentForward

template<typename cmp>
bool isPresentBackward (typename set<event, cmp>::reverse_iterator scanLeftIt,
			typename set<event, cmp>::reverse_iterator rend,
			set<event, cmp> &neighbours) {
  ++scanLeftIt;
  typename set<event, cmp>::reverse_iterator nnIt = neighbours.rbegin();
  while ((scanLeftIt != rend) && (nnIt != neighbours.rend())) {
    if (scanLeftIt->n.id == nnIt->n.id) return true;
    else if (cmp::scanLess (*scanLeftIt, *nnIt)) ++nnIt;
    else ++scanLeftIt;
  }//end while
  return false;
}//end isPresentBackward

//removes the passed event e is found
template<typename cmp>
inline void removeEvent (const event &e, set<event, cmp> &es) {
  typename set<event, cmp>::iterator ites = es.begin();
  for (; ites != es.end(); ++ites) {
    if (ites->n.id == e.n.id) {
      es.erase (ites);
      return;
    }//end if
  }//end for
}//end removeEvent

//peforms a sweep of yevents, resolving overlaps
template <typename cmp>
void FastOverlapRemoval::sweep (vector<event> &events) {
  set<event, cmp> scanline;
  pass++;
  unsigned int counter = 0;
  hash_map<unsigned int, set<event, cmp> > left;
  hash_map<unsigned int, set<event, cmp> > right;
  vector<event>::iterator eventIt = events.begin();
  for (; eventIt != events.end(); ++eventIt) {
    event curEvent = *eventIt;
    if (curEvent.isOpen) {
      //cout << "processing event of " << curEvent.n.id << endl;
      //insert the current event into the scanline
      pair <typename set<event, cmp>::iterator, bool> inserted = 
	scanline.insert (curEvent);
      if (!inserted.second) {
	cerr << "Key colision!  We could be having problems in a moment" 
	     << endl;
      }//end if

      /*typename set<event, cmp>::iterator eventsOpen = scanline.begin();
      cout << "events currently open " << endl;
      for (; eventsOpen != scanline.end(); ++eventsOpen)
	cout << eventsOpen->n.id << " " << flush;
      cout << endl;
      */
      //if there is something to overlap with...
      if (scanline.size() > 1) {
	//get references to the spot for the left and right neighbours
	set<event, cmp> &leftNeighbours = left[curEvent.n.id];
	set<event, cmp> &rightNeighbours = right[curEvent.n.id];

	//get left and right neigbours if they exist (more than one node)
	if (inserted.first != scanline.begin())
	  getLeftNeigbours<cmp> (leftNeighbours, inserted.first, 
				 scanline.begin(), curEvent);
	if (inserted.first != scanline.end())
	  getRightNeigbours<cmp> (rightNeighbours, inserted.first, 
				  scanline.end(), curEvent);
	/*cout << "before remove trans " << endl;
	for (unsigned int i = 0; i < nodeOrder.size(); i++) {
	  cout << "left neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator leftPrint = left[nodeOrder[i].id].begin();
	  for (; leftPrint != left[nodeOrder[i].id].end(); ++leftPrint)
	    cout << leftPrint->n.id << endl;
	  cout << "right neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator rightPrint = right[nodeOrder[i].id].begin();
	  for (; rightPrint != right[nodeOrder[i].id].end(); ++rightPrint)
	    cout << rightPrint->n.id << endl;
	}//end for i
	cout << endl;
	cout << leftNeighbours.size() << " " << rightNeighbours.size() << endl;
	cout << endl;*/
	removeTransitives<cmp> (leftNeighbours, rightNeighbours, left, right);
	
	/*cout << "after remove trans " << endl;
	for (unsigned int i = 0; i < nodeOrder.size(); i++) {
	  cout << "left neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator leftPrint = left[nodeOrder[i].id].begin();
	  for (; leftPrint != left[nodeOrder[i].id].end(); ++leftPrint)
	    cout << leftPrint->n.id << endl;
	  cout << "right neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator rightPrint = right[nodeOrder[i].id].begin();
	  for (; rightPrint != right[nodeOrder[i].id].end(); ++rightPrint)
	    cout << rightPrint->n.id << endl;
	}//end for i
	cout<< endl;
	*/
		
	//add the current event to the left and right node iterators
	typename set<event, cmp>::iterator leftIt = leftNeighbours.begin();
	vector<typename set<event, cmp>::iterator> toEraseFromLeft;
	for (; leftIt != leftNeighbours.end(); ++leftIt) {
	  if (!isPresentForward<cmp> (leftIt, leftNeighbours.end(),
				      right[leftIt->n.id]))
	    right[leftIt->n.id].insert (curEvent);
	  else toEraseFromLeft.push_back (leftIt);
	}//end for
	eraseNeighbours<cmp> (leftNeighbours, toEraseFromLeft);

	typename set<event, cmp>::reverse_iterator rrightIt = 
	  rightNeighbours.rbegin();
	vector<typename set<event, cmp>::iterator> toEraseFromRight;
	for (; rrightIt != rightNeighbours.rend(); ++rrightIt) {
	  if (!isPresentBackward<cmp> (rrightIt, rightNeighbours.rend(),
				       left[rrightIt->n.id])) 
	    left[rrightIt->n.id].insert (curEvent);
	  else {
	    typename set<event, cmp>::reverse_iterator bkp = rrightIt; 
	    toEraseFromRight.push_back ((++bkp).base());
	  }//end else
	}//end for
	eraseNeighbours<cmp> (rightNeighbours, toEraseFromRight);

	/*cout << "after opposite inserts " << endl;
	for (unsigned int i = 0; i < nodeOrder.size(); i++) {
	  cout << "left neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator leftPrint = left[nodeOrder[i].id].begin();
	  for (; leftPrint != left[nodeOrder[i].id].end(); ++leftPrint)
	    cout << leftPrint->n.id << endl;
	  cout << "right neighbours of " << nodeOrder[i].id << endl;
	  typename set<event, cmp>::iterator rightPrint = right[nodeOrder[i].id].begin();
	  for (; rightPrint != right[nodeOrder[i].id].end(); ++rightPrint)
	    cout << rightPrint->n.id << endl;
	}//end for i
	cout << endl;*/
      }//end if
    }//end if
    
    //otherwise, we have a close event and need to remove a node from the
    //scanline.
    else {
      vpscClause clause;
      set<event, cmp> &leftSet = left[curEvent.n.id];
      typename set<event, cmp>::iterator leftIt = leftSet.begin();
      for (; leftIt != leftSet.end(); ++leftIt) {
	clause.nleft = index[leftIt->n.id];
	clause.nright = index[curEvent.n.id];
	clause.left = cmp::getScan (layoutProxy->getNodeValue (leftIt->n));
	clause.right = cmp::getScan (layoutProxy->getNodeValue (curEvent.n));
	clause.gap = (cmp::getSizeScan (sizes->getNodeValue (leftIt->n)) + 
		      cmp::getSizeScan (sizes->getNodeValue (curEvent.n))) 
	  / 2.0;
	clause.index = vpsc.size();
	vpscOut[clause.nleft].push_back (vpsc.size());
	vpscIn[clause.nright].push_back (vpsc.size());
	vpsc.push_back (clause);
	removeEvent (curEvent, right[leftIt->n.id]);
      }//end for
      set<event, cmp> &rightSet = right[curEvent.n.id];
      typename set<event, cmp>::iterator rightIt = rightSet.begin();
      for (; rightIt != rightSet.end(); ++rightIt) {
	clause.nleft = index[curEvent.n.id];
	clause.nright = index[rightIt->n.id];
	clause.right = cmp::getScan (layoutProxy->getNodeValue (rightIt->n));
	clause.left = cmp::getScan (layoutProxy->getNodeValue (curEvent.n));
	clause.gap = (cmp::getSizeScan (sizes->getNodeValue (rightIt->n)) +
		      cmp::getSizeScan (sizes->getNodeValue (curEvent.n)))
	  / 2.0;
	clause.index = vpsc.size();
	vpscOut[clause.nleft].push_back (vpsc.size());
	vpscIn[clause.nright].push_back (vpsc.size());
	vpsc.push_back (clause);
	removeEvent (curEvent, left[rightIt->n.id]);
      }//end for
      scanline.erase (open[curEvent.n.id]);
      open.erase (curEvent.n.id);
      left.erase (curEvent.n.id);
      right.erase (curEvent.n.id);
    }//end else
  }//end for eventIt
}//end sweep

void FastOverlapRemoval::mergeBlock (block *p, vpscClause *clause,
				     block *b, double dist) {
  p->wposn = p->wposn + b->wposn - dist*b->weight;
  p->weight = p->weight + b->weight;
  p->posn = p->wposn/p->weight;
  p->active.push_back (clause->index);

  //merge active clause lists 
  vector<unsigned int>::iterator bactive = b->active.begin();
  for (; bactive != b->active.end(); ++bactive)  
    p->active.push_back (*bactive);

  //merge variable lists
  set<unsigned int>::iterator bvarit = b->vars.begin();
  for (; bvarit != b->vars.end(); ++bvarit) {
    blocks[*bvarit] = p;
    offset[*bvarit] = dist + offset[*bvarit];
    //cout << "new offset of " << *bvarit << " is " << offset[*bvarit] << endl;
    p->vars.insert (*bvarit);
  }//end for

  //merge in and out priority queues
  //cout << "number of vars " << p->vars.size() << endl;
  while (!b->in.empty()) {
    if ((p->vars.find(b->in.top()->nleft) == p->vars.end()) ||
	(p->vars.find(b->in.top()->nright) == p->vars.end()))
      p->in.push (b->in.top());
    b->in.pop();
  }//end while
  while (!b->out.empty()) {
    if ((p->vars.find(b->out.top()->nleft) == p->vars.end()) ||
	(p->vars.find(b->out.top()->nright) == p->vars.end()))
      p->out.push (b->out.top());
    b->out.pop();
  }//end while

  //DEBUG
  /*priority_queue<vpscClausePtr, vector<vpscClausePtr>, violationCmp> in;
  while (!p->in.empty()) {
    cout << "in queue " << p->in.top()->nleft << " " 
	 << p->in.top()->nright << " " << violation (p->in.top()) << endl;
    in.push (p->in.top());
    p->in.pop();
  }
  p->in = in;
  */
  delete b;
}//end mergeBlock

template <typename cmp>
void FastOverlapRemoval::mergeLeft (block *b) {
  while (!b->in.empty() && (violation (blocks[b->in.top()->nleft]->posn +
				       offset[b->in.top()->nleft],
				       blocks[b->in.top()->nright]->posn +
				       offset[b->in.top()->nright],
				       b->in.top()->gap,
				       b->in.top()->nleft, 
				       b->in.top()->nright) > 0)) {
    //cout << "IN HERE" << endl;
    vpscClause *clause = b->in.top();
    //cout << "clause " << b->in.top()->nleft << " " 
    //	 << b->in.top()->nright << endl;
    b->in.pop();
    block *bl = blocks[clause->nleft];
    if (b != bl) { //if not internal constraint
      double dist = offset[clause->nleft] + clause->gap 
	- offset[clause->nright];
      if (b->vars.size() > bl->vars.size()) {
	//cout << "merging " << clause->nleft << " onto " 
	//  << clause->nright << endl;
	mergeBlock (b, clause, bl, -dist);
      }//end if
      else {
	//cout << "merging " << clause->nright << " onto " 
	//   << clause->nleft << endl;
	mergeBlock (bl, clause, b, dist);
	b = bl;
      }//end else
      /*set<unsigned int>::iterator i = b->vars.begin();
      for (; i != b->vars.end(); ++i) {
	Coord c = layoutProxy->getNodeValue (nodeOrder[*i]);
	cmp::setScan (c, b->posn + offset[*i]);
	layoutProxy->setNodeValue (nodeOrder[*i], c); 
      }//end for i 
      pluginProgress->progress(1, 100);
      char buff [10];
      cin >> buff;*/
    }//end if
  }//end while
}//end mergeLeft

void FastOverlapRemoval::mergeRight (block *b) {
  while (!b->out.empty() && (violation (b->out.top()) 
			     + offset[b->out.top()->nleft]
			     - offset[b->out.top()->nright] > 0)) {
    vpscClause *clause = b->out.top();
    b->out.pop();
    block *br = blocks[clause->nright];
    if (b != br) { //if not internal constraint
      double dist = offset[clause->nleft] + clause->gap 
	- offset[clause->nright];
      if (b->vars.size() > br->vars.size())
	mergeBlock (b, clause, br, dist);
      else {
	mergeBlock (br, clause, b, -dist);
	b = br;
      }//end else
    }//end if
  }//end while
}//end mergeLeft

void FastOverlapRemoval::dsearch (unsigned int visit,
				  unsigned int &totalOrderPos,
				  unsigned int *totalOrder,
				  bool *marked) {
  if (marked[visit]) return;
  marked[visit] = true;
  vector<unsigned int>::iterator vpscOutIt = vpscOut[visit].begin();
  for (; vpscOutIt != vpscOut[visit].end(); ++vpscOutIt) {
    unsigned int nextNode = vpsc[*vpscOutIt].nright;
    if (!marked[nextNode]) 
      dsearch (nextNode, totalOrderPos, totalOrder, marked);
  }//end for
  totalOrder[totalOrderPos--] = visit;
}//end dsearch

void 
FastOverlapRemoval::computeTotalOrder (unsigned int *totalOrder) {
  //starting from all nodes with no in edges
  unsigned int totalOrderPos = nodeOrder.size() - 1;
  bool marked [nodeOrder.size()];
  for (unsigned int i = 0; i < nodeOrder.size(); i++) marked[i] = false;
  for (unsigned int i = 0; i < nodeOrder.size(); i++)
    if (!marked[i]) dsearch (i, totalOrderPos, totalOrder, marked);
}//end computeTotalOrder

template <typename cmp>
void FastOverlapRemoval::satisfyVPSC() {
  unsigned int totalOrder [nodeOrder.size()];
  computeTotalOrder (totalOrder);
  for (unsigned int i = 0; i < nodeOrder.size(); i++) {
    blocks[i] = NULL;
    offset[i] = 0;
  }//end for

  for (unsigned int i = 0; i < nodeOrder.size(); i++) {
    //cout << "processing " << totalOrder[i] << endl;
    blocks[totalOrder[i]] = 
      new block (totalOrder[i], 
		 cmp::getScan 
		 (layoutProxy->getNodeValue (nodeOrder[totalOrder[i]])),
		 1.0,
		 vpscIn, vpscOut, vpsc);
    mergeLeft<cmp> (blocks[totalOrder[i]]);
  }//end for
}//end satisfyVPSC


void
FastOverlapRemoval::splitVars (block *b, unsigned int start,
			       unsigned int wall, block *newb) {
  newb->vars.insert (start);
  blocks[start] = newb;

  vector<unsigned int>::iterator activeit = b->active.begin();
  for (; activeit != b->active.end(); ++activeit) {
    if ((vpsc[*activeit].nleft == start) && 
	(vpsc[*activeit].nright != wall)) {
      splitBlock (b, vpsc[*activeit].nright, start, newb);
      newb->active.push_back (*activeit);
    }//end if
    else if ((vpsc[*activeit].nright == start) && 
	     (vpsc[*activeit].nleft != wall)) {
      splitBlock (b, vpsc[*activeit].nleft, start, newb);
      newb->active.push_back (*activeit);
    }//end else if
  }//end for
}//end splitVars 

void
FastOverlapRemoval::splitBlock (block *b, unsigned int start,
				unsigned int wall, block *newb) {
  splitVars (b, start, wall, newb);
  newb->posn = b->posn;
  newb->weight = newb->vars.size();
  newb->wposn = newb->posn*newb->weight;

  vector<vpscClausePtr> vSaveIn;
  while (!b->in.empty()) {
    if (newb->vars.find (b->in.top()->nright) != newb->vars.end())
      newb->in.push (b->in.top());
    else vSaveIn.push_back (b->in.top());
    b->in.pop();
  }//end while
  vector<vpscClausePtr>::iterator vSaveInIt = vSaveIn.begin();
  for (; vSaveInIt != vSaveIn.end(); ++vSaveInIt)
    b->in.push (*vSaveInIt);
  
  vector<vpscClausePtr> vSaveOut;
  while (!b->out.empty()) {
    if (newb->vars.find (b->out.top()->nleft) != newb->vars.end())
      newb->out.push (b->out.top());
    else vSaveOut.push_back (b->out.top());
    b->out.pop();
  }//end while
  vector<vpscClausePtr>::iterator vSaveOutIt = vSaveOut.begin();
  for (; vSaveOutIt != vSaveOut.end(); ++vSaveOutIt) 
    b->out.push (*vSaveOutIt);
}//end splitBlock

template <typename cmp>
double FastOverlapRemoval::dsearchlm (unsigned int visit,
				    vector<unsigned int> &active,
				    unsigned int visited) {
  double dFdvisit = blocks[visit]->posn + offset[visit] - 
    cmp::getScan (layoutProxy->getNodeValue (nodeOrder[visit]));
  vector<unsigned int>::iterator activeit = active.begin();
  for (; activeit != active.end(); ++activeit) {
    if ((vpsc[*activeit].nleft == visit) && 
	(vpsc[*activeit].nright != visited)) {
      lm[*activeit] = dsearchlm<cmp> (vpsc[*activeit].nright, active, visit);
      dFdvisit += lm[*activeit];
    }//end if
    else if ((vpsc[*activeit].nright == visit) && 
	     (vpsc[*activeit].nleft != visited)) {
      lm[*activeit] = -dsearchlm<cmp> (vpsc[*activeit].nleft, active, visit);
      dFdvisit -= lm[*activeit];
    }//end if
  }//end for
  return dFdvisit;
}//end dsearch

template <typename cmp>
void FastOverlapRemoval::computeLM() {
  for (unsigned int i = 0; i < lm.size(); i++) lm[i] = 0;
  for (unsigned int i = 0; i < nodeOrder.size(); i++) 
    blocks[i]->marked = false;
  for (unsigned int i = 0; i < nodeOrder.size(); i++) {
    if (!blocks[i]->marked) {
      dsearchlm<cmp> (*(blocks[i]->vars.begin()), blocks[i]->active,
		      *(blocks[i]->vars.begin()));
      blocks[i]->marked = true;
    }//end if
  }//end for
}//end computeLM

template <typename cmp>
void FastOverlapRemoval::solveVPSC() {
  /* computeLM<cmp>();
  double minLM;
  //keep looping while there is a lagrange multiplier below zero
  //this indicates non-optimal displacement of nodes.
  do {
    minLM = 1;
    unsigned int minpos = 0;
    vector<double>::iterator lmit = lm.begin();
    unsigned int counter = 0;
    for (; lmit != lm.end(); ++lmit) {
      if (*lmit < minLM) {
	minLM = *lmit;
	minpos = counter;
      }//end if
      counter++;
    }//end for
    
    cerr << "Min Lagrange " << minLM << endl;
    
    //we need to split if the lagrange multiplier is less than zero
    if (minLM < 0) {
      //split the block along the violating constraint
      block *b = blocks[vpsc[minpos].nleft];
      block *lb = new block;
      splitBlock (b, vpsc[minpos].nleft, vpsc[minpos].nright, lb);
      block *rb = new block;
      splitBlock (b, vpsc[minpos].nright, vpsc[minpos].nleft, rb);
      delete b;
      
      lb->posn = b->posn + minLM;
      lb->wposn = lb->posn*lb->weight;
      mergeLeft (lb);
      
      rb = blocks[vpsc[minpos].nright];
      rb->wposn = 0;
      set<unsigned int>::iterator itvars = rb->vars.begin();
      for (; itvars != rb->vars.end(); ++itvars) 
	rb->wposn +=  
	  cmp::getScan (layoutProxy->getNodeValue (nodeOrder[*itvars])) - 
	  offset [*itvars];
      double oldPosn = rb->posn;
      rb->posn = rb->wposn/rb->weight;
      mergeRight (rb);
      computeLM<cmp>();
      }//end if
      } while (minLM < 0);*/
}//end solveVPSC

template<typename cmp>
void FastOverlapRemoval::setFinalPositions () {
  Iterator<node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord curPos = layoutProxy->getNodeValue (curNode);
    cmp::setScan (curPos,
		  blocks[index[curNode.id]]->posn + offset[index[curNode.id]]);
    layoutProxy->setNodeValue (curNode, curPos);
  } delete nodes;
}//end setFinalPositions

void FastOverlapRemoval::cleanBlocks () {
  for (unsigned int i = 0; i < nodeOrder.size(); i++) {
    if (blocks[i] != NULL) {
      block *bkp = blocks[i];
      set<unsigned int>::iterator varit = bkp->vars.begin();
      for (; varit != bkp->vars.end(); ++varit) blocks[*varit] = NULL;
      delete bkp;
    }//end if
  }//end for
}//end cleanBlocks

bool lessInX (const nodePlacePair &p1, const nodePlacePair &p2) {
  return p1.second.getX() < p2.second.getX();
}//end lessInX

bool lessInY (const nodePlacePair &p1, const nodePlacePair &p2) {
  return p1.second.getY() < p2.second.getY();
}//end lessInY

#define EPSILON (1e-3)
template <typename cmp>
void 
FastOverlapRemoval::perturbEqual (bool 
				  (*sortFn) (const nodePlacePair &p1, 
					     const nodePlacePair &p2)) {
  vector<nodePlacePair> pairs;
  Iterator<node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    nodePlacePair p;
    p.first = curNode;
    p.second = layout->getNodeValue (curNode);
    pairs.push_back (p);
  } delete nodes; //end while
  sort (pairs.begin(), pairs.end(), sortFn);
  vector<nodePlacePair>::iterator evit = pairs.begin();
  vector<nodePlacePair>::iterator evitnext = evit;
  for (++evitnext; evitnext != pairs.end(); ++evitnext, ++evit) {
    double newPos = cmp::getScan(evitnext->second);
    Coord previous = layout->getNodeValue (evit->first);
    if (cmp::getScan (evitnext->second) - cmp::getScan (previous) < 0)
      newPos = cmp::getScan (previous) + EPSILON;
    else if (cmp::getScan (evitnext->second) - 
	     cmp::getScan (previous) < EPSILON)
      newPos += EPSILON;
    Coord c = layout->getNodeValue (evitnext->first);
    cmp::setScan (c, newPos);
    layout->setNodeValue (evitnext->first, c);
  }//end for
}//end perturbEqual

bool FastOverlapRemoval::run() {
  vector<event> events;
  unsigned int counter = 0;
  pass = 0;
  vpscIn.resize (superGraph->numberOfNodes());
  vpscOut.resize (superGraph->numberOfNodes());
  //adjust node size for rotated nodes
  layout = NULL;
  if (dataSet != NULL) dataSet->get ("layout", layout);
  if (layout == NULL)
    layout = superGraph->getProperty<LayoutProxy> ("viewLayout");
  SizesProxy *viewSize = superGraph->getProperty<SizesProxy> ("viewSize");
  MetricProxy *viewRot = superGraph->getProperty<MetricProxy> ("viewRotation");
  SizesProxy localSizes (superGraph);
  Iterator<node> *snodes = superGraph->getNodes();
  while (snodes->hasNext()) {
    node curNode = snodes->next();
    Size curSize = viewSize->getNodeValue (curNode);
    double curRot = viewRot->getNodeValue (curNode);
    Size newSize (curSize.getW()*fabs (cos (curRot*M_PI/180.0)) +
		  curSize.getH()*fabs (sin (curRot*M_PI/180.0)),
		  curSize.getW()*fabs (sin (curRot*M_PI/180.0)) +
		  curSize.getH()*fabs (cos (curRot*M_PI/180.0)));
    localSizes.setNodeValue (curNode, newSize);
  }delete snodes; //end while
  sizes = &localSizes; 

  //perturb nodes with same x and y coord
  perturbEqual<compY> (lessInY);
  perturbEqual<compX> (lessInX);

  //create events
  Iterator<node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    event eo, ec;
    node curNode = nodes->next();
    index[curNode.id] = nodeOrder.size();
    eo.isOpen = true;
    ec.isOpen = false;
    eo.n = ec.n = curNode;
    Coord pos = layout->getNodeValue (curNode);
    layoutProxy->setNodeValue (curNode, pos);
    Size size = sizes->getNodeValue (curNode); 
    eo.coord = Coord (pos.getX() + size.getW()/2.0,
		      pos.getY() - size.getH()/2.0, 0);
    ec.coord = Coord (pos.getX() - size.getW()/2.0,
		      pos.getY() + size.getH()/2.0, 0);
    eo.nodepos = ec.nodepos = pos;
    open[curNode.id] = eo;
    events.push_back (eo);
    events.push_back (ec);
    nodeOrder.push_back (curNode);
  } delete nodes;

  //resolve x overlaps
  sort (events.begin(), events.end(), compY::eventSort);
  sweep<compX> (events);
  if (!vpsc.empty()) {
    blocks.resize (superGraph->numberOfNodes());
    offset.resize (superGraph->numberOfNodes());
    lm.resize (vpsc.size());
    satisfyVPSC<compX> ();
    //solveVPSC<compX> ();
    setFinalPositions<compX> ();
    cleanBlocks ();
  }//end if
  
  //clean up the vpsc clauses
  vpsc.clear();
  vector <vector<unsigned int> >::iterator vpscinit = vpscIn.begin();
  for (; vpscinit != vpscIn.end(); ++vpscinit) vpscinit->clear();
  vector <vector<unsigned int> >::iterator vpscoutit = vpscOut.begin();
  for (; vpscoutit != vpscOut.end(); ++vpscoutit) vpscoutit->clear();

  //update positions of nodes in the events and fill open again
  vector<event>::iterator eventit = events.begin(); 
  for (; eventit != events.end(); ++eventit) {
    node curNode = eventit->n;
    Coord pos = layoutProxy->getNodeValue (curNode);
    Size size = sizes->getNodeValue (curNode);
    eventit->nodepos = pos;
    if (eventit->isOpen) {
      eventit->coord = Coord (pos.getX() + size.getW()/2.0,
			      pos.getY() - size.getH()/2.0, 0);
      open[curNode.id] = *eventit;
    }//end if
    else
      eventit->coord = Coord (pos.getX() - size.getW()/2.0,
			      pos.getY() + size.getH()/2.0, 0);
  }//end for

  //resolve y overlaps
  sort (events.begin(), events.end(), compX::eventSort);
  sweep<compY> (events);
  if (!vpsc.empty()) {
    blocks.resize (superGraph->numberOfNodes());
    offset.resize (superGraph->numberOfNodes());
    lm.resize (vpsc.size());
    satisfyVPSC<compY> ();
    //solveVPSC<compY> ();
    setFinalPositions<compY> ();
    cleanBlocks ();
  }//end if 
  return true;
}//end run
