//-*-c++-*-
#ifndef _RESOLVE_OVERLAP_H
#define _RESOLVE_OVERLAP_H

#if (__GNUC__ < 3)
#include <hash_map>
#else
#include <ext/hash_map>
#endif
#include <queue>
#include <set>
#include <tulip/TulipPlugin.h>
#include <tulip/MutableContainer.h>

/** \addtogroup layout */
/*@{*/
/// ReslolveOverlap.h - Resolves all node overlaps.
/** 
 * The algorithm finds and applies the minimum scale factor to ensure
 * that no two pairs of nodes overlap in the graph
 *
 * HISTORY:
 *
 * 25/11/2004 Verson 0.0.1: Initial release
 *
 * NOTES:
 *
 *   The algorithm works on general graphs.
 *   Let n be the number of nodes, the algorithm complexity is in O(n).
 *
 *   If two nodes start out at the same postion the algorithm cannot
 *   succeed.  It will return failure.
 *   
 * AUTHOR:
 *
 * Daniel Archambault, The University of British Columbia 
 * Email:archam@cs.ubc.ca
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */

//This structure is used for storing the events in the event queue
typedef struct event {
  bool isOpen;
  node n;
  Coord coord;
  Coord nodepos;
} event;

//a type for perturbing nodes
typedef std::pair<node, Coord> nodePlacePair;

//a structure to store the nodes involved in the VPSC clause
typedef struct vpscClause {
  unsigned int index;
  unsigned int nleft, nright;
  double left, gap, right;
} vpscClause; 
typedef vpscClause *vpscClausePtr;

//computes violation of a constraint
inline double static violation (const vpscClausePtr &v1) {
  return v1->left + v1->gap - v1->right;
}//end violation

//computes violation of a constraint
inline double static violation (double left, double right, double gap,
				unsigned int nleft, unsigned int nright) {
  //std::cout << "nleft nright " << nleft << " " << nright << std::endl;
  //std::cout << "left right " << left << " " << right << std::endl;
  //std::cout << "violation " << left + gap - right << std::endl;
  return left + gap - right;
}//end violation

//a comparator for violation
typedef struct violationCmp {
  inline bool operator() (const vpscClausePtr &v1, const vpscClausePtr &v2) {
    return  violation (v1) < violation (v2);
  }//end operator()
} violationCmp;

class FastOverlapRemoval:public Layout { 
class block;
private:
  LayoutProxy *layout;
  SizesProxy *sizes;
  __gnu_cxx::hash_map<unsigned int, event> open;
  std::vector<vpscClause> vpsc;
  std::vector<std::vector<unsigned int> > vpscIn;
  std::vector<std::vector<unsigned int> > vpscOut;
  std::vector<block *> blocks;
  std::vector<double> offset;
  std::vector<double> lm;
  std::vector<node> nodeOrder;
  __gnu_cxx::hash_map<unsigned int, unsigned int> index;
  unsigned int pass;

  //a block of variables
  class block {
  public:
    bool marked;
    std::set<unsigned int> vars;
    std::vector<unsigned int> active;
    std::priority_queue<vpscClausePtr, std::vector<vpscClausePtr>,
			violationCmp> in;
    std::priority_queue<vpscClausePtr, std::vector<vpscClausePtr>,
			violationCmp> out;
    double posn;
    double wposn;
    double weight;
    
    inline block (unsigned int variable, double nposn, double nweight,
		  std::vector<std::vector<unsigned int> > &vpscIn,
		  std::vector<std::vector<unsigned int> > &vpscOut,
		  std::vector<vpscClause> &vpsc) {
      vars.insert (variable);
      posn = nposn;
      //std::cout << "at position " << posn << std::endl;
      wposn = posn;
      weight = nweight;
      std::vector<unsigned int> &vpscin = vpscIn[variable]; 
      std::vector<unsigned int>::iterator itvpscin = vpscin.begin();
      for (; itvpscin != vpscin.end(); ++itvpscin) 
	in.push (&vpsc[*itvpscin]);
      
      std::vector<unsigned int> &vpscout = vpscOut[variable];
      std::vector<unsigned int>::iterator itvpscout = vpscout.begin();
      for (; itvpscout != vpscout.end(); ++itvpscout) 
	out.push (&vpsc[*itvpscout]);
      
    }//end initBlock
    inline ~block () {}
    inline block () {}
  };

  template <typename cmp>
  void getLeftNeigbours (std::set<event, cmp> &leftNeighbours,
			 typename std::set<event, cmp>::iterator left, 
			 typename std::set<event, cmp>::iterator begin,
			 event &insertEvent);
  template <typename cmp>
  void getRightNeigbours (std::set<event, cmp> &rightNeighbours, 
			  typename std::set<event, cmp>::iterator right, 
			  typename std::set<event, cmp>::iterator end,
			  event &insertEvent);
  template <typename cmp>
  void sweep (std::vector<event> &events);
  
  void mergeBlock (block *p, vpscClause *clause,
		   block *b, double dist);
  template<typename cmp>
  void mergeLeft (block *b);
  void mergeRight (block *b);
  void dsearch (unsigned int visit,
		unsigned int &totalOrderPos,
		unsigned int *totalOrder,
		bool *marked);
  void computeTotalOrder (unsigned int *totalOrder);

  template <typename cmp>
  void satisfyVPSC();

  void splitVars (block *b, unsigned int start,
		  unsigned int wall, block *newb);
  void splitBlock (block *b, unsigned int start,
		   unsigned int wall, block *newb);
  template <typename cmp>
  double dsearchlm (unsigned int visit,
		    std::vector<unsigned int> &active,
		    unsigned int visited);
  template <typename cmp>
  void computeLM();
  template <typename cmp>
  void solveVPSC();
  template <typename cmp>
  void setFinalPositions ();
  void cleanBlocks ();

  template <typename cmp>
  void perturbEqual (bool (*sortFn) (const nodePlacePair &p1, 
				     const nodePlacePair &p2));

public:
  inline FastOverlapRemoval(const PropertyContext &context):Layout(context) {};
  inline ~FastOverlapRemoval() {};
  inline bool check (std::string &errMsg) { errMsg = ""; return true; }
  inline void reset () {}
  bool run();
};
/*@}*/
#endif
