/**
 Author: David Auber
 Email : auber@labri.fr
 Last modification : 20/08/2001
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of Compact Graph to improve information density.  For nodes, 
 * history, author and licence please see the header file CompactTree.h
 */

#include <math.h>
#include <float.h>
#include "GrowRoots.h"
#include <tulip/TreeTest.h>
#include <tulip/MetaGraphProxy.h>

SIZESPLUGIN(GrowRoots,"Grow Roots","David Auber/ Daniel Archambault","25/11/2004","Ok","0","1");

using namespace std;

GrowRoots::GrowRoots(const PropertyContext &context):Sizes(context){}

GrowRoots::~GrowRoots() {}

//This function grows the roots of the branches of the tree so
//that their size takes up the appropriate amount of space.
void GrowRoots::growBranchRoots (node &root) {
  //if the out degree is zero, we have reached a leaf and its size
  //should not be adjusted.
  if (superGraph->outdeg (root) == 0) return;

  //this variable holds the max allowable scale factor
  double max_scale_factor = DBL_MAX;

  //this is a proxy for the positions of the nodes of the tree
  LayoutProxy *pos = superGraph->getLocalProperty <LayoutProxy> ("viewLayout");

  //get the root size and radius
  Coord rootPos;
  Size rootSize;
  rootPos = pos->getNodeValue (root);
  rootSize = sizesProxy->getNodeValue (root);
  double rootRad = sqrt ((rootSize.getW()/2.0)*(rootSize.getW()/2.0) + 
			 (rootSize.getH()/2.0)*(rootSize.getH()/2.0));
  
  //go through the children of root and compact them
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();

    //if we run into ourself, skip
    if (curNode == root) continue;

    //get the positions and sizes of the current node
    Coord curPos; 
    Size curSize;
    curPos = pos->getNodeValue (curNode);
    curSize = sizesProxy->getNodeValue (curNode);
    double curRad = sqrt ((curSize.getW()/2.0)*(curSize.getW()/2.0) + 
			  (curSize.getH()/2.0)*(curSize.getH()/2.0));
  
    //calculate the distance between the root and the current node
    double dist = (curPos - rootPos).norm();
  
    //slide the current node as far in close to the centre
    //node as possible, before overlap
    double scale_factor = dist/(curRad + rootRad);

    //if the scale factor is less than the current max_scale_factor,
    //it is the new max_scale_factor
    if (scale_factor < max_scale_factor) max_scale_factor = scale_factor;
  } delete nodes; //end while

  //scale the size of the root node accordingly if it contains a subgraph
  MetaGraphProxy *metagraphs = 
    superGraph->getProperty <MetaGraphProxy> ("viewMetaGraph");
  if (metagraphs->getNodeValue (root) != NULL) {
    rootSize.setW (max_scale_factor*rootSize.getW());
    rootSize.setH (max_scale_factor*rootSize.getH());
    sizesProxy->setNodeValue (root, rootSize);
  }//end if

  //grow all the other roots of the branches in the tree
  nodes = superGraph->getOutNodes(root);
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    growBranchRoots (curNode);
  } delete nodes; //end while
}//end compactBranch  
  
bool GrowRoots::run() {
  //proxy for root sizes
  SizesProxy *sizes = superGraph->getProperty <SizesProxy> ("viewSize");

  //set the original sizes and the depth size to zero.
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Size curSize = sizes->getNodeValue (curNode);
    sizesProxy->setNodeValue (curNode, curSize);
  } delete nodes; //end while
  
  //This is the root of the tree
  node root;
  
  //find the root of the tree.
  nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    if (superGraph->indeg (curNode) == 0) {
      root = curNode;
      break;
    }//end if
  } delete nodes; //end while

  //grow the roots of the branches
  growBranchRoots (root);
  
  //return success
  return true;
}//end run

bool GrowRoots::check (std::string &errMsg) {
    if (!TreeTest::isTree (superGraph)) {
      errMsg = "The passed graph to Compact Tree must be a tree.\n";
      return false;
    }//end if
    return true;
}//end check

void GrowRoots::reset () {
}//end reset
