//-*-c++-*-
#ifndef __HIERLAYOUT_H
#define __HIERLAYOUT_H

/// TopoLayout.h
/** 
 *
 *  An implementation of layout by topological feature.  The plugin
 *  takes a hierarchy created by TopoClustering and computes a layout
 *  based on the passed hierarchy of topological features embedded in
 *  the graph.
 *
 *  HISTORY:
 *
 *  The code was first implemented in Nov/Dec 2004 by Daniel Archambault.
 *
 *  NOTES:
 *
 *  If the plugin does not find a hierarchy it calls Topological Feature
 *  Clustering to create one.
 *
 *  AUTHOR:
 *
 *  Daniel Archambault, University of British Columbia, CA: 
 *  Email: archam@cs.ubc.ca
 *  Version 0.1: 08 December 2004: Initial version
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */

#include <cstring>
#include <vector>
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <tulip/TulipPlugin.h>
#include <tulip/MutableContainer.h>
#include <HierarchyMap.h>

class TopoLayout:public Layout { 
public:
  //A variable to store the node/edge threshold
  double nodeEdgeThresh;

  //A function to recursively set the spoke order around all branch
  //roots.  Thetas will the the theta polar coordinate of the nodes at the
  //level below the root.
  void setSpokeOrderAroundRoot (node root);

  TopoLayout(const PropertyContext &);
  ~TopoLayout();
  bool run();
  bool check(std::string &);
  void reset();

private:
  //time to layout non topology segements of the algorithm
  double timeTorques, timeCentreGlobalPos, 
    timeFastOverlaps, timeClusterBuster;

  //This contains the previous positions of the internal nodes of the
  //quotient graph
  MutableContainer <Coord> prevPos;

  //a proxy to store the position of the hierarchy meta graphs
  MetaGraphProxy *hierarchyMetaGraphs;

  //a proxy that stores the graph a particular node is contained within
  MetaGraphProxy *containedInGraph;

  //This contains a pointer to where the final layout of the graph
  //is stored.
  SuperGraph *finalLayoutGraph;

  //This contains a pointer to the hierarchy
  SuperGraph *hierarchy;

  //This contains a pointer to the graph at the current level
  //and a layout proxy to it
  SuperGraph *featureGraph;
  LayoutProxy *featureLayout;

  //This contains a pointer to the untouched graph (only valid at root)
  SuperGraph *untouched;

  //variables to store if we run certain passes
  bool clusterbuster, fastOverlapRun, torquesRun;

  //When the drawing operator is passed using a plugin, 
  //this contains the topological feature types and algorithms
  StringCollection drawingOperator;

  //These hash maps handle pointers to plugin names and pointers to
  //functions which layout a topological feature type
  //----------------------------------------------------------------------
  struct eqTopoType {
    bool operator() (const char *s1, const char *s2) const {
      return std::strcmp (s1, s2) == 0;
    }//end operator()
  };
  
  class LayoutAlgorithm {
  public:
    //Plugins to run for this algorithm, time to run in two formats,
    //a test to see which format the data is in.
    std::string pluginHigh;
    std::string pluginLow; 
    double time;
    unsigned int longTime;
    bool (*highLowTest) (SuperGraph *featureGraph, TopoLayout *topoLayout);
    
    LayoutAlgorithm () {
      pluginHigh = pluginLow = "";
      time = longTime = 0;
      highLowTest = NULL;
    }//end constructor
  };
  typedef __gnu_cxx::hash_map<const char *, LayoutAlgorithm,
			      __gnu_cxx::hash<const char *>, 
			      eqTopoType> AlgorithmTable;
  AlgorithmTable layoutAlgorithmTable;
  typedef  __gnu_cxx::hash_map<const char *, LayoutAlgorithm,
		      __gnu_cxx::hash<const char *>, eqTopoType>::iterator
  layoutAlgorithmTableIterator;
  //----------------------------------------------------------------------

  //This contains a pointer to the list of original edges in the graph
  //that correspond to this edge of the graph
  MutableContainer <std::vector <edge> > originalEdges;
  
  //a function that finds the subgraph with the passed name
  SuperGraph *findGraph (std::string graphName);

  //A function to recursively lay out a graph hierarchy created by
  //TopoClustering.  Returns true if it succeeds without error.
  bool recurseOnSubHierarchy (SuperGraph *curFeatureGraph);

  //A function to draw the hierarchy
  bool drawHierarchy ();

  //This function sets up the layout proxy of the graph
  //at this level
  void setLayoutProxy ();

  //selects layout algorithm and preforms layout of graph at this level.
  //It returns true if it succeeds without error.
  bool layout ();
  bool graphLayout (LayoutAlgorithm &layoutAlgorithm);
  bool performLayout (const std::string &algorithm, clock_t start, 
		      double &cumulativeTime);

  //function to layout an interior node of the hierarchy
  bool layoutFeatureNode (SuperGraph *feature);

  //a function to generate the originalEdges table at this level of the
  //hierarchy
  void genOriginalEdges ();

  //A function to find the position of an original node of the graph
  //at this level of the hierarchy.  That positition is returned 
  //by the function.  The value of foundIn is set to the node in which
  //originalNode was found.
  Coord findHierarchyPosition (node originalNode, node &foundIn);

  //the following applies torques to the nodes that contain subgraphs
  //at this level.  finalLayout is the graph that stores the final layout
  void applyTorques ();

  
  //This function computes a centre of the graph weighted by node
  //area.  It is required by torques for them to be computed on
  //large layouts.
  void centreByArea (SuperGraph *feature, LayoutProxy *featureLayout);

  //A function that computes the torque exerted on the quotient graph
  //node at radBase by the (radNode, opposite) edge.
  inline double computeTorque (node radNode, node opposite, Coord radBase,
			       double rad);

  //The following function sets the node size of the hierarchy node n
  void setNodeSize (node n, SuperGraph *containedIn);

  //function to handle all user promps to the plugin
  void userPrompts (); 
  void setHighLowTest (const std::string &featureType,
		  LayoutAlgorithm &featurAlgo);
  void functionTableFromUserPrompts ();
  void functionTableFromDrawOperator (std::string drawingOperator);
  void setOtherPassOptions ();

  //function that reports the final times of the components in the algorithm
  //start -- the time the plugin began executing
  inline void reportTimes (clock_t start, time_t startLongTime); 

  //The following function updates node positions according to scaling
  //and translation executed on the virtual node table
  void setFinalPositions (SuperGraph *hierarchy,
			  double cumulativeRotation,
			  Coord cumulativeTrans);
};

//determines if the tree is bushy or not (maxDegree > maxDepth)
bool bushy (node n, node cameFrom, SuperGraph *featureGraph,
	    unsigned int curDepth, 
	    unsigned int &maxDepth, unsigned int &maxDegree);

//A test to determine if a tree is bushy or not
bool isBushy (SuperGraph *tree, TopoLayout *topoLayout);

//A test to see if we have a high node/edge ratio
bool hasHighNodeEdgeRatio (SuperGraph *graph, TopoLayout *topoLayout);

//defines for node/edge ratio thresholds
#define NODE_EDGE_THRESH_DATASET ("Node/Edge")
#define DEFAULT_NODE_EDGE_THRESH ("1.0")

//These define the dataset entries for each topological feature
//----------------------------------------------------------------------
#define DATASET_BUSHY_TREE ("Bushy Tree")
#define DATASET_DEEP_TREE ("Deep Tree")
#define DATASET_CONNECTED ("Connected")
#define DATASET_EMBEDDER ("Near-Mesh")
#define DATASET_STRENGTH_LOW ("Stren Low")
#define DATASET_STRENGTH_HIGH ("Stren High")
#define DATASET_DEFAULT_LOW ("Unkwn Low")
#define DATASET_DEFAULT_HIGH ("Unkwn High")
#define DATASET_LGL_LOW ("LGL Low")
#define DATASET_LGL_HIGH ("LGL High")
//----------------------------------------------------------------------

//These define the set of layout algorithms to be used for a particular
//topological feature
//----------------------------------------------------------------------
#define LAYOUTS_BUSHY_TREE ("Bubble Tree;RINGS Area;Hierarchical Tree (R-T Extended);GEM (Frick) Area;")
#define LAYOUTS_DEEP_TREE ("Hierarchical Tree (R-T Extended);RINGS Area;Bubble Tree;GEM (Frick) Area;")
#define LAYOUTS_CONNECTED ("Connected Componnent Packing;Circular;GEM (Frick) Area;Connected Componnent Packing;")
#define LAYOUTS_EMBEDDER ("Area HDE;High Dimensional Embedder;FM^3;GEM (Frick) Area;")
#define LAYOUTS_STRENGTH_LOW ("GEM (Frick) Area;GRIP;FM^3;Circular;")
#define LAYOUTS_STRENGTH_HIGH ("Circular;GEM (Frick) Area;GRIP;")
#define LAYOUTS_DEFAULT_LOW ("GEM (Frick) Area;GRIP;FM^3;Circular;")
#define LAYOUTS_DEFAULT_HIGH ("Circular;GEM (Frick) Area;GRIP;")
#define LAYOUTS_LGL_HIGH ("LGL Area;GEM (Frick) Area;")
#define LAYOUTS_LGL_LOW ("GEM (Frick) Area;LGL Area;")
#define LAYOUT_FAST_OVERLAP_REMOVAL ("Fast Overlap Removal")
//----------------------------------------------------------------------
#endif





