/**
 Author:  Daniel Archambault
 Email :  archam@cs.ubc.ca
 Last modification : 08/12/2004
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of TopoLayout.  For nodes, history, author and licence
 * please see the header file TopoLayout.h
 */

#include "TopoLayout.h"

using namespace std;

//function to handle all user promps to the plugin
void TopoLayout::userPrompts () {
  addParameter<double> (NODE_EDGE_THRESH_DATASET, 
			NULL, DEFAULT_NODE_EDGE_THRESH);
  addParameter<StringCollection> (DATASET_BUSHY_TREE, NULL,
				  LAYOUTS_BUSHY_TREE);
  addParameter<StringCollection> (DATASET_DEEP_TREE, NULL,
				  LAYOUTS_DEEP_TREE);
  addParameter<StringCollection> (DATASET_CONNECTED, NULL,
				  LAYOUTS_CONNECTED);
  addParameter<StringCollection> (DATASET_EMBEDDER, NULL,
				  LAYOUTS_EMBEDDER);
  addParameter<StringCollection> (DATASET_STRENGTH_LOW, NULL,
				  LAYOUTS_STRENGTH_LOW);
  addParameter<StringCollection> (DATASET_STRENGTH_HIGH, NULL,
				  LAYOUTS_STRENGTH_HIGH);
  addParameter<StringCollection> (DATASET_DEFAULT_LOW, NULL,
				  LAYOUTS_DEFAULT_LOW);
  addParameter<StringCollection> (DATASET_DEFAULT_HIGH, NULL,
				  LAYOUTS_DEFAULT_HIGH);
  addParameter<StringCollection> (DATASET_LGL_LOW, NULL,
				  LAYOUTS_LGL_LOW);
  addParameter<StringCollection> (DATASET_LGL_HIGH, NULL,
				  LAYOUTS_LGL_HIGH);
  addParameter<bool> (DATASET_CLUSTER_BUSTER, NULL, "false");
  addParameter<bool> (DATASET_FAST_OVERLAPS, NULL, "false");
  addParameter<bool> (DATASET_TORQUES, NULL, "true");
}//end userPrompts

//A function to determine which test is associated with this
//topology type
void TopoLayout::setHighLowTest (const string &featureType,
				 LayoutAlgorithm &featureAlgo) {
  //bushy/deep test for trees
  if (isType (featureType, TREE_IDENT) || 
      isType (featureType, BICONNECTED_IDENT))
    featureAlgo.highLowTest = isBushy;
  //connectivity test for clusters
  else if (isType (featureType, STRENGTH_IDENT) || 
	   isType (featureType, UNKNOWN_IDENT))
    featureAlgo.highLowTest = hasHighNodeEdgeRatio;
  //no test for anything else
  featureAlgo.highLowTest = NULL;
}//end setHighLowTest

//Gets the user prompts from the TopoLayout Dialogue
void TopoLayout::functionTableFromUserPrompts () {
  StringCollection selectedString(";"); //the string selected by a combo box
  LayoutAlgorithm &tree = layoutAlgorithmTable[TREE_IDENT];
  LayoutAlgorithm &connected = layoutAlgorithmTable[CONNECTED_IDENT];
  LayoutAlgorithm &embedder = layoutAlgorithmTable[EMBEDDER_IDENT];
  LayoutAlgorithm &biconnected = layoutAlgorithmTable[BICONNECTED_IDENT];
  LayoutAlgorithm &strength = layoutAlgorithmTable[STRENGTH_IDENT];
  LayoutAlgorithm &singletons = layoutAlgorithmTable[SINGLETONS_IDENT];
  LayoutAlgorithm &lgl = layoutAlgorithmTable[LGL_IDENT];
  LayoutAlgorithm &unknown = layoutAlgorithmTable[UNKNOWN_IDENT];
  setHighLowTest (TREE_IDENT, tree);
  setHighLowTest (CONNECTED_IDENT, connected);
  setHighLowTest (EMBEDDER_IDENT, embedder);
  setHighLowTest (BICONNECTED_IDENT, biconnected);
  setHighLowTest (STRENGTH_IDENT, strength);
  setHighLowTest (UNKNOWN_IDENT, unknown);
  setHighLowTest (SINGLETONS_IDENT, singletons);
  setHighLowTest (LGL_IDENT, lgl);

  dataSet->get (DATASET_BUSHY_TREE, selectedString);
  tree.pluginHigh = selectedString.getCurrentString();
  dataSet->get (DATASET_DEEP_TREE, selectedString);
  tree.pluginLow = selectedString.getCurrentString();
  biconnected = tree;
  singletons = tree;
  dataSet->get (DATASET_CONNECTED, selectedString);
  connected.pluginHigh = selectedString.getCurrentString();
  dataSet->get (DATASET_EMBEDDER, selectedString);
  embedder.pluginHigh = selectedString.getCurrentString();
  dataSet->get (DATASET_STRENGTH_LOW, selectedString);
  strength.pluginLow = selectedString.getCurrentString();
  dataSet->get (DATASET_STRENGTH_HIGH, selectedString);
  strength.pluginHigh = selectedString.getCurrentString();
  dataSet->get (DATASET_DEFAULT_LOW, selectedString);
  unknown.pluginLow = selectedString.getCurrentString();
  dataSet->get (DATASET_DEFAULT_HIGH, selectedString);
  unknown.pluginHigh = selectedString.getCurrentString();
  dataSet->get (DATASET_LGL_LOW, selectedString);
  lgl.pluginLow = selectedString.getCurrentString();
  dataSet->get (DATASET_LGL_HIGH, selectedString);
  lgl.pluginHigh = selectedString.getCurrentString();
}//end functionTableFromUserPrompts

//Get the function values from the drawing operator string collection
//feature; high layout; low layout ...
void TopoLayout::
functionTableFromDrawOperator (const string layoutOperator) {
  drawingOperator = StringCollection (layoutOperator);
  StringCollection::iterator drawOpIt = drawingOperator.begin();
  while (drawOpIt != drawingOperator.end()) {
    string pluginHigh = "", pluginLow = "";
    const char *featureType = drawOpIt->c_str();
    ++drawOpIt;
    if (drawOpIt != drawingOperator.end()) {
      pluginHigh = *drawOpIt;
      ++drawOpIt;
    }//end if
    if (drawOpIt != drawingOperator.end()) {
      pluginLow = *drawOpIt;
      ++drawOpIt;
    }//end if

    if (string (featureType) != "") {
      LayoutAlgorithm &featureAlgo = layoutAlgorithmTable[featureType];
      featureAlgo.pluginHigh = pluginHigh;
      featureAlgo.pluginLow = pluginLow;
      setHighLowTest (featureType, featureAlgo);
    }//end if
  }//end for
}//end drawingOperator

//A function that gets passed options from the user
void TopoLayout::setOtherPassOptions() {
  dataSet->get (DATASET_FAST_OVERLAPS, fastOverlapRun);
  dataSet->get (DATASET_CLUSTER_BUSTER, clusterbuster);
  dataSet->get (DATASET_TORQUES, torquesRun);
  dataSet->get (NODE_EDGE_THRESH_DATASET, nodeEdgeThresh);
}//end void
