#include <tulip/MutableContainer.h>
#include <tulip/MetaGraphProxy.h>
#include <tulip/TlpTools.h>
#include <math.h>
#include "HierLayoutOneLevel.h"

using namespace std;

LAYOUTPLUGIN(HierLayoutOneLevel,"HierLayoutOneLevel","Daniel Archambault","11/16/2004","Alpha","0","1");

HierLayoutOneLevel::HierLayoutOneLevel(const PropertyContext &context):Layout(context){
}

HierLayoutOneLevel::~HierLayoutOneLevel() {
}

//a function to do squares
inline double sqr (double x) {
  return x*x;
}//end sqr

//an inline function to help prevent tortional oscillations
inline double attenu_tau (double tau, double prevTau, double time) {
  //return quadratic damping
  return tau*time*time;
}//end attenu_tau

//a member function to compute torque
inline double HierLayoutOneLevel::computeTorque (node radNode, node opposite, 
						 Coord radBase, double rad) {
  Coord f = prevPos.get (opposite.id) - prevPos.get (radNode.id);
  f.setZ (0);
  if (f.norm() > 1e-3) f /= f.norm (); 
  Coord r = prevPos.get (radNode.id) - radBase;
  r.setZ (0);
  if (rad > 1e-3) r /= rad;

  //compute the angle between the two vectors
  //r * f bugs gives nan (I do not know why???)
  double angle = acosf ((r.getX()*f.getX() + r.getY()*f.getY()) / r.norm ()); //f is unit (or should be)
  
  
  //compute the cross for the correct sign
  double cross = r.getX()*f.getY() - r.getY()*f.getX();

  //return a half angle torque (high twist at M_PI not M_PI/2 with the
  //right sign.
  return ((cross < 0) ? -1 : 1)*(r.norm() * sinf (angle/2.0));
}//end computeTorque

//a member function to apply the torque to the set
//of inside nodes of a quotient graph
#define MAX_TURN (M_PI) //do not allow rotation over 180 at a timestep
inline void HierLayoutOneLevel::rotateByTau (SuperGraph *rotGraph, 
					     double tau, Coord trans) {
  if (tau > MAX_TURN) tau = MAX_TURN;
  Iterator <node> *nodes = rotGraph->getNodes();
  while (nodes->hasNext ()) {
    node curNode = nodes->next ();
    Coord result = layoutProxy->getNodeValue (curNode);
    result -= trans;
    Coord resultCopy = result;
    result.setX (resultCopy.getX()*cos(tau) - resultCopy.getY()*sin (tau));
    result.setY (resultCopy.getX()*sin(tau) + resultCopy.getY()*cos (tau));
    result.setZ (0);
    result += trans;
    layoutProxy->setNodeValue (curNode, result);
  }//end while
  delete nodes;
}//end rotateByTau

//This inline function centres a cluster in its own, local layoutProxy.
//The centering does not take into account node size.  It simply
//matches up the origin of the local coordinate frame with the origin
//of the quotient coordinate frame
inline void HierLayoutOneLevel::centreGraph (SuperGraph *cluster) {
  LayoutProxy *localLayout = 
    cluster->getLocalProperty<LayoutProxy> ("viewLayout");
  Coord max = localLayout->getMax (cluster);
  Coord min = localLayout->getMin (cluster);
  Coord centre;
  centre.setX ((max.getX() + min.getX())/2.0);
  centre.setY ((max.getY() + min.getY())/2.0);
  centre.setZ (0);
  
  //adjust the locations of the coordinates
  Iterator <node> *nodes = cluster->getNodes ();
  while (nodes->hasNext ()) {
    node curNode = nodes->next();
    Coord shifted = localLayout->getNodeValue(curNode) - centre;
    localLayout->setNodeValue (curNode, shifted);
  }//end while
  delete nodes; 
}//end 

//This function sets up the layout proxies of the passed graph
void HierLayoutOneLevel::setLayoutProxies (SuperGraph *graph) {
  
  //copy over the global positions of the nodes
  LayoutProxy *inputLayout = 
    superGraph->getProperty<LayoutProxy> ("viewLayout");
  Iterator <node> *nodes = superGraph->getNodes ();
  while (nodes->hasNext ()) {
    node curNode = nodes->next ();
    layoutProxy->setNodeValue (curNode, 
			       inputLayout->getNodeValue (curNode));
  }//end while 
  delete nodes;
  

  //copy these globals into local positions as well
  Iterator <SuperGraph *> *superGraphs = graph->getSubGraphs ();
  while (superGraphs->hasNext ()) {
    SuperGraph *curGraph = superGraphs->next ();
    LayoutProxy *inputSGLayout =
      curGraph->getProperty<LayoutProxy> ("viewLayout");
    Iterator <node> *sgNodes = curGraph->getNodes ();
    while (sgNodes->hasNext ()) {
      node curNode = sgNodes->next();
      inputSGLayout->setNodeValue (curNode, 
				   layoutProxy->getNodeValue (curNode));
    }//end while
    delete sgNodes;
  }//end while
  delete superGraphs;  
}//end setLayoutProxies

//the following algorithm computes the layout of the subgraphs of
//curGraph.  Right now, it does spring-electrical for arbitrary graphs
//and circular for graphs with no edges.
void HierLayoutOneLevel::layoutClusters (SuperGraph *clusterRoot) {
  //a string for error messages
  string errMsg;

  //go through the subgraphs and create cluster nodes
  Iterator <SuperGraph *> *clusters = clusterRoot->getSubGraphs();
  while (clusters->hasNext ()) {
    //get the cluster
    SuperGraph *curCluster = clusters->next();

    //if there are no edges, layout circular
    if (curCluster->numberOfEdges() == 0) {
      //layout graph with only nodes
      curCluster->computeProperty ("Circular",
				   curCluster->getLocalProperty<LayoutProxy> 
				   ("viewLayout"), errMsg);
    }//end if

    //otherwise do an initial circular layout and then perform spring
    //electrical.
    else {
      LayoutProxy *clusterLayout = curCluster->getLocalProperty<LayoutProxy> ("viewLayout");

      //do the two layouts
      curCluster->computeProperty ("Circular", clusterLayout, errMsg);
      LayoutProxy tmpLayout(curCluster);
      curCluster->computeProperty ("Spring_Electrical", &tmpLayout, errMsg);
      
      //set the local positions of the nodes.
      Iterator <node> *nodes = curCluster->getNodes();
      while (nodes->hasNext ()) {
	node curNode = nodes->next();
	Coord curCoord = tmpLayout.getNodeValue(curNode); 
	curCoord.setZ(0);
	clusterLayout->setNodeValue(curNode,curCoord);
      } //end while
      delete nodes;
    }//end else
    centreGraph (curCluster);
  }//end while
  delete clusters;
}//end layoutClusters

//The following function unravels the hierarchy to set final global 
//positions of the nodes
void HierLayoutOneLevel::setGlobalPositions (SuperGraph *rootGraph, 
					     SuperGraph *quotientGraph) {
  //go through all t" "he nodes of the quotient graph
  Iterator <node> *quotientGraphNodes = quotientGraph->getNodes ();
  LayoutProxy *quotientLayout = 
    quotientGraph->getLocalProperty<LayoutProxy> ("viewLayout");
  MetaGraphProxy *metaGraphs =
    quotientGraph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  while (quotientGraphNodes->hasNext ()) {
    node curQuotientNode = quotientGraphNodes->next();

    //get the translation vector to the position of the quotient
    //graph node for this root graph node.
    Coord quotientTrans = quotientLayout->getNodeValue(curQuotientNode);

    //go inside the quotient graph and get its inside nodes here somehow.
    //the inside subgraph of the quotient node is quotientSubGraph
    //get those nodes and put them inside insideNodes
    //get a layout proxy and put it inside insideNodesLayout
    SuperGraph *curQuoSubGraph = metaGraphs->getNodeValue(curQuotientNode);

    //if it is null there is no inside subgraph
    if (curQuoSubGraph == NULL) continue;
    Iterator <node> *insideNodes = curQuoSubGraph->getNodes ();
    LayoutProxy *insideNodesLayout = 
      curQuoSubGraph->getLocalProperty<LayoutProxy> ("viewLayout");
    while (insideNodes->hasNext ()) {
      node curInsideNode = insideNodes->next ();
      Coord localTrans = insideNodesLayout->getNodeValue (curInsideNode);
      layoutProxy->setNodeValue (curInsideNode, quotientTrans + localTrans);
    }//end while
    delete insideNodes;
  }//end while
  delete quotientGraphNodes;
}//end setGlobalPositions 

//the following applies torques to the quotient graph nodes
//in order to prevent edge overlap with elements of the quotient
//graph
void HierLayoutOneLevel::applyTorques (SuperGraph *rootGraph, 
				       SuperGraph *quotientGraph) {
  //get the layout and metagraph proxies of the quotient graph
  LayoutProxy *quotientLayout = 
    quotientGraph->getLocalProperty<LayoutProxy> ("viewLayout");
  SizesProxy *quotientSizes = 
    quotientGraph->getProperty<SizesProxy> ("viewSize");
  MetaGraphProxy *metaGraphs =
    quotientGraph->getProperty<MetaGraphProxy> ("viewMetaGraph");

  //set all previous torques to 0
  prevTau.setAll (0);
  
  //we compute the torque over all nodes of the quotent graph, numNodes times
  unsigned int iterations = quotientGraph->numberOfNodes();
  for (unsigned int i = 0; i < iterations; i++) {

    //go through all the nodes of the quotient graph
    Iterator <node> *quotientGraphNodes = quotientGraph->getNodes ();

    //back up the positions of the nodes
    Iterator <node> *nodeBackup = superGraph->getNodes ();
    while (nodeBackup->hasNext()) {
      node curNode = nodeBackup->next ();
      prevPos.set (curNode.id, layoutProxy->getNodeValue (curNode));
    }//end while
    delete nodeBackup;
   
    while (quotientGraphNodes->hasNext ()) {
      node curQuotientNode = quotientGraphNodes->next();
      
      //get the translation vector to the position of the quotient
      //graph node for this root graph node.
      Coord quotientTrans = quotientLayout->getNodeValue(curQuotientNode);

      //get the radius of the bounding circle of the quotent node
      double rad;
      Size quoSize = quotientSizes->getNodeValue (curQuotientNode);
      rad = sqrt (sqr(quoSize.getW()/2.0) + sqr(quoSize.getH()/2.0));
    
      //go inside the quotient graph and get its inside nodes here somehow.
      //the inside subgraph of the quotient node is quotientSubGraph
      //get those nodes and put them inside insideNodes
      //get a layout proxy and put it inside insideNodesLayout
      SuperGraph *curQuoSubGraph = metaGraphs->getNodeValue (curQuotientNode);
      
      //if it is null there is no inside subgraph
      if (curQuoSubGraph == NULL) continue;
      Iterator <node> *insideNodes = curQuoSubGraph->getNodes ();
      LayoutProxy *insideNodesLayout = 
	curQuoSubGraph->getLocalProperty<LayoutProxy> ("viewLayout");
      double tau = 0; //the total torque for each quotient graph node
      while (insideNodes->hasNext ()) {
	node curInsideNode = insideNodes->next ();
	Iterator <edge> *nodeEdges = superGraph->getInOutEdges (curInsideNode);
	while (nodeEdges->hasNext ()) {
	  edge curEdge = nodeEdges->next();
	  if (!curQuoSubGraph->isElement (curEdge)) {
	    node opposite = superGraph->opposite (curEdge, curInsideNode);
	    tau += computeTorque (curInsideNode, opposite, 
				  quotientTrans, rad);
	  }//end if
	}//end while
	delete nodeEdges;
      }//end while
      delete insideNodes;

      //the time to go as %
      double time = (iterations - i)/static_cast<double> (iterations);

      //dampen to prevent oscillations
      tau = attenu_tau (tau, prevTau.get(curQuotientNode.id), time);
      
      //apply the torque to all the inside nodes
      rotateByTau (curQuoSubGraph, tau, quotientTrans);

      //record the applied torque
      prevTau.set (curQuotientNode.id, tau);

    }//end while
    delete quotientGraphNodes;
  }//end for
}//end applyTorques

bool HierLayoutOneLevel::run() {
  //for the error message
  string errMsg;

  //map the input layout to the layout proxys
  setLayoutProxies (superGraph);

  //Layout the clusters of the superGraph
  layoutClusters (superGraph);

  //compute the quotient graph of the layed-out clusters with size set
  DataSet dataSet;
  tlp::clusterizeGraph (superGraph, errMsg, &dataSet, "QuotientClustering");
  SuperGraph *quotientGraph;
  dataSet.get<SuperGraph *>("quotientGraph", quotientGraph);
  LayoutProxy *quotientLayout = quotientGraph->getLocalProperty<LayoutProxy> ("viewLayout");
  Iterator <node> *nodes;

  //Draw the quotent graph using spring-electrical (draws out
  //nodes with area dependance) with a cirular staring position
  quotientGraph->computeProperty ("Circular",
				  quotientLayout, errMsg);
  LayoutProxy tmpLayout(quotientGraph);
  quotientGraph->computeProperty ("Spring_Electrical",
				  &tmpLayout, errMsg);
  nodes = quotientGraph->getNodes();
  while (nodes->hasNext ()) {
    node curNode = nodes->next();
    Coord curCoord = tmpLayout.getNodeValue(curNode); 
    curCoord.setZ(0);
    layoutProxy->setNodeValue(curNode,curCoord);
    quotientLayout->setNodeValue(curNode,curCoord);
  } delete nodes;
  //centreGraph (quotientGraph);

  //set the global positions of the nodes 
  setGlobalPositions (superGraph, quotientGraph);

  //compute the required torques for each node of the quotient graph.
  //and rotate them.
  applyTorques (superGraph, quotientGraph);

  return true;

}//end run

bool HierLayoutOneLevel::check(string &erreurMsg){
  erreurMsg="";
  return true;
}

void HierLayoutOneLevel::reset(){}

