/**
 Author: Daniel Archambault
 Email : archam@cs.ubc.ca
 Last modification : 08/12/2004
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/
/**
 * implementation of Resolve Overlap.  For nodes, history, 
 * author and licence please see the header file ResolveOverlap.h
 */

#include <math.h>
#include <vector>
#include "ResolveOverlap.h"
#include <BoundingObjects.h>

LAYOUTPLUGIN(ResolveOverlap,"Resolve Overlap","Daniel Archambault","08/12/2004","Ok","0","1");

using namespace std;

ResolveOverlap::ResolveOverlap(const PropertyContext &context):Layout(context){}
ResolveOverlap::~ResolveOverlap() {}

//this function centres a graph around the passed node centre
void ResolveOverlap::centreGraph (node centre) {
  //get the layout proxy and the position of the centre
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  Coord centrePos = pos->getNodeValue (centre);
  
  //centre nodes around centrePos
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    Coord position = pos->getNodeValue (curNode);
    pos->setNodeValue (curNode, position - centrePos);
  }//end while
  delete nodes;
  
  //centre edge bends around centerPos
  Iterator <edge> *edges = superGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next ();
    vector <Coord> bends = pos->getEdgeValue (curEdge);
    for (unsigned int i = 0; i < bends.size (); i++) 
      bends[i] = bends[i] - centrePos;
    pos->setEdgeValue (curEdge, bends);
  }//end while
  delete edges;
}//end centre

bool ResolveOverlap::run() {
  //proxy for sizes and positions
  SizesProxy *nodeSizes = superGraph->getProperty <SizesProxy> ("viewSize");
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  
  //this is the minimum scale factor required to ensure no two nodes are 
  //overlapping.
  double max_scale_factor = 1;
  
  //this is the node we centre around for the scaling
  unsigned int centre_node;
  
  //A vector of nodes in the graph
  vector <node> v_nodes;
  
  //put all the nodes in the vector and set all z to zero
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext ()) {
    node curNode = nodes->next();
    v_nodes.push_back (curNode);
    Coord position = pos->getNodeValue (curNode);
    pos->setNodeValue (curNode, Coord (position.getX(), position.getY(), 0));
  }//end while
  delete nodes;
  
  //load all node positions into the layout proxy
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++)
    layoutProxy->setNodeValue (v_nodes[i], 
			       pos->getNodeValue (v_nodes[i]));
  
  //go through every pair of nodes in the vector and determine the
  //max node overlap.
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    for (unsigned int j = i + 1; j < superGraph->numberOfNodes (); j++) {
      //get the positions and sizes of the nodes
      Coord pos_i = pos->getNodeValue (v_nodes[i]);
      Coord pos_j = pos->getNodeValue (v_nodes[j]);
      Size size_i = nodeSizes->getNodeValue (v_nodes[i]);
      Size size_j = nodeSizes->getNodeValue (v_nodes[j]);
      
      //calculate the distance between the two nodes
      double dist = (pos_i - pos_j).norm();
      
      //if nodes are nearly at the same position, the algorithm fails
      if (dist < 1e-6) {
	cerr << "Zero distance between a pair of nodes initially.\n" 
	     << "Resolve Overlap fails." << endl;
	cerr << "Acutal Distance between " << v_nodes[i].id << " and "
	     << v_nodes[j].id << " is " << dist << endl;
	return false;
      }//end else
	    
      //compute the scale factor to take the current distance
      //between the nodes to a distance that just separates them
      double scale_factor = getScaleFactor (size_i, size_j, pos_i, pos_j);
      
      //determine the max overlap
      if (scale_factor > max_scale_factor) {
	max_scale_factor = scale_factor;
	centre_node = (rad_i > rad_j) ? i : j;
      }//end if
    }//end for i
  }//end for j 
    
  //if we have a scale factor greater than one, we apply it.
  if (max_scale_factor > 1) {
    
    //report scale factor and scale centre
    cerr << "Found scale centre " << v_nodes[centre_node].id << endl;
    cerr << "Scaling by " << max_scale_factor << " around it." << endl;
    
    //centre the graph around the centre_node
    centreGraph (v_nodes[centre_node]);
    
    //scale all the nodes out from the centre node
    for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
      //scale the nodes out about the centre
      Coord newPos = pos->getNodeValue (v_nodes[i]);
      newPos *= max_scale_factor;
      layoutProxy->setNodeValue (v_nodes[i], newPos);
    }//end for i
    
    //set the new positions for the edge bends
    Iterator <edge> *edges = superGraph->getEdges();
    while (edges->hasNext()) {
      edge curEdge = edges->next ();
      vector <Coord> bends = pos->getEdgeValue (curEdge);
      for (unsigned int i = 0; i < bends.size (); i++)
	bends[i] *= max_scale_factor;
      layoutProxy->setEdgeValue (curEdge, bends); 
    }//end while
    delete edges;
  }//end if
  
  return true;
}//end run
