/**
 Author: David Auber
 Email : auber@labri.fr
 Last modification : 20/08/2001
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/

/**
 * implementation of OverlapsFinalLayout to resolve overlaps.  For nodes, 
 * history, author and licence please see the header file 
 * OverlapsFinalLayout.h
 */

#include <math.h>
#include <float.h>
#include <algorithm>
#include "OverlapsFinalLayout.h"
#include <tulip/TlpTools.h>
#include <tulip/StableIterator.h>

SIZESPLUGIN(OverlapsFinalLayout,"Overlaps Final Layout","Daniel Archambault","08/12/2004","Ok","0","1");

using namespace std;

typedef vector<Coord>::iterator coordIt;
typedef vector<node>::iterator nodeIt;
typedef vector<double>::iterator doubleIt;

OverlapsFinalLayout::OverlapsFinalLayout(const PropertyContext &context):Sizes(context) {
}

OverlapsFinalLayout::~OverlapsFinalLayout() {
}

//Resolve overlaps at a particular level of the final hierarchical layout.
bool OverlapsFinalLayout::overlapsLevel(SuperGraph *graph,
					string &compactType) {
  //decide if atLeastOneHierarchyNode is required
  string name = graph->getAttribute<string> ("name");
  bool atLeastOneHierarchyNode;
  if (overlapsAllPairs == NULL)
    atLeastOneHierarchyNode = false;
  else {
    atLeastOneHierarchyNode = true;
    for (unsigned int i = 0; i < overlapsAllPairs->size(); i++)
      if (name.find ((*overlapsAllPairs)[i]) == 0)
	atLeastOneHierarchyNode = false;
  }//end else

  //a shrink factor for each node in the subgraph
  vector <double> minShrinkFactors;

  //put all the nodes of the graph inside a vector
  vector<node> nodeList;
  vector<Coord> minShrinkRads;
  Iterator<node> *nodes = graph->getNodes();
  while (nodes->hasNext()) {
    nodeList.push_back (nodes->next());
    minShrinkFactors.push_back (1.0);
    minShrinkRads.push_back (Coord (0, 0, 0));
  } delete nodes; //end while

  //pairwise shrink every pair of nodes
  MetaGraphProxy *metagraphs = 
    graph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  SizesProxy *nodeSizes = 
    graph->getProperty<SizesProxy> ("viewSize");
  LayoutProxy *nodeLayout = 
    graph->getLocalProperty<LayoutProxy> ("viewLayout");
  MetricProxy *nodeRotation = 
    graph->getLocalProperty<MetricProxy> ("viewRotation");
  MetricProxy *nodeScale = 
    graph->getLocalProperty<MetricProxy> ("viewScale");
  nodeScale->setAllNodeValue (1.0);
  for (unsigned int i = 0; i < nodeList.size(); i++) {
    //only compute shrink if one node is a hierarchy node 
    //atLeastOneHierarchyNode is true
    node &node_i = nodeList[i];
    if (atLeastOneHierarchyNode && 
	(metagraphs->getNodeValue (node_i) == NULL)) continue;

    for (unsigned int j = i + 1; j < nodeList.size(); j++) {

      node &node_j = nodeList[j];
      
      //get the positions and sizes of the nodes
      Coord pos_i = nodeLayout->getNodeValue (node_i);
      Coord pos_j = nodeLayout->getNodeValue (node_j);
      double rot_i = nodeRotation->getNodeValue (node_i);
      double rot_j = nodeRotation->getNodeValue (node_j);

      //compute the distance between the nodes
      double dist = (pos_i - pos_j).norm();

      //if nodes are nearly at the same position, the algorithm fails
      if (dist < 1e-16) {
	cerr << "Zero distance between a pair of nodes initially.\n" 
	     << "Shrink Resolve Overlap fails for this pair." << endl;
	cerr << "Acutal Distance between " << " is " << dist << endl;
	continue;
      }//end else
      

      //compute the scale factor (1.0 are scales as the nodes at this
      //level should not have been scaled yet).
      double scale_factor = getShrinkFactor (compactType, pos_i, pos_j,
					     1.0, 1.0, rot_i, rot_j, 
					     (*boundaryTable)[node_i.id], 
					     (*boundaryTable)[node_j.id]);

      //the min shrink that needs to be applied
      if (scale_factor < minShrinkFactors[i]) {
	minShrinkFactors[i] = scale_factor;
	minShrinkRads[i] = getRad (compactType, pos_j - pos_i, 1.0, 
				   rot_i, (*boundaryTable)[node_i.id]);
      }
      if (scale_factor < minShrinkFactors[j]) {
	minShrinkFactors[j] = scale_factor;
	minShrinkRads[j] = getRad (compactType, pos_i - pos_j, 1.0, 
				   rot_j, (*boundaryTable)[node_j.id]);
      }
    }//end for j
  }//end for i

  //There was no overlap resolution until one was executed
  (*shrinkResolveTable) [name.c_str()] = false;

  //go through and shrink pairs of nodes as necissary.
  nodeIt curNode = nodeList.begin();
  doubleIt curShrink = minShrinkFactors.begin();
  coordIt curRad = minShrinkRads.begin();
  for (; curNode != nodeList.end(); ++curNode, ++curShrink, ++curRad) {
    //shrink around node centre if scale factor is less than one
    if (*curShrink < 1) {
      cout << "shrinking  " << curNode->id << " " << *curShrink << endl;
      nodeScale->setNodeValue (*curNode, *curShrink);
      Size curNodeSize = nodeSizes->getNodeValue (*curNode);
      nodeSizes->setNodeValue (*curNode, Size (*curShrink*curNodeSize.getW(),
					       *curShrink*curNodeSize.getH(),
					       *curShrink*curNodeSize.getD()));
      Coord position = nodeLayout->getNodeValue (*curNode);
      //drawRad (graph, *curShrink*(*curRad), position);
      (*shrinkResolveTable) [name.c_str()] = true;
    }//end if
  }//end for i
  return true;
}//end overlapsLevel

//Recursively calls OverlapsFinalLayout to compact all the subgraphs and
//the final graph.
bool OverlapsFinalLayout::recurseOnSubGraphs (SuperGraph *graph,
					      string &compactType) {
  //for error messages
  string errMsg;
  //success of recursive calls
  bool success = true;

  //reecurively shrink the subgraphs
  MetaGraphProxy *metagraphs = 
    graph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  SizesProxy *sizes =
    graph->getProperty<SizesProxy> ("viewSize");
  Iterator <node> *nodes = graph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    SuperGraph *curGraph = metagraphs->getNodeValue (curNode);

    //the size is either decided by the subgraph or, if there is no
    //subgraph, the node itself
    if (curGraph != NULL) {
      success = recurseOnSubGraphs (curGraph, compactType);
      if (!success) {
	string name = curGraph->getAttribute <string> ("name");
	cerr << "Error recursing on subgraph " << name << endl;
	return false;
      }//end if
      updateBoundary (*boundaryTable, curNode, curGraph, compactType);
    }//end if
    else (*boundaryTable)[curNode.id] = 
	   leafNodeBoundary (sizes->getNodeValue (curNode));
  } delete nodes; //end while

  //resolve overlaps at the level
  string name = graph->getAttribute <string> ("name");
  cerr << "resolving overlaps of " << name << endl;
  bool level_success = overlapsLevel (graph, compactType);
  if (!level_success) {
    cerr << "Error resolving overlaps of graph " << name << endl;
    return false;
  }//end if
  return true;
}//end recurseOnSubGraphs


bool OverlapsFinalLayout::run() {

  //load the sizes into a sizes proxy
  SizesProxy *sizes = superGraph->getProperty<SizesProxy> ("viewSize");
  Iterator <node> *bkpnodes = superGraph->getNodes();
  while (bkpnodes->hasNext()) {
    node curNode = bkpnodes->next();
    sizesProxy->setNodeValue (curNode, sizes->getNodeValue (curNode));
  } delete bkpnodes; //end while
  
  //get the virtual node table
  bool madeBT = !dataSet->get <BoundaryTable *> (DATASET_BOUNDARY_TABLE,
						 boundaryTable);
  if (madeBT) {
    cerr << "Boundary table not specified.  Making one and deleting it at"
	 << "the end of plugin execution " << endl;
    boundaryTable = new BoundaryTable;
    dataSet->set <BoundaryTable *> (DATASET_BOUNDARY_TABLE,
				    boundaryTable);
  }//end if

  //get the shrinkResolve table
  bool madeRT = !dataSet->get <shrinkResTable *> (DATASET_SHRINK_RESOLVE_TABLE,
						  shrinkResolveTable);
  if (madeRT) {
    shrinkResolveTable = new shrinkResTable;
    dataSet->set <shrinkResTable *> (DATASET_SHRINK_RESOLVE_TABLE,
				     shrinkResolveTable);
  }//end if

  //get the type of compaction required
  string compactType;
  bool success = dataSet->get <string> (DATASET_BOUNDING_TYPE, 
					compactType);
  if (!success) {
    cerr << "Bounding type not specified.  Using bounding circle by default."
	 << endl;
    compactType = BOUNDING_CIRCLE;
  }//end if

  //get the list of graphs that require compaction of all pairs
  bool compactAllPairsSuccess = 
    dataSet->get <vector <string> *> (DATASET_ALL_PAIRS, overlapsAllPairs);
  if (!compactAllPairsSuccess) overlapsAllPairs = NULL;
  
  //Recurse on the the other sub graphs
  bool recursive_success = recurseOnSubGraphs (superGraph, compactType);

  //if we made the virtual node table (at root) delete it
  if (madeBT) delete boundaryTable;

  //if we made the shrink resolve table (at root) delete it
  if (madeRT) delete shrinkResolveTable;

  //reslove overlaps the level and return the result
  return recursive_success;
}//end run

bool OverlapsFinalLayout::check(string &erreurMsg){
  erreurMsg="";
  return true;
}

void OverlapsFinalLayout::reset(){}
