/**
 Author: Daniel Archambault
 Email : archam@cs.ubc.ca
 Last modification : 08/12/2004
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/
/**
 * implementation of Resolve Overlap.  For nodes, history, 
 * author and licence please see the header file ShrinkResolveOverlap.h
 */

#include <vector>
#include "ShrinkResolveOverlap.h"
#include <BoundingObjects.h>

SIZESPLUGIN(ShrinkResolveOverlap,"Shrink Resolve Overlap","Daniel Archambault","08/12/2004","Ok","0","1");

using namespace std;

ShrinkResolveOverlap::ShrinkResolveOverlap(const PropertyContext &context):Sizes(context){}

ShrinkResolveOverlap::~ShrinkResolveOverlap() {}

bool ShrinkResolveOverlap::run() {
  //proxy for sizes and positions
  SizesProxy *nodeSizes = superGraph->getProperty <SizesProxy> ("viewSize");
  LayoutProxy *pos = superGraph->getProperty <LayoutProxy> ("viewLayout");
  
  //a table of shrink factors
  vector <double> minShrinkFactors;
  for (unsigned int i = 0; i < superGraph->numberOfNodes(); i++)
    minShrinkFactors.push_back (1.0);
  
  //A vector of nodes in the graph
  vector <node> v_nodes;
  
  //put all the nodes in the vector and set all z to zero
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext ()) {
    node curNode = nodes->next();
    v_nodes.push_back (curNode);
    Coord position = pos->getNodeValue (curNode);
    pos->setNodeValue (curNode, Coord (position.getX(), position.getY(), 0));
  }//end while
  delete nodes;
  
  //load all the initial sizes into the sizes proxy
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    sizesProxy->setNodeValue (v_nodes[i], 
			      nodeSizes->getNodeValue (v_nodes[i]));
  }//end for i
  
  //go through every pair of nodes in the vector and determine the
  //max node overlap.
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    for (unsigned int j = i + 1; j < superGraph->numberOfNodes (); j++) {
      //get the positions and sizes of the nodes
      Coord pos_i = pos->getNodeValue (v_nodes[i]);
      Coord pos_j = pos->getNodeValue (v_nodes[j]);
      Size size_i = nodeSizes->getNodeValue (v_nodes[i]);
      Size size_j = nodeSizes->getNodeValue (v_nodes[j]);
      vector <Coord> convexHull_i;
      vector <Coord> convexHull_j;

      //compute the distance between the nodes
      double dist = (pos_i - pos_j).norm();
      
      //if nodes are nearly at the same position, the algorithm fails
      if (dist < 1e-6) {
	cerr << "Zero distance between a pair of nodes initially.\n" 
	     << "Shrink Resolve Overlap fails for this pair." << endl;
	cerr << "Acutal Distance between " << v_nodes[i].id << " and "
	     << v_nodes[j].id << " is " << dist << endl;
	continue;
      }//end else
      
      //compute the factor to shrink
      //you probably want to keep this at circle for torques could cause
      //overlaps
      double scale_factor = getShrinkFactor (BOUNDING_CIRCLE,size_i, size_j, 
					     pos_i, pos_j,
					     0, 0,
					     convexHull_i,
					     convexHull_j);
      
      //the min shrink that needs to be applied
      if (scale_factor < minShrinkFactors[i])
	minShrinkFactors[i] = scale_factor;
      if (scale_factor < minShrinkFactors[j])
	minShrinkFactors[j] = scale_factor;
    }//end for i
  }//end for j 
  
  //go through and shrink pairs of nodes as necissary.
  for (unsigned int i = 0; i < superGraph->numberOfNodes (); i++) {
    //report scale factor and scale centre
    if (floor (minShrinkFactors[i]) < 1) {
      Size nodeSize = nodeSizes->getNodeValue (v_nodes[i]);
      nodeSize.setW (minShrinkFactors[i]*nodeSize.getW());
      nodeSize.setH (minShrinkFactors[i]*nodeSize.getH());
      sizesProxy->setNodeValue (v_nodes[i], nodeSize);
    }//end if
  }//end for i
  return true;
}//end run

