#include <tulip/MutableContainer.h>
#include "SpringElectrical.h"
#include <math.h>
#include <time.h>
#include <stdlib.h>

LAYOUTPLUGIN(SpringElectrical,"Spring_Electrical","David Auber","14/01/2002","Alpha","0","1");

using namespace std;

inline double sqr(double x) {
  return (x*x);
}

//attenuates a force f the force applied on the previous iteration.
inline Coord attenu_force (Coord f, Coord prevF) {
  //no force applied previously, return the force
  if (f.norm () < 1e-3) return Coord (0, 0, 0);
  if (prevF.norm () < 1e-3) return f;
 
  //do the cross product of the forces normalized
  f /= f.norm ();
  prevF /= prevF.norm ();
  double cross = f.getX()*prevF.getY() - f.getY()*prevF.getX();
  if (1 + cross > 1e-3) return Coord ((1 + cross)*f.getX(), 
				      (1 + cross)*f.getY(),
				      0);
  else return Coord (0, 0, 0);
}


SpringElectrical::SpringElectrical(const PropertyContext &context):Layout(context){
}

SpringElectrical::~SpringElectrical() {
}

Coord SpringElectrical::repulsiveForces(node n) {
  node u = n;
  Coord uCoord = prevPos.get(u.id);
  Coord result(0,0,0);
  Iterator<node> *itN = superGraph->getNodes();

  while(itN->hasNext()) {
    node v = itN->next();
    if (u == v) continue;
    Coord uv = prevPos.get(v.id) - uCoord;
    double dist = uv.norm();
    if (dist > 1.0E-3) {
      uv /= dist;
      double length = 10 + (sizeNorm.get(u.id) + sizeNorm.get(v.id));
     
      //dist -= (sizeNorm.get(u.id) + sizeNorm.get(v.id));
      //cout << "repulsive " << sqr (length) << " " << sqr (dist) << endl;
      uv *= -1*sqr (length)/sqr (dist);
    } else {
      uv[0] += 0.01 * (rand() / static_cast <double> (RAND_MAX));
      uv[1] += 0.01 * (rand() / static_cast <double> (RAND_MAX));
    }
    result += uv;
  } delete itN;
  return result;
}

Coord SpringElectrical::attractiveForces(node n) {
  node nu = n;
  Coord u = prevPos.get(nu.id);
  Coord result(0,0,0);
  //spring forces
  Iterator<node> *itN=superGraph->getInOutNodes(n);
  while(itN->hasNext()) {
    node nv = itN->next();
    Coord v = prevPos.get(nv.id);
    Coord uv = v-u;
    double dist = uv.norm();
    if (dist > 1e-3) {
      uv /= dist;
      //dist -= (sizeNorm.get(nu.id) + sizeNorm.get(nv.id));
      double length = 10;
      dist -= sizeNorm.get(nu.id) + sizeNorm.get(nv.id);
      //dist -= (sizeNorm.get(nu.id) + sizeNorm.get(nv.id));
      //cout << "attr " << sqr (length) << " " << sqr (dist) << endl;
      if (dist > 0)
	uv *= sqr (dist)/sqr (length);
      else
	uv *= 1e-3;
    }

    else {
      uv[0] += 0.01 * (rand() / static_cast <double> (RAND_MAX));
      uv[1] += 0.01 * (rand() / static_cast <double> (RAND_MAX));
    }
    uv /= double(superGraph->deg(nu)+1);//+superGraph->deg(nv);
    result += uv;
  } delete itN;
  return result;
}

static const double MAXFORCE = 100.0;
Coord maxForce(Coord force) {
  Coord result(force);
  if (result.norm() > MAXFORCE) {
    result /= result.norm();
    result *= MAXFORCE;
  }
  return result;
}

bool SpringElectrical::run() {

  //for error messages
  string errMsg;

  //get layout proxies
  int iterations = superGraph->numberOfNodes() >? 500;
  inputSelection = superGraph->getProperty<SelectionProxy>("viewSelection");
  inputSize = superGraph->getProperty<SizesProxy>("viewSize");
  LayoutProxy *inputLayout = superGraph->getProperty<LayoutProxy>("viewLayout");
  Iterator<node> *itN = superGraph->getNodes();
  while(itN->hasNext()){
    node itn = itN->next();
    layoutProxy->setNodeValue(itn,inputLayout->getNodeValue(itn));
    prevPos.set(itn.id, inputLayout->getNodeValue(itn));
    prevForce.set (itn.id, Coord (0, 0, 0));
    sizeNorm.set(itn.id,inputSize->getNodeValue(itn).norm()/2.0);
  } delete itN;

  //if the dataset says we need to prefix with circular, do so
  bool doCircular;
  bool hasCircular = dataSet->get ("Circular", doCircular);
  if (hasCircular && doCircular) {
    LayoutProxy circLayout (superGraph);
    bool success = superGraph->computeProperty ("Circular",
						&circLayout, errMsg,
						NULL, dataSet);
    if (!success) {
      cerr << "Circular in Spring_Electrical returned the following error\n:" 
	   << errMsg << endl;
      return false;
    }//end if	

    itN = superGraph->getNodes();
    while (itN->hasNext()) {
      node itn = itN->next();
      layoutProxy->setNodeValue (itn, circLayout.getNodeValue(itn));
    }delete itN;

  }//end if

  k = sqrt(double(superGraph->numberOfNodes()));
  k2 = k;

  for (int count = 0; count < iterations; ++count) {
    if (pluginProgress->progress(count,iterations)!=TLP_CONTINUE) break;
    itN = superGraph->getNodes();
    while(itN->hasNext()){
      node itn = itN->next();
      Coord att = attractiveForces(itn); //maxForce(attractiveForces(itn));
      Coord rep = repulsiveForces(itn); //maxForce(repulsiveForces(itn));
      Coord f = att + rep;
      f = attenu_force (f, prevForce.get (itn.id));
      layoutProxy->setNodeValue(itn, layoutProxy->getNodeValue(itn) + f);
      prevForce.set (itn.id, f);
    } delete itN;
    //set the previous positions to the current postions
    itN = superGraph->getNodes();
    while(itN->hasNext()) {
      node curNode = itN->next();
      prevPos.set (curNode.id, layoutProxy->getNodeValue(curNode));
    }//end while
    delete itN;
  }
  return pluginProgress->state()!=TLP_CANCEL;
}

bool SpringElectrical::check(string &erreurMsg){
  erreurMsg="";
  return true;
}

void SpringElectrical::reset(){}

