/**
 Author: Daniel Archambault
 Email : archam@cs.ubc.ca
 Last modification : 08/12/2004
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by  
 the Free Software Foundation; either version 2 of the License, or     
 (at your option) any later version.
*/
/**
 * implementation of Topological Feature Clustering.  For nodes, history, 
 * author and licence please see the header file TopoClustering.h
 */
#include <time.h>
#include <vector>
#include <tulip/StableIterator.h>
#include <tulip/TlpTools.h>
#include <tulip/ForEach.h>
#include "TopoClustering.h"

//These are the topological feature test functions
//------------------------------------------------------------------
#include <tulip/TreeTest.h>
//------------------------------------------------------------------

//common functions used by the topological feature plugin
//------------------------------------------------------------------
#define TREE_COMPONENT_METRIC ("Connected_Tree Component 2")
#define SINGLETONS_METRIC ("Singletons")
#define CONNECTED_COMPONENT_METRIC ("Connected Component")
#define BICONNECTED_COMPONENT_METRIC ("Bicon Filter")
#define STRENGTH_COMPONENT_METRIC ("Strength Filter")
//------------------------------------------------------------------

//Colours of each topological feature
#define NODE_ALPHA (255)
#define EDGE_ALPHA (5)
#define CONNECTED_COLOUR_NODE (Color (0, 0, 255, NODE_ALPHA))
#define CONNECTED_COLOUR_EDGE (Color (0, 0, 255, EDGE_ALPHA))
#define TREE_COLOUR_NODE (Color (255, 0, 0, NODE_ALPHA))
#define TREE_COLOUR_EDGE (Color (255, 0, 0, EDGE_ALPHA))
#define SINGLETON_COLOUR_NODE (Color (170, 0, 127, NODE_ALPHA))
#define SINGLETON_COLOUR_EDGE (Color (170, 0, 127, EDGE_ALPHA))
#define BICONNECTED_COLOUR_NODE (Color (190, 127, 0, NODE_ALPHA))
#define BICONNECTED_COLOUR_EDGE (Color (190, 127, 0, EDGE_ALPHA))
#define NEARMESH_COLOUR_NODE (Color (50, 0, 75, NODE_ALPHA))
#define NEARMESH_COLOUR_EDGE (Color (50, 0, 75, EDGE_ALPHA))
#define UNKNOWN_COLOUR_NODE (Color (0, 85, 0, NODE_ALPHA))
#define UNKNOWN_COLOUR_EDGE (Color (0, 85, 0, EDGE_ALPHA))
#define CLUSTER_COLOUR_NODE (Color (139, 139, 139, NODE_ALPHA))
#define CLUSTER_COLOUR_EDGE (Color (139, 139, 139, EDGE_ALPHA))

CLUSTERINGPLUGIN(TopoClustering,
		 "Topological Feature Clustering",
		 "Daniel Archambault","08/12/2004","Alpha","0","1");

using namespace std;

//The following function initializes the decomposition operator and
//the decomposition function pointer table
inline void TopoClustering::initDecompOperator() {
  //initialize default decomposition operator
  RunOnce = StringCollection (string (CONNECTED_IDENT) + ";");
  RunRecursively = StringCollection (string (TREE_IDENT) + ";" + 
				     string (BICONNECTED_IDENT) + 
				     ";" +
				     string (EMBEDDER_IDENT) + 
				     ";" +
				     string (STRENGTH_IDENT) + 
				     ";");
  featureFunctionTable[CONNECTED_IDENT] = connected;
  featureFunctionTable[TREE_IDENT] = tree;
  featureFunctionTable[BICONNECTED_IDENT] = biconnected;
  featureFunctionTable[EMBEDDER_IDENT] = embedder;
  featureFunctionTable[STRENGTH_IDENT] = strength;
  featureFunctionTable[SINGLETONS_IDENT] = singletons;
}//end initDecompOperator

TopoClustering::TopoClustering(ClusterContext context):Clustering(context) {
  //initialize timing and counters to zero
  timeTree = 0;
  timeConnected = 0;
  timeBiconnected = 0;
  timeStrength = 0;
  timeEmbedder = 0;
  timeSingletons = 0;
  treeCounter = 0;
  singletonsCounter = 0;
  connectedCounter = 0;
  biconnectedCounter = 0;
  strengthCounter = 0;
  embedderCounter = 0;
  unknownCounter = 0;
  keepColours = false;
  initDecompOperator();
}//end TopoClustering constructor

TopoClustering::~TopoClustering() {}

//a function to generate the originalEdges table at this level of the
//hierarchy
void genOriginalEdges (SuperGraph *graph, vector <edge> &originalEdges) {
  //a proxy for the original edge attribute
  StringProxy *originalEdgesProxy = 
    graph->getProperty <StringProxy> (ORIGINAL_EDGE_NAME);
  
  //for all the edges at this level
  Iterator <edge> *edges = graph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();

    //if there is a null string, there are no original edges
    //(we do not want to apply torques).
    if (originalEdgesProxy->getEdgeValue (curEdge) != "") {
      istringstream idparse (originalEdgesProxy->getEdgeValue (curEdge));
      unsigned int edgeid;
      while (!idparse.eof()) {
	idparse >> edgeid;
	originalEdges.push_back (edge (edgeid));
      }//end while
    }//end if
  } delete edges; //end while
}//end genOriginalEdges

//an inline function to set graph colour
inline void setGraphColour (SuperGraph *graph, 
			    const Color nodeColour, const Color edgeColour) {
  ColorsProxy *colours = 
    graph->getProperty <ColorsProxy> ("viewColor");
  Iterator <node> *nodes = graph->getNodes();
  while (nodes->hasNext())
    colours->setNodeValue (nodes->next(), nodeColour);
  delete nodes;
  Iterator <edge> *edges = graph->getEdges();
  while (edges->hasNext())
    colours->setEdgeValue (edges->next(), edgeColour);
  delete edges;
  vector <edge> originalEdges;
  genOriginalEdges (graph, originalEdges);
  for (unsigned int i = 0; i < originalEdges.size(); i++)
    colours->setEdgeValue (originalEdges[i], edgeColour);
}//end setGraphColour

//A function that claims validity of any subgraph
bool allValid (SuperGraph *g) {
  return true;
}//end allValid

//This function enqueues the working graph if it needs to be recursed on.
inline void TopoClustering::moreRecurseWorking (SuperGraph *workingGraph) {
  string name = workingGraph->getAttribute <string> ("name");
  if ((isType (workingGraph, UNKNOWN_IDENT) || 
       isType (workingGraph, STRENGTH_IDENT)))
    furtherDecompGraphs.push (workingGraph);
}//end moreRecurseWorking

//This function recursively decomposes the sub hierarchies that need
//further decomposition.  It returns true if successful on all components
bool TopoClustering::recurseOnSubHierarchies () {
  //a string for error messages
  string errMsg;

  //a boolean for recursion success
  bool recursionSuccess = true;

  //for all of the graphs that need further decompostion, recurse on them.
  while (!furtherDecompGraphs.empty()) {
    //recurse on this subhierarchy
    SuperGraph *curGraph = furtherDecompGraphs.front();
    bool success = decomposeGraph (curGraph);
    furtherDecompGraphs.pop();

    //if we do not have success, say so
    if (!success) {
      cerr << "Topological Feature Clustering on recursive call returned "
	   << "the following error:\n" << errMsg << endl;
      recursionSuccess = false;
      break;
    }//end if
  }//end for i
  return recursionSuccess;
}//end recurseOnSubHierarchies

//Does topological feature clustering on the passed graph
bool TopoClustering::decomposeGraph (SuperGraph *workingGraph) {
  //name of the graph we are recursing on
  cerr << "recursing on " 
       << workingGraph->getAttribute <string> ("name") << endl;

 //get the number of nodes in the hierarchy at the start of this stage
  //of decomposition.
  unsigned int startNumberOfNodes = workingGraph->numberOfNodes();

  //Try to find topological features and cluster them at this level
  bool success = runOperator (RunRecursively, workingGraph, false);
  if (!success) {
    cerr << "Error encountered in recursive operator" << endl;
    return false;
  }//end if
    
  //if we have made progress see if we have to work on this graph as well
  if (startNumberOfNodes > workingGraph->numberOfNodes()) 
    moreRecurseWorking (workingGraph);
  return true;
}//end decomposeGraph

//a function that add the nodes and edges of the feature into its
//newSubGraph
inline void
copyFeature (SuperGraph *feature, SuperGraph *newSubGraph) {
  StableIterator <node> nodes (feature->getNodes());
  while (nodes.hasNext()) newSubGraph->addNode (nodes.next());
  StableIterator <edge> edges (feature->getEdges());
  while (edges.hasNext()) newSubGraph->addEdge (edges.next());
}//end loadFeature

//a function that deletes the nodes and edges of graph from remainder
inline void
deleteFeatureGraphFromWorking (SuperGraph *feature, SuperGraph *workingGraph) {
  //delete the nodes from the working graph edges are deleted on node deletion
  StableIterator <node> nodes (feature->getNodes());
  while (nodes.hasNext ()) workingGraph->delNode (nodes.next());
}//end deleteFeatureGraphFromWorking

//this function, given a feature graph, sets the connections from
//of the feature node to the nodes of the working graph
inline void 
TopoClustering::setFeatureNodeConnections (SuperGraph *workingGraph,
					   SuperGraph *featureGraph,
					   node featureNode) {
  //a mutable container to indicate when an edge has already been
  //drawn.
  MutableContainer <bool> edgeDone;
  MutableContainer <edge> createdEdge;
  edgeDone.setAll (false);
  
  //The new featureNode is contained in the working graph
  containedInGraph->setNodeValue (featureNode, workingGraph);

  //for all the nodes in the featureGraph
  StableIterator<node> nodes (featureGraph->getNodes());
  while (nodes.hasNext ()) {
    node curNode = nodes.next();
    //This node is now contained in the feature graph
    containedInGraph->setNodeValue (curNode, featureGraph);

    //copy over the edges
    StableIterator<edge> edges (workingGraph->getInOutEdges (curNode));
    while (edges.hasNext()) {
      edge curEdge = edges.next();
      //if the edge is not an element of the feature graph,
      //create a new edge in the working graph with the proper
      //endpoints, delete the old edge from the working graph
      if (!featureGraph->isElement (curEdge)) {
	node opposite = workingGraph->opposite (curEdge, curNode);
	if (!edgeDone.get (opposite.id)) {
	  assert(opposite != curNode);
	  edge created = workingGraph->addEdge (featureNode, opposite);
	  edgeDone.set (opposite.id, true);
	  createdEdge.set (opposite.id, created);
	  originalEdges->setEdgeValue (created, "");
	}//end if
	//append the ids of edges in the original graph that made this
	//edge exist.
	if (originalEdges->getEdgeValue (curEdge) != "") {
	  if (originalEdges
	      ->getEdgeValue (createdEdge.get (opposite.id)) != "")
	    originalEdges->setEdgeValue (createdEdge.get (opposite.id),
					 originalEdges->getEdgeValue 
					 (createdEdge.get (opposite.id)) 
					 + " " +
					 originalEdges->getEdgeValue 
					 (curEdge));
	  
	  else 
	    originalEdges->setEdgeValue (createdEdge.get (opposite.id),
					 originalEdges
					 ->getEdgeValue (curEdge));
	}//end if
	//if we are processing bicon components, copy the bridge proxy
	if ((bridges != NULL) && (bridges->getEdgeValue (curEdge) != "")) 
	  bridges->setEdgeValue (createdEdge.get (opposite.id), 
				 bridges->getEdgeValue (curEdge));
     
	//if this endpoint 
	workingGraph->delEdge (curEdge);
      }//end if
    }//end for i
  }//end while
}//end setFeatureNodeConnections

//This function finds and clusters topological features.
//It takes the working graph in the clusterization step
//and modifies it after the algorithm has done its clustering
//workingGraph -- The SuperGraph object on which clustering is run
//metricClusterAlgo -- The metric clustering algorithm to be run
//checkValidity -- a function that checks feature validity
//featureName -- The name of the feature (appears in subgraph label)
//nextFeatureNumber -- The next available feature number
//childrenFurtherDecomp -- if set to true, we recurse on the children of
//                         the decomposition it will not do the decomp
//                         if there is only one subgraph (infinite recursion)
//allowingSingleNodes -- allows singletons to be considered valid
//The function returns true if it succeeds without error.
#define CLUSTER_LABEL ("c_")
bool
TopoClustering::clusterTopologicalFeature (SuperGraph *workingGraph,
					   string metricClusterAlgo,
					   bool (*checkValidity) (SuperGraph *g),
					   string featureName,
					   unsigned int &featureNumber,
					   const Color featureNodeColour,
					   const Color featureEdgeColour, 
					   bool childrenFurtherDecomp,
					   bool allowSingleNodes,
					   DataSet &localData) {

  //A proxy for colours of the final layout graph
  ColorsProxy *colours = 
    workingGraph->getProperty <ColorsProxy> ("viewColor");

  //string for error messages
  string errMsg;
  SuperGraph *cloneWorkingGraph = superGraph->addSubGraph();
  copyFeature(workingGraph, cloneWorkingGraph);
  //get the viewMetric property context
  MetricProxy topologicalClusters(cloneWorkingGraph);
 
  //try the passed algorithm on the working graph
  bool success;
  success = 
    cloneWorkingGraph->computeProperty (metricClusterAlgo, 
					&topologicalClusters, errMsg,
					NULL, &localData);

  //make sure the plugin ran.
  if (!success) {
    cerr << "The metric plugin " << metricClusterAlgo 
	 << " encountered the following error:\n" << errMsg 
	 << "\n(Is the plugin installed?)" << endl;
    return false;
  }//end if

  //cluster them
  localData.set("Metric", &topologicalClusters);
  success = tlp::clusterizeGraph (cloneWorkingGraph, errMsg, &localData, 
				  "Equal Value");
  if (!success) {
    cerr << "Equal Value clustering returned the following error:\n"
	 << errMsg << "\n(Is the plugin properly installed?)" << endl;
    cerr << "in clusterTopologicalFeature" << endl;
    return false;
  }//end if

  //update working graph so it is the same as cloneWorkingGraph
  StableIterator <node> workingNodes (workingGraph->getNodes());
  while (workingNodes.hasNext())
    workingGraph->delNode (workingNodes.next());
  StableIterator <edge> workingEdges (workingGraph->getEdges());
  while (workingEdges.hasNext())
    workingGraph->delEdge (workingEdges.next());
  copyFeature(cloneWorkingGraph, workingGraph);

  if (bridges != NULL) {
    StableIterator<edge> workingBridgeEdges (workingGraph->getEdges());
    StringProxy *cloneWorkingGraphBridges =
      cloneWorkingGraph->getProperty<StringProxy> (BRIDGE_PROXY);
    while (workingBridgeEdges.hasNext()) {
      edge curBridgeEdge = workingBridgeEdges.next();
      bridges->setEdgeValue (curBridgeEdge, 
			     cloneWorkingGraphBridges->getEdgeValue (curBridgeEdge));
    }//end while
  }//end if

  //count the number of valid subgraphs and singletons
  unsigned int numberOfValidClusterSubGraphs = 0;
  unsigned int numberOfCreatedSubGraphs = 0;
  StableIterator <SuperGraph *> subgs (cloneWorkingGraph->getSubGraphs ());
  while (subgs.hasNext()) {
    SuperGraph *subgraph = subgs.next();
    if (isType (subgraph, CLUSTER_LABEL)) {
      numberOfCreatedSubGraphs++;
      if (((subgraph->numberOfNodes () > 1) || allowSingleNodes) &&
	  checkValidity (subgraph))
	numberOfValidClusterSubGraphs++;
    }//end if
  }//end while

  //if we are futher decomposing the children, everything is that topology
  //type until shown otherwise.
  if (childrenFurtherDecomp && (numberOfValidClusterSubGraphs > 1))
    if (!keepColours)
      setGraphColour (workingGraph, 
		      featureNodeColour, featureEdgeColour);

  //go through all the possible features, pruning out the non-features
  //and flagging them as remainders
  cerr << "Creating Hierarchy " << endl;
  StableIterator <SuperGraph *> possibleFeatures (cloneWorkingGraph->getSubGraphs ());
  while (possibleFeatures.hasNext () && 
	 ((!childrenFurtherDecomp && (numberOfCreatedSubGraphs > 1))|| 
	  (childrenFurtherDecomp && (numberOfValidClusterSubGraphs > 1)))) {
    SuperGraph *curGraph = possibleFeatures.next();
    cerr << "." << flush;

    //make sure this was created by equal value clustering
    if (!isType (curGraph, CLUSTER_LABEL)) continue;
    
    //if this is not singleton, and it is a valid 
    //type of the feature we are looking for, we cluster it.
    if ((curGraph->numberOfNodes () > 1) && checkValidity (curGraph)) {

      //create a new subgraph on the supergraph
      SuperGraph *newSubGraph = superGraph->addSubGraph ();

      //change the new subgraph name to the feature name
      char charFeatureNumber [10];
      if (!childrenFurtherDecomp) {
	sprintf (charFeatureNumber, "%d", featureNumber);
	newSubGraph->setAttribute <string> ("name", 
					    featureName + charFeatureNumber);
	featureNumber++;
      }//end if
      else {
	sprintf (charFeatureNumber, "%d", unknownCounter);
	newSubGraph->setAttribute <string> ("name", 
					    static_cast <string> (UNKNOWN_IDENT) + 
					    charFeatureNumber);
	unknownCounter++;
      }//end else

      //copy the feature into the new sub graph
      copyFeature (curGraph, newSubGraph);
      
      //colour the subgraph if required
      if (!keepColours) {
	if (!childrenFurtherDecomp) 
	  setGraphColour (newSubGraph, featureNodeColour, featureEdgeColour);
	else 
	  setGraphColour (newSubGraph, UNKNOWN_COLOUR_NODE,
			  UNKNOWN_COLOUR_EDGE);
      }//end if

      //create a cluster node in the working graph and set connections
      node featureNode = workingGraph->addNode();
      featureNodes->setNodeValue (featureNode, newSubGraph);
      setFeatureNodeConnections (workingGraph, newSubGraph, featureNode);
      //set the parent node of the newSubGraph to the id of the featureNode
      newSubGraph->setAttribute <unsigned int> (PARENT_NODE_ATTRIBUTE, 
						featureNode.id);

      //delete the feature graph from the working graph use the
      //backup for deletion as we don't want to be modifying
      //both graphs.
      deleteFeatureGraphFromWorking (newSubGraph, workingGraph);

      //if the components require further decomp, push them back on
      //the decomp list
      if (childrenFurtherDecomp) 
	furtherDecompGraphs.push (newSubGraph);
    }//end if
  } //end while
  cerr << endl << "done" << endl;

  //delete the clone working graph
  superGraph->delAllSubGraphs(cloneWorkingGraph);
  
  //if the children are unknown -- not the current component -- and
  //we have detected the topological feature
  //set the current component name and increment the feature number
  if (childrenFurtherDecomp && (numberOfValidClusterSubGraphs > 1)) {
    char charFeatureNumber [10];
    sprintf (charFeatureNumber, "%d", featureNumber);
    workingGraph->setAttribute <string> ("name",  
					 featureName + charFeatureNumber);
    featureNumber++;
  }//end if

  //if the valid number of subgraphs is equal to one and there is no
  //further decomp.  Relable the graph name appropriately
  if (!childrenFurtherDecomp && (numberOfValidClusterSubGraphs == 1) &&
      (numberOfCreatedSubGraphs == 1)) {
    char charFeatureNumber [10];
    sprintf (charFeatureNumber, "%d", featureNumber);
    workingGraph->setAttribute <string> ("name", 
					 featureName + charFeatureNumber);
    if (!keepColours) 
      setGraphColour (workingGraph, 
		      featureNodeColour, featureEdgeColour);
    featureNumber++;
  }//end if
  return true;
}//end clusterTopologicalFeature

//This function stops the clock timer, reports the time, and
//add the time to the appropriate timer.
inline void stopClock (string type, clock_t start, 
		       double &cumulativeTime) {
  clock_t end = clock();
  double elapsedTime = (double) (end - start) / (double) (CLOCKS_PER_SEC);
  cout << "Time spent on " << type << " " << elapsedTime << endl;
  cumulativeTime += elapsedTime;
}//end stopClock

//A function to detect connected components.  Returns
//true if multiple connected components were found.
bool connected (SuperGraph *workingGraph, bool &featuresFound,
		TopoClustering *topoClust) {
  cerr << "Starting Connected Component Detection" << endl;
  clock_t connectedStart =  clock();

  //find and make Connected Component clusters
  //The subgraphs are set to "Unknown" as they require
  //further decomposition.
  DataSet localData;
  bool success = 
    topoClust->clusterTopologicalFeature (workingGraph,
					  CONNECTED_COMPONENT_METRIC,
					  allValid,
					  CONNECTED_IDENT,
					  topoClust->connectedCounter,
					  CONNECTED_COLOUR_NODE,
					  CONNECTED_COLOUR_EDGE,
					  true, true, localData);
  if (!success) {
    cerr << "Error encountered while trying to cluster connected components "
	 << endl;
    return false;
  }//end if

  //stop the clock and and record the time
  stopClock ("connected components", connectedStart, topoClust->timeConnected);
 
  //If the graph's name has changed to connected, we have found 
  //at least a pair of connected components.  Return
  if (isType (workingGraph, CONNECTED_IDENT)) featuresFound = true;
  return true;
}//end connectedComponents

//A function to detect trees.  Returns true if multiple trees were found.
bool tree (SuperGraph *workingGraph, bool &featuresFound,
	   TopoClustering *topoClust) {
    cerr << "Starting Tree Detection" << endl;
  clock_t treeStart = clock();
  unsigned int startNodes = workingGraph->numberOfNodes();
  DataSet treeData;
  treeData.set ("Non-Tree Roots", true);
  //find and make tree clusters
  bool success = 
    topoClust->clusterTopologicalFeature (workingGraph,
					  TREE_COMPONENT_METRIC,
					  TreeTest::isTree,
					  TREE_IDENT,
					  topoClust->treeCounter,
					  TREE_COLOUR_NODE,
					  TREE_COLOUR_EDGE,
					  false, false, treeData);
  if (!success) {
    cerr << "Error encountered while trying to cluster trees." << endl;
    return false;
  }//end if

  //stop the clock and and record the time
  stopClock ("trees", treeStart, topoClust->timeTree);
  if ((startNodes > workingGraph->numberOfNodes()) ||
      isType (workingGraph, TREE_IDENT)) featuresFound = true;
  return true;
}//end trees

//A function to detect singletons.  Returns true if singletons were found.
bool singletons (SuperGraph *workingGraph, bool &featuresFound,
		 TopoClustering *topoClust) {
  cerr << "Starting Singleton Detection" << endl;
  clock_t singletonStart = clock();
  unsigned int startNodes = workingGraph->numberOfNodes();
  DataSet localData;
  bool success = 
    topoClust->clusterTopologicalFeature (workingGraph,
					  SINGLETONS_METRIC,
					  TreeTest::isTree,
					  SINGLETONS_IDENT,
					  topoClust->singletonsCounter,
					  SINGLETON_COLOUR_NODE,
					  SINGLETON_COLOUR_EDGE,
					  false, false, localData);
  if (!success) {
    cerr << "Error encountered while trying to cluster singletons." << endl;
    return false;
  }//end if

  //stop the clock and and record the time
  stopClock ("singletons", singletonStart, topoClust->timeSingletons);
  if ((startNodes > workingGraph->numberOfNodes()) ||
      isType (workingGraph, SINGLETONS_IDENT)) featuresFound = true;
  return true;
}//end singletons

//A function to detect biconnected components.  Returns
//true if multiple biconnected components were found.
bool biconnected (SuperGraph *workingGraph, bool &featuresFound,
		  TopoClustering *topoClust) {
  string errMsg; //for error messages
  cerr << "Starting Biconnected Component Detection" << endl;
  clock_t biconStart = clock();
  topoClust->bridges = 
    workingGraph->getLocalProperty<StringProxy> (BRIDGE_PROXY);

  //find and make Biconnected Component clusters
  //The subgraphs are set to "Unknown" as they require
  //further decomposition.
  DataSet localData;
  bool success = 
    topoClust->clusterTopologicalFeature (workingGraph,
					  BICONNECTED_COMPONENT_METRIC,
					  allValid,
					  BICONNECTED_IDENT,
					  topoClust->biconnectedCounter,
					  BICONNECTED_COLOUR_NODE,
					  BICONNECTED_COLOUR_EDGE,
					  true, true, localData);
  if (!success) {
    cerr << "Error encountered while trying to cluster biconnected components "
	 << endl;
    return false;
  }//end if
  
  //If the graph's name has changed to biconnected, we have found 
  //at least a pair of components.  Set root of tree to max degree node
  if (isType (workingGraph, BICONNECTED_IDENT)) {
    node curNode, maxDegNode;
    unsigned int maxDeg = 0;
    forEach (curNode, workingGraph->getNodes()) {
      if (workingGraph->deg (curNode) > maxDeg) {
	maxDeg = workingGraph->deg (curNode);
	maxDegNode = curNode;
      }//end if
    }//end forEach
    TreeTest::makeRootedTree (workingGraph, maxDegNode);
    featuresFound = true;
  }//end if
  else workingGraph->delLocalProperty (BRIDGE_PROXY);

  topoClust->bridges = NULL;

  //stop the clock and and record the time
  stopClock ("biconected components", biconStart, topoClust->timeBiconnected);
  return true;
}//end biconnected

#define CLOSE_ENOUGH_VAR (0.05)
#define MAX_PERCENT_VAR (0.60)
#define MIN_PERCENT_VAR (0.10)
inline bool goodVariance (vector<double> &lambdas,
			  vector<vector<double> > &evecs) {
  //if the percent variance in the two axes is about the same, 
  //layout is good
  if (lambdas[0] - lambdas[1] < CLOSE_ENOUGH_VAR) return true; 
  
  //could be good in three dimensions
  double sumFirstThree = lambdas[0] + lambdas[1] + lambdas[2];
  return (sumFirstThree > (1 - CLOSE_ENOUGH_VAR)) && 
    (lambdas[2] > MIN_PERCENT_VAR) && 
    (lambdas[2] < MIN_PERCENT_VAR + CLOSE_ENOUGH_VAR) && 
    (lambdas[0] < MAX_PERCENT_VAR);
}//end goodVariance

bool embedder (SuperGraph *workingGraph, bool &featuresFound,
	       TopoClustering *topoClust) {
  string errMsg; //for error messages
  cerr << "Checking for Embedder likeness" << endl;
  clock_t embedderStart = clock();
  
  LayoutProxy localEmbedder (workingGraph);
  DataSet localDataSet;
  bool success = 
    workingGraph->computeProperty ("High Dimensional Embedder",
				   &localEmbedder, errMsg, NULL, 
				   &localDataSet);
  if (!success) {
    cerr << "Error encountered while trying to call High Dimensional "
	 << "Embedder " << endl << errMsg << endl;
    return false;
  }//end if
  
  vector<double> lambdas;
  vector<vector<double> > evecs;
  if (!localDataSet.get ("lambdas", lambdas)) {
    cerr << "Error:  eigenvalues not returned." << endl;
    return false;
  }//end if
  if (!localDataSet.get ("evecs", evecs)) {
    cerr << "Error:  eigenvectors not returned." << endl;
    return false;
  }//end if
  
  //if we have good variance, use HDE
  if (goodVariance (lambdas, evecs)) {
      ColorsProxy *colours = 
	workingGraph->getProperty <ColorsProxy> ("viewColor");
      if (!topoClust->keepColours)
	setGraphColour (workingGraph, NEARMESH_COLOUR_NODE, 
			NEARMESH_COLOUR_EDGE);
      
      char charFeatureNumber [10];
      sprintf (charFeatureNumber, "%d", topoClust->embedderCounter);
      workingGraph->setAttribute <string> ("name", 
					   static_cast <string> (EMBEDDER_IDENT) + 
					   charFeatureNumber);
      ++(topoClust->embedderCounter);
      //stop the clock and and record the time
      stopClock ("near-mesh components", embedderStart, 
		 topoClust->timeEmbedder);
      featuresFound = true;
  }//end if
  //stop the clock and and record the time
  stopClock ("near-mesh components", embedderStart, topoClust->timeEmbedder); 
  return true;
}//end embedder

//A function to detect if there are strength clusters.  Returns
//true if there are.
bool strength (SuperGraph *workingGraph, bool &featuresFound,
	       TopoClustering *topoClust) {
  if (!highNodeEdgeRatio (workingGraph->numberOfNodes(),
			  workingGraph->numberOfEdges(),
			  topoClust->nodeEdgeThresh)) {
    cerr << "Starting Strength Detection" << endl;
    clock_t strengthStart =  clock();
    
    //Decompose the rest using a strength metric
    DataSet localData;
    bool success = 
      topoClust->clusterTopologicalFeature (workingGraph,
					    STRENGTH_COMPONENT_METRIC,
					    allValid,
					    STRENGTH_IDENT,
					    topoClust->strengthCounter,
					    CLUSTER_COLOUR_NODE,
					    CLUSTER_COLOUR_EDGE,
					    true, false, localData);

    if (!success) {
      cerr << "Error encountered while trying to execute strength clustering "
	   << endl;
      return false;
     }//end if
    
    //stop the clock and and record the time
    stopClock ("strength components", strengthStart, topoClust->timeStrength);
    if (isType (workingGraph, STRENGTH_IDENT)) featuresFound = true;
  }//end if
  return true;
}//end strength

//The following function runs a TopoClustering operator on the workingGraph
//represented by the passed StringCollection.  It returns false if no
//algorithm of the operator finds anything.  When runAll is true, it forces
//all algorithms to be run.  When it is false, the operator terminates on
//the first successful application of an algorithm in the operator.
bool TopoClustering::runOperator (StringCollection &decompOperator,
				  SuperGraph *workingGraph,
				  bool runAll) {

  bool featuresFound = false; //did the algorithm find features?
  bool success = true; //did the algorithm succeed without error?
  StringCollection::iterator decompOpIt = decompOperator.begin();
  for (; decompOpIt != decompOperator.end() && 
	 (runAll || !featuresFound); ++decompOpIt) {
    const char *detectionAlgo = decompOpIt->c_str();
    if (!featureFunctionTable[detectionAlgo]) {
      cerr << "The function to decompose " << detectionAlgo 
	   << " features was not found in the feature function table." << endl;
      return false;
    }//end if
    else {
      success = 
	featureFunctionTable[detectionAlgo] (workingGraph,
					     featuresFound, this);
      if (!success) {
	cerr << "Error while detecting " << detectionAlgo << endl;
	return false;
      }//end if
    }//end else
  }//end for
  return success;
}//end runOperator

//The following function updates the hierarcy pannel under finalLayoutGraph
//workingG -- the top of the working graph hierarchy.
//finalLayoutG -- the top of the final layout hierarchy
void TopoClustering::updateFinalLayoutPanel(SuperGraph *workingG,
					    SuperGraph *finalLayoutG) {
  //make a new subgraph under the final layout graph
  SuperGraph *newFinalLayoutG = finalLayoutG->addSubGraph();

  //Traverse the hierarchy adding original nodes of the graph when
  //we see them
  double maxOffset = 0;
  StableIterator <node> nodes (workingG->getNodes());
  while (nodes.hasNext()) {
    node curNode = nodes.next();
    SuperGraph *curGraph = featureNodes->getNodeValue (curNode);

    //if we have a standard node, add it to new final layout graph
    //as long as it is not an artificial node
    if (curGraph == NULL) {
      if (finalLayoutGraph->isElement (curNode)) {	
	newFinalLayoutG->addNode (curNode);
	leafid->setNodeValue (curNode, newFinalLayoutG->getId());
      }//end if
    }//end if
    else 
      updateFinalLayoutPanel (curGraph, newFinalLayoutG);
  }//end while

  //Now that we have all the nodes at this level of the hierarchy
  //add the edges
  StableIterator <node> nodeEdges (newFinalLayoutG->getNodes());
  while (nodeEdges.hasNext()) {
    node curNode = nodeEdges.next();
    StableIterator <edge> edges (finalLayoutGraph->getInOutEdges (curNode));
    while (edges.hasNext()) {
      edge curEdge = edges.next();
      node opposite = finalLayoutGraph->opposite (curEdge, curNode);
      if (newFinalLayoutG->isElement (opposite))
	newFinalLayoutG->addEdge (curEdge);
    }//end while
  }//end while

  //set the name of the new final layout graph
  char cstrNodesAndEdges [40];
  sprintf (cstrNodesAndEdges, ": n %d e %d (Hier: n %d e %d)", 
	   newFinalLayoutG->numberOfNodes(), newFinalLayoutG->numberOfEdges(),
	   workingG->numberOfNodes(), workingG->numberOfEdges());
  string name = workingG->getAttribute <string> ("name");
  string finalName = name + cstrNodesAndEdges;
  newFinalLayoutG->setAttribute <string> ("name", finalName);
  workingG->setAttribute <string> ("name", finalName);
}//end updateHierarchyPannel

//function that reports the final times of the components in the algorithm
//start -- the time the plugin began executing
inline void TopoClustering::reportTimes (clock_t start) {
  clock_t end = clock();
  cout << "-------------------------------------------------------" << endl;
  cout << "Total execution time " 
       << ((double) (end - start) / (double) CLOCKS_PER_SEC) << endl;
  cout << "-------------------------------------------------------" << endl;
  cout << "Trees " << timeTree << endl;
  cout << "Biconnected Components " << timeBiconnected << endl;
  cout << "Connected Components " << timeConnected << endl;
  cout << "Strength Components " << timeStrength << endl;
  cout << "Embedder Components "<< timeEmbedder << endl;
  cout << "-------------------------------------------------------" << endl;
}//end reportTimes

//Runs topological fearue clustering
bool TopoClustering::run() {
  //begin timing the decomposition
  clock_t startTime =  clock();

  //take out a meta graph proxy for the cluster name
  featureNodes = superGraph->getProperty<MetaGraphProxy> (FEATURE_NODE_NAME);

   //take out a metric proxy for the leaf level id
  leafid = superGraph->getProperty<MetricProxy> (LEAF_ID_NAME);

  //take out a string proxy for the original edge ids
  originalEdges = superGraph->getProperty <StringProxy> (ORIGINAL_EDGE_NAME); 

  //A meta graph proxy for the graph that contains this node
  containedInGraph = 
    superGraph->getProperty <MetaGraphProxy> (CONTAINED_IN_GRAPH);

  //the proxy for bridges will be NULL unless we are processing bicon comps.
  bridges = NULL;
  
  if (dataSet != NULL) {
    //check to see if there is a near-complete threshold.
    if (!dataSet->get<double> ("nodeEdgeThresh", nodeEdgeThresh))
      nodeEdgeThresh = 1.0;
    
    //check to see if decomposition operators have been passed into the graph.
    string runOnceString, runRecursivelyString;
    if (dataSet->get<string> ("TopoLayout Operator Once", runOnceString))
      RunOnce = StringCollection (runOnceString);
    if (dataSet->get<string> ("TopoLayout Operator Recursive", 
			      runRecursivelyString))
      RunRecursively = StringCollection (runRecursivelyString);
    dataSet->get<bool> (DATASET_KEEPCOLOURS, keepColours);
  }//end not null dataset

  //check to see if there is an untouched graph, if there is not one,
  //make one and put it in dataset (if one does not exist, we should be at 
  //root, otherwise bug).  TopoClustering assumes no untouched graph
  //exists.  It will not find the graph even if you supply it.
  bool existUntouched = dataSet->get<SuperGraph *>("untouched", untouched);
  if (!existUntouched) {
    untouched = tlp::newCloneSubGraph (superGraph);
    untouched->setAttribute <string> ("name", UNTOUCHED_SUBGRAPH_NAME);
    ColorsProxy *untouchedColours = 
      untouched->getLocalProperty<ColorsProxy> ("viewColor");
    ColorsProxy *colours = 
      superGraph->getProperty<ColorsProxy> ("viewColor");
    Iterator <node> *nodes = superGraph->getNodes();
    while (nodes->hasNext()) {
      node curNode = nodes->next();
      untouchedColours->setNodeValue (curNode,
				      colours->getNodeValue (curNode));
    } delete nodes;
    Iterator<edge> *edges = superGraph->getEdges();
    while (edges->hasNext()) {
      edge curEdge = edges->next();
      untouchedColours->setEdgeValue (curEdge,
				      colours->getEdgeValue (curEdge));
    } delete edges;
    dataSet->set <SuperGraph *>("untouched", untouched);
  }//end if

  //check to see if there is a final layout graph, if there is not one,
  //make one and put it in dataset (if one does not exist, we should be at 
  //root, otherwise bug).  TopoClustering assumes no final layout graph
  //exists.  It will not find the graph, even if you supply it.
  bool existFinalLayout =
    dataSet->get<SuperGraph *>("finalLayoutGraph", finalLayoutGraph);
  if (!existFinalLayout) {
    //make a new one
    finalLayoutGraph = tlp::newCloneSubGraph (superGraph);
    finalLayoutGraph
      ->setAttribute <string> ("name", FINAL_LAYOUT_SUBGRAPH_NAME);

    //set the location of all nodes of the final layout graph to
    //zero to begin
    LayoutProxy *finalLayout = 
      finalLayoutGraph->getLocalProperty <LayoutProxy> ("viewLayout");
    finalLayout->setAllNodeValue (Coord (0, 0, 0));

    //make sure final layout does not not have meta graph attributes
    finalLayoutGraph->getLocalProperty <MetaGraphProxy> ("viewMetaGraph");

    //A proxy for colours of the final layout graph
    ColorsProxy *colours = 
      finalLayoutGraph->getProperty <ColorsProxy> ("viewColor");

    //Initially all colours should be unknown if not kept
    if (!keepColours) {
      colours->setAllNodeValue (UNKNOWN_COLOUR_NODE);
      colours->setAllEdgeValue (UNKNOWN_COLOUR_EDGE);
    }//end if

    //put the pointer to it in dataSet
    dataSet->set <SuperGraph *>("finalLayoutGraph", finalLayoutGraph);

    //set all the orignal edges of the finalLayoutGraph to their own id
    Iterator <edge> *edges = finalLayoutGraph->getEdges();
    char charID [10]; //the id of the edge as a string
    while (edges->hasNext()) {
      edge curEdge = edges->next();
      sprintf (charID, "%d", curEdge.id);
      originalEdges->setEdgeValue (curEdge, charID);
    } delete edges; //end while
  }//end if

  //check to see if there is a hierarchy, if there is not one,
  //make one and put it in dataset (if one does not exist, we should be at 
  //root, otherwise bug).  TopoClustering assumes no hierarchy
  //exists.  It will not find the graph even if you supply it.
  bool existHierarchy = 
    dataSet->get<SuperGraph *>("hierarchy", hierarchy);
  if (!existHierarchy) {
    hierarchy = tlp::newCloneSubGraph (superGraph);
    hierarchy->setAttribute <string> ("name", HIERARCHY_SUBGRAPH_NAME);
    hierarchy = tlp::newCloneSubGraph (hierarchy);
    hierarchy->setAttribute <string> ("name", UNKNOWN_IDENT);
    dataSet->set <SuperGraph *>("hierarchy", hierarchy);
    //set the working graph parent attribute to NO_PARENT
    hierarchy->setAttribute <unsigned int> (PARENT_NODE_ATTRIBUTE, 
					    NO_PARENT);
  }//end if

  //set the final layout node sizes to the same sizes as untouched
  SizesProxy *finalSizes = 
    finalLayoutGraph->getLocalProperty <SizesProxy> ("viewSize");
  SizesProxy *untouchedSizes = 
    untouched->getProperty <SizesProxy> ("viewSize");
  Iterator <node> *nodes = superGraph->getNodes();
  while (nodes->hasNext()) {
    node curNode = nodes->next();
    finalSizes->setNodeValue (curNode, 
			      untouchedSizes->getNodeValue (curNode));
  } delete nodes; //end while
  
  //set all the containedInGraph values of the nodes in this
  //graph to workingGraph
  Iterator <node> *containedInHierarchy = hierarchy->getNodes();
  while (containedInHierarchy->hasNext())
    containedInGraph->setNodeValue (containedInHierarchy->next(), hierarchy);
  delete containedInHierarchy;

  //construct the hierarchy by applying runOnce and the recursive operator
  if (RunOnce.size() > 0) {
    bool success = runOperator (RunOnce, hierarchy, true);
    if (!success) {
      cerr << "Error encountered while running initial pass operator " 
	   << "on the hierarchy." << endl;
      return false;
    }//end if
  }//end if
  moreRecurseWorking (hierarchy);

  if (RunRecursively.size() > 0) {
    cout << RunRecursively.size() << endl;
    cout << "in recurse on queue " << endl;
    bool success = recurseOnSubHierarchies ();
    if (!success) {
      cerr << "Error encountered while recursing on hierarchy." << endl;
      return false;
    }//end if
  }//end if
  
  //update the hierarchy pannel under final layout
  updateFinalLayoutPanel (hierarchy, finalLayoutGraph);

  //if we are at the end report the final times
  reportTimes (startTime);
  
  return true;
}//end run

bool TopoClustering::check(string &erreurMsg) {
  erreurMsg="";
  return true;
}

void TopoClustering::reset() {
}
