//-*-c++-*-
#ifndef _TOPO_CLUSTERING_H
#define _TOPO_CLUSTERING_H

/// TopoClustering.h
/** 
 *
 *  This cluster performes topological feature clustering.  It looks 
 *  for certain topological features in the graph and clusters them
 *  recursively into a hierarchy.
 *
 *  HISTORY:
 *
 *  The code was first implemented in Nov/Dec 2004 by Daniel Archambault.
 *
 *  NOTES:
 *
 *  The algorithm will place a pointer to the SuperGraph representing
 *  the created hierarchy in the passed dataSet under "hierarchy."
 *  It will also create a finalLayoutGraph with all nodes set to 
 *  (0, 0, 0) and and untouched copy of the input graph, placing pointers
 *  to them in "finalLayoutGraph" and "untouched" respectively.
 *
 *  The plugin does not check the validity of "hierarchy," 
 *  "untouched," and "finalLayoutGraph."  Please ensure they are not
 *  defined in the passed dataSet before you call the plugin.  If you
 *  do not the plugin will fail badly.  Also, please ensure that dataSet
 *  does not have a variable "TopologyCounters" defined or variables
 *  defined with any of the variable names in the topological feature
 *  counters section below.  Subgraph labeling may be affected if you do.
 * 
 *  AUTHOR:
 *
 *  Daniel Archambault, University of British Columbia, CA: 
 *  Email: archam@cs.ubc.ca
 *  Version 0.1: 09 December 2004: Initial version
 *
 *  LICENCE:
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by  
 *  the Free Software Foundation; either version 2 of the License, or     
 *  (at your option) any later version.
 *
 */

#include <string>
#include <cstring>
#include <queue>
#include <tulip/TulipPlugin.h>
#include <tulip/MetaGraphProxy.h>
#include <tulip/ColorsProxy.h>
#include "HierarchyMap.h"
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif

/** \addtogroup clustering */
/*@{*/
class TopoClustering : public Clustering { 
public:
  TopoClustering(ClusterContext);
  ~TopoClustering();

  //A variable to store the node/edge threshold
  double nodeEdgeThresh;

  //counters for all the topological features
  //----------------------------------------------------------------------
  unsigned int treeCounter; 
  unsigned int connectedCounter; 
  unsigned int biconnectedCounter;
  unsigned int strengthCounter;
  unsigned int embedderCounter;
  unsigned int singletonsCounter;
  unsigned int unknownCounter; //for unidentified components
  //----------------------------------------------------------------------
  
  //The total amount of time spent on a particular operation in microsecs.
  //----------------------------------------------------------------------
  double timeTree;
  double timeConnected;
  double timeBiconnected;
  double timeStrength;
  double timeEmbedder;
  double timeSingletons;
  //----------------------------------------------------------------------

  //A string proxy to indicate the nodes inside biconnected components
  //which act like bridges.
  StringProxy *bridges;

  //a dataset parameter to define if we colour by topological feature
  bool keepColours;

  //the hybrid graph on which we are working is workingGraph
  //This function finds and clusters topological features.
  //It takes the working graph in the clusterization step
  //and modifies it after the algorithm has done its clustering
  //workingGraph -- The SuperGraph object on which clustering is run
  //metricClusterAlgo -- The metric clustering algorithm to be run
  //checkValidity -- a function that checks feature validity
  //featureName -- The name of the feature (appears in subgraph label)
  //nextFeatureNumber -- The next available feature number
  //childrenFurtherDecomp -- if set to true, we recurse on the children of
  //                         the decomposition it will not do the decomp
  //                         if there is only one subgraph (infinite recursion)
  //allowingSingleNodes -- allows singletons to be considered valid
  //it returns true if it completes without error
  bool clusterTopologicalFeature (SuperGraph *workingGraph,
				  std::string metricClusterAlgo,
				  bool (*checkValidity) (SuperGraph *g),
				  std::string featureName,
				  unsigned int &featureNumber,
				  const Color featureNodeColour,
				  const Color featureEdgeColour, 
				  bool childrenFurtherDecomp,
				  bool allowSingleNodes,
				  DataSet &localData);
  
  bool run();
  bool check(std::string &);
  void reset();

private:
  //proxy that will esentially hold the graph hierarchy
  MetaGraphProxy *featureNodes;

  //a metric proxy that holds the leaf level node id
  MetricProxy *leafid;

  //proxy that holds the ids of the edges in the original graph
  //that gave rise to this edge
  StringProxy *originalEdges;

  //A meta graph proxy for the graph that contains this node
  MetaGraphProxy *containedInGraph;

  //the untouched graph
  SuperGraph *untouched;

  //the final layout graph (stored for HierLayout and not used in this
  //plugin)
  SuperGraph *finalLayoutGraph;

  //The hierarchy
  SuperGraph *hierarchy;

  //a vector of hierarchy components that need work.
  std::queue <SuperGraph *> furtherDecompGraphs;

  //The topoclustering decomposition operator and the hash table
  //which contains function pointers to each decomposition operator
  //----------------------------------------------------------------------
  StringCollection RunOnce; 
  StringCollection RunRecursively;
  struct equstr {
    bool operator() (const char *s1, const char *s2) const {
      return std::strcmp (s1, s2) == 0;
    }//end operator()
  };
  __gnu_cxx::hash_map<const char *, 
		      bool (*) (SuperGraph *workingGraph, 
				bool &featuresFound,
				TopoClustering *topoClust),
		      __gnu_cxx::hash<const char *>, equstr>
  featureFunctionTable;
  //----------------------------------------------------------------------

  //A function to intialize the feature function table and the default
  //decomposition operator
  inline void initDecompOperator();

  //This function recursively decomposes the sub hierarchies that need
  //further decomposition.  It returns true if successful on all
  //sub hierarchies
  bool recurseOnSubHierarchies ();

  //this function, given a feature graph, sets the connections from
  //of the feature node to the nodes of the working graph
  inline void 
  setFeatureNodeConnections (SuperGraph *workingGraph,
			     SuperGraph *featureGraph,
			     node featureNode);

  //Finds multiple connected components in working graph and returns 
  //true if found 
  bool connectedComponents (SuperGraph *workingGraph);


  //Executes topological feature clustering on the passed graph
  bool decomposeGraph (SuperGraph *workingGraph);

  //This function enqueues the working graph if it needs to be recursed on.
  //startNumberOfNodes is the number of nodes in the subgraph before the
  //decomposition was applied.
  inline void moreRecurseWorking (SuperGraph *workingGraph);

  //The following function runs a TopoClustering operator on the workingGraph
  //represented by the passed StringCollection.  It returns false if no
  //algorithm of the operator finds anything.  When runAll is true, it forces
  //all algorithms to be run.  When it is false, the operator terminates on
  //the first successful application of an algorithm in the operator.
  bool runOperator (StringCollection &decompOperator, SuperGraph *workingGraph,
		    bool runAll);

  //This function updates the hierarchy under the finalLayoutGraph pannel
  //after the hierarchy has been constructed under workingGraph.
  //workingG -- a pointer to the constructed hierarchy
  //finalLayoutG -- a pointer to the finalLayoutGraph
  void updateFinalLayoutPanel (SuperGraph *workingG, 
			       SuperGraph *finalLayoutG);

  //function that reports the final times of the components in the algorithm
  //start -- the time the plugin started executing
  inline void reportTimes (clock_t start);

};

//these functions perform topological feature clustering.  It should
//be evident what feature is sought from the function name.  
//WorkingGraph is the graph in which the features are found and
//FeaturesFound is set to true if more than one feature was found.
//returns true if the algorithm completes without error
//-------------------------------------------------------------------
bool connected (SuperGraph *workingGraph, bool &featuresFound,
		TopoClustering *topoClust);
bool tree (SuperGraph *workingGraph, bool &featuresFound,
	   TopoClustering *topoClust);
bool singletons (SuperGraph *workingGraph, bool &featuresFound,
		 TopoClustering *topoClust);
bool biconnected (SuperGraph *workingGraph, bool &featuresFound,
		  TopoClustering *topoClust);
bool embedder (SuperGraph *workingGraph, bool &featuresFound,
	       TopoClustering *topoClust);
bool strength (SuperGraph *workingGraph, bool &featuresFound,
	       TopoClustering *topoClust);
//-------------------------------------------------------------------
/*@}*/
#endif
