#include <deque>
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include "tulip/TlpTools.h"
#include "tulip/SuperGraph.h"
#include "tulip/MetricProxy.h"
#include "tulip/SortIterator.h"
#include "tulip/ForEach.h"
#include "tulip/MutableContainer.h"
#include "tulip/GraphMeasure.h"
#include "tulip/KruskalMST.h"

using namespace std;
using namespace __gnu_cxx;

void bfsSpanningTree (SuperGraph *graph, SuperGraph *tree, 
		      deque<node> &bfsDeque,
		      MutableContainer<bool> &visited) {
  while (!bfsDeque.empty()) {
    node visitNode = bfsDeque.front();
    bfsDeque.pop_front();
    edge curEdge;
    forEach (curEdge, graph->getInOutEdges (visitNode)) {
      node opposite = graph->opposite (curEdge, visitNode);
      if (!visited.get (opposite.id)) {
	visited.set (opposite.id, true);
	tree->addEdge (curEdge);
	bfsDeque.push_back (opposite);
      }//end if
    }//if not visited
  }//end while
}//end bfsSpanningTree

MetricProxy *getWeights (SuperGraph *graph) {
  MetricProxy *weights = NULL;
  Iterator<string> *properties = graph->getProperties();
  while (properties->hasNext()) {
    string curProperty = properties->next();
    if (curProperty == WEIGHTS_PROXY_NAME) {
      weights = graph->getProperty<MetricProxy> (WEIGHTS_PROXY_NAME);
      delete properties;
      return weights;
    }//end if
  } delete properties;
  return NULL;
}//end doesHaveWeights

//Merges the components source and target are a part of.  The
//smallest component is merged onto the largest.
void mergeComps (SuperGraph *tree, node source, node target,
		 hash_map<node, unsigned int> &component,
		 hash_map<node, deque<node> *> &subTrees) {
  deque<node> *maxComp = subTrees[source];
  deque<node> *minComp = subTrees[target];
  node maxNode = source;
  if (maxComp->size() < minComp->size()) {
    deque<node> *tmpComp = maxComp;
    maxComp = minComp;
    minComp = tmpComp;
    maxNode = target;
  }//end if

  unsigned int maxLabel = component[maxNode];
  while (!minComp->empty()) {
    node minCompNode = minComp->front();
    component[minCompNode] = maxLabel;
    subTrees[minCompNode] = maxComp;
    maxComp->push_back (minCompNode);
    minComp->pop_front();
  }//end for

  //delete the minComp Deque
  delete minComp;
}//end mergeComps

/**
 * Sets the selection proxy mst so it corresponds with the
 * edges and nodes contained in the actual subgraph tree
 */
void setSelectionProxy (SuperGraph *mstGraph, SelectionProxy *tree) {
  tree->setAllNodeValue (true);
  tree->setAllEdgeValue (false);
  edge curEdge;
  forEach (curEdge, mstGraph->getEdges()) 
    tree->setEdgeValue (curEdge, true);
}//end setSelectionProxy

void KruskalMST::mst (SuperGraph *graph, SelectionProxy *tree) {
  cout << "in kruskal mst " << endl;
  SuperGraph *mstGraph = tlp::newSubGraph (graph);
  MetricProxy *weights = getWeights (graph);
  node curNode;
  forEach (curNode, graph->getNodes()) mstGraph->addNode (curNode);

  //if no weights or weights same, hand back a bfs spann tree
  //as all spanning trees are equal in cost.
  if (weights == NULL) {
    node centre = graph->getOneNode();
    deque<node> bfsDeque;
    MutableContainer<bool> visited;
    bfsDeque.push_back (centre);
    visited.setAll (false);
    visited.set (centre.id, true);
    bfsSpanningTree (graph, mstGraph, bfsDeque, visited);
    setSelectionProxy (mstGraph, tree);
    graph->delAllSubGraphs (mstGraph);
    return;
  }//end if

  //add all the nodes to the tree and set their component ids, sizes
  //and roots
  unsigned int index = 0;
  hash_map<node, unsigned int> component;
  hash_map<node, deque<node> *> subTrees;
  forEach (curNode, graph->getNodes()) {
    component[curNode] = index;
    subTrees[curNode] = new deque<node> ();
    subTrees[curNode]->push_back (curNode);
    ++index;
  }//end forEach

  //sort the edges by increasing weight
  SortEdgeIterator edgesSortedByWeight (graph->getEdges(), weights);
  while (mstGraph->numberOfEdges() < graph->numberOfNodes() - 1) {
    edge curEdge = edgesSortedByWeight.next();
    node source = graph->source (curEdge);
    node target = graph->target (curEdge);

    //if we are merging two trees or branching off into unknown node
    //terratory, add the edge
    unsigned int sourceComp = component[source];
    unsigned int targetComp = component[target];
    if ((sourceComp != targetComp)) {
      mstGraph->addEdge (curEdge);
      mergeComps (mstGraph, source, target, component, subTrees);
    }//end if
  }//end while
  delete (subTrees.begin())->second;

  //cleanup and load the mst into the selection proxy
  setSelectionProxy (mstGraph, tree);
  graph->delAllSubGraphs (mstGraph);
}//end kruskalMST
