#include <cstdlib>
#include <tulip/TreeTest.h>
#include <tulip/ForEach.h>
#include <tulip/TlpTools.h>
#include <tulip/SortIterator.h>
#include "AreaRINGS.h"

using namespace std;
using namespace __gnu_cxx;

double polygonAreaFrac (unsigned int n) {
  if ((n == 0) || (n == 1) || (n == 2)) return 0.0;
  else {
    double theta = M_PI / (double)(n);
    double num = (1.0 - sin(theta)) * (1.0 - sin(theta));
    double denom = (1.0 + sin(theta)) * (1.0 + sin(theta));
    return num/denom;
  }//end else
}//end polygonAreaFrac

#define LEVEL_TO_BENDS_MULT (-5.0)
void AreaRINGS::setEdgeBend (node root, node child, unsigned int level) {
  edge rootChildEdge = superGraph->existEdge (root, child);
  if (rootChildEdge.isValid()) {
    Coord edgeBend (0, 0, LEVEL_TO_BENDS_MULT*level);
    vector<Coord> bend;
    bend.push_back (edgeBend);
    rootRelative->setEdgeValue (rootChildEdge, bend);
  }//end if
}//end setEdgeBend

//Function to layout hains of nodes in the hierarchy
void AreaRINGS::layoutChain (node &root, node &chainNode, double rootRadius,
			     unsigned int level) {
  double chainRadius = radius->getNodeValue (chainNode);
  Coord initOffset = rootRelativeOffset->getNodeValue (root);
  rootRelative->setNodeValue (chainNode,
			      Coord (rootRadius + chainRadius, 0, 0) -
			      initOffset);
  Coord centre (chainRadius, 0, 0);
  rootRelativeOffset->setNodeValue (root, initOffset - centre);
  radius->setNodeValue (root, rootRadius + chainRadius);
  setEdgeBend (root, chainNode, level);
}//end layout chain

//Triangle centre function for the outer Soddy Center for triangle
//with sides a, b, c
inline double outSoddyTriCentreFunc (double a, double b, double c,
				     double circumRadius) {
  return (b*c)/(2*(-a + b + c)*circumRadius) - 1;
}//end outSoddyTriCentreFunc

//The normalization value of trilinear coordinates alpha:beta:gamma and
//sides a, b, c
inline double normValue (double a, double b, double c,
			 double alpha, double beta, double gamma) {
  return 1.0/(a*alpha + b*beta + c*gamma);
}//end kValue

//The circumradius of triangle with sides a, b, c
inline double circumRadius (double a, double b, double c) {
  return (a*b*c)/sqrt ((a + b + c)*(b + c - a)*
		       (c + a - b)*(a + b - c));
}//end circumRadius

//The Soddy Outer Radius of a set of circles with radii r1, r2, and r3
inline double soddyOuterRadius (double r1, double r2, double r3) {
  return -(r1*r2*r3)/(r1*r2 + r1*r3 + r2*r3 - 
		      2*sqrt (r1*r2*r3*(r1 + r2 + r3)));

}//end soddyOuterRadius

//Layout two pronged forks
void AreaRINGS::layoutFork (node &root, node &node1, node &node2,
			    double rootRadius, unsigned int level) {
  node leftNode = node1;
  node rightNode = node2;
  double leftRadius = radius->getNodeValue (node1);
  double rightRadius = radius->getNodeValue (node2);
  if (leftRadius > rightRadius) {
    leftRadius = radius->getNodeValue (node2);
    rightRadius = radius->getNodeValue (node1);
    leftNode = node2;
    rightNode = node1;
  }//end if
  
  double cl = rootRadius + leftRadius;
  double cr = rootRadius + rightRadius;
  double lr = leftRadius + rightRadius;
  //assume root at (0,0,0).  Assume right at (cr, 0, 0).  Find left
  double cosTheta = (cl*cl + cr*cr - lr*lr)/(2*cl*cr);
  double sinTheta = sin(acos(cosTheta));
  Coord left (cl*cosTheta, cl*sinTheta, 0);
  Coord right (cr, 0, 0);

  double circumRad = circumRadius (lr, cl, cr);
  double lrTriLinear = outSoddyTriCentreFunc (lr, cl, cr, circumRad);
  double clTriLinear = outSoddyTriCentreFunc (cl, cr, lr, circumRad);
  double crTriLinear = outSoddyTriCentreFunc (cr, lr, cl, circumRad);
 
  //The soddy circle is used if we have an accute triangle.  Otherwise,
  //the bounding circle at the centre of the two largest circles is used
  Coord centre;
  double boundingRadius;
  if ((lrTriLinear > 0) || (crTriLinear > 0) || (clTriLinear > 0)) {
    centre = Coord (rightRadius, 0, 0);
    boundingRadius = cr;
    if (lr > cr) {
      double cosAlpha = (cr*cr + lr*lr - cl*cl)/(2*cr*lr);
      double sinAlpha = sin (acos (cosAlpha));
      centre = Coord (cr - leftRadius*cosAlpha, leftRadius*sinAlpha, 0);
      boundingRadius = lr;
    }//end if
  }//end if
  else {
    double norm = normValue(lr, cl, cr, lrTriLinear, clTriLinear, crTriLinear);
    lrTriLinear *= norm;
    clTriLinear *= norm;
    crTriLinear *= norm;
    centre = left*(crTriLinear*cr) + right*(clTriLinear*cl);
    boundingRadius = 
      soddyOuterRadius (rootRadius, leftRadius, rightRadius);
  }//end else

  Coord initOffset = rootRelativeOffset->getNodeValue (root);
  rootRelative->setNodeValue (leftNode, left - initOffset);
  rootRelative->setNodeValue (rightNode, right - initOffset);
  rootRelativeOffset->setNodeValue (root, initOffset - centre);
  radius->setNodeValue (root, boundingRadius);
  setEdgeBend (root, leftNode, level);
  setEdgeBend (root, rightNode, level);
}//end layoutFork

void genOriginalEdges (string orgEdgeString, vector<edge> &orgEdges) {
  if (orgEdgeString != "") {
    istringstream idparse (orgEdgeString);
    unsigned int edgeid;
    while (!idparse.eof()) {
      idparse >> edgeid;
      orgEdges.push_back (edge (edgeid));
    }//end while
  }//end if
}//end genOrginalEdges

double AreaRINGS::setPlacedChildrenOrder (node root, 
					  vector<node> &placedChildren) {
  MetaGraphProxy *biconComp = 
    superGraph->getProperty<MetaGraphProxy> ("viewMetaGraph");
  MetricProxy *aveAngle = 
    superGraph->getLocalProperty<MetricProxy> ("aveAngle");
  aveAngle->setAllNodeValue (0);
  SuperGraph *rootSubGraph = biconComp->getNodeValue (root);
  if (rootSubGraph == NULL) return 0;
  Coord rootCentre;
  LayoutProxy *rootLayout = 
    rootSubGraph->getProperty<LayoutProxy> ("viewLayout");
  Coord max = rootLayout->getMax();
  Coord min = rootLayout->getMin();
  rootCentre = (max + min)/2.0;
  
  node maxEdgeNode;
  unsigned int maxEdges = 0;
  vector<node>::iterator placeIt = placedChildren.begin();
  for (; placeIt != placedChildren.end(); ++placeIt) {
    SuperGraph *childSubGraph = biconComp->getNodeValue (*placeIt);
    if (childSubGraph == NULL) continue;
    edge rootEdge = superGraph->existEdge (root, *placeIt);
    if (rootEdge.isValid()) {
      const vector<edge> &orgEdge = originalEdges[rootEdge];
      if (maxEdges < orgEdge.size()) {
	maxEdges = orgEdge.size();
	maxEdgeNode = *placeIt;
      }//end if
      double sumAngle = 0;
      vector<edge>::const_iterator orgEdgeIt = orgEdge.begin();
      for (; orgEdgeIt != orgEdge.end(); ++orgEdgeIt) {
	node source = superGraph->getRoot()->source (*orgEdgeIt);
	node target = superGraph->getRoot()->target (*orgEdgeIt);
	node rootSubGraphNode = 
	  (rootSubGraph->isElement (source)) ? source : target;
	Coord rootVec = 
	  rootLayout->getNodeValue (rootSubGraphNode) - rootCentre;
	rootVec /= rootVec.norm();
	sumAngle += acos(rootVec.getX())*((rootVec.getY() < 0) ? -1 : 1);
      }//end for
      double rootAveAngle = sumAngle/static_cast<double> (orgEdge.size());
      aveAngle->setNodeValue (*placeIt, rootAveAngle);
    }//end if
  }//end for
  LessThan aveAngleSort;
  aveAngleSort.metric = aveAngle;
  sort (placedChildren.begin(), placedChildren.end(), aveAngleSort);
  vector<node> behind, newPlacedChildren;
  node nodeBehind;
  placeIt = placedChildren.begin();
  for (; placeIt != placedChildren.end(); ++placeIt) {
    if (*placeIt != maxEdgeNode) behind.push_back (*placeIt);
    else break;
  }//end for
  double startAngle = aveAngle->getNodeValue (maxEdgeNode);
  for (; placeIt != placedChildren.end(); ++placeIt)
    newPlacedChildren.push_back (*placeIt);
  vector<node>::iterator behindIt = behind.begin();
  for (; behindIt != behind.end(); ++behindIt)
    newPlacedChildren.push_back (*behindIt);
  placedChildren.clear();
  placedChildren = newPlacedChildren;
  superGraph->delLocalProperty ("aveAngle");
  return startAngle;
}//end setPlacedChildrenOrder

//less than this number children is special case
#define SPECIAL_CASE_CHILDREN (3) 
void AreaRINGS::layoutUsingRatio (node &root, deque<node> &children,
				  double centreRadius,
				  bool edgeOrderChildren) {
  vector<node> placedChildren;
  unsigned int numLevels = 0; 
  while (!children.empty()) {
    node maxChild;
    double maxChildRadius = 0;
    double secondMaxChildRadius = 0;
    double centreRadiusSquared = centreRadius*centreRadius;
    double r1Squared = centreRadiusSquared;
    while (!children.empty() && (centreRadiusSquared/r1Squared > 
				 polygonAreaFrac (placedChildren.size()))) {
      node curChild = children.front();
      children.pop_front();
      placedChildren.push_back (curChild);
      if (radius->getNodeValue (curChild) > maxChildRadius) {
	secondMaxChildRadius = maxChildRadius;
	maxChild = curChild;
	maxChildRadius = radius->getNodeValue (curChild);
	r1Squared = centreRadius + 2*maxChildRadius;
	r1Squared *= r1Squared;
      }//end if
      else if (radius->getNodeValue (curChild) > secondMaxChildRadius)
	secondMaxChildRadius = radius->getNodeValue (curChild);
    }//end while

    //if we ran out of room, we put the last placed child back in children
    if (centreRadiusSquared/r1Squared <
	polygonAreaFrac (placedChildren.size())) {
      if (maxChild == placedChildren.back()) 
	maxChildRadius = secondMaxChildRadius;
      children.push_front (placedChildren.back());
      placedChildren.pop_back();
    }//end if

    if (placedChildren.size() < SPECIAL_CASE_CHILDREN) {
      radius->setNodeValue (root, centreRadius);
      if (placedChildren.size() == 1) 
	layoutChain (root, placedChildren[0], centreRadius, numLevels + 1);
      else 
	layoutFork (root, placedChildren[0], 
		    placedChildren[1], centreRadius, numLevels + 1);
      centreRadius = radius->getNodeValue (root);
    }//end if
    else {
      double placedCounter = 0;
      double ringRadius = centreRadius + maxChildRadius;
      Coord initOffset = rootRelativeOffset->getNodeValue (root);
      double startAngle = 0;
      if (edgeOrderChildren) 
	startAngle = setPlacedChildrenOrder (root, placedChildren);
      vector<node>::iterator placedIt = placedChildren.begin();
      for (; placedIt != placedChildren.end(); ++placedIt, ++placedCounter) {
	Coord curPos;
	curPos.setX(ringRadius*
		    cos (startAngle + 
			 (2*M_PI*placedCounter)/
			 static_cast<double> (placedChildren.size())));
	curPos.setY(ringRadius*
		    sin (startAngle + 
			 (2*M_PI*placedCounter)/
			 static_cast<double> (placedChildren.size())));
	rootRelative->setNodeValue (*placedIt, curPos - initOffset);
	setEdgeBend (root, *placedIt, numLevels + 1);
      }//end for
      centreRadius = ringRadius + maxChildRadius;
    }//end else
    placedChildren.clear();
    ++numLevels;
  }//end while
  radius->setNodeValue (root, centreRadius);
}//end layoutUsingRatio

void AreaRINGS::handleChain (node &root, std::deque<node> &chain) {
  if (chain.size() == 1) 
    layoutChain (root, chain.front(), radius->getNodeValue (root));
  else if (chain.size() == 2) {
    node left, right;
    left = chain.front();
    chain.pop_front();
    right = chain.front();
    layoutFork (root, left, right, radius->getNodeValue (root));
    Coord initOffset = rootRelativeOffset->getNodeValue (right);
    Coord leftRootRelative = rootRelative->getNodeValue(left);
    rootRelativeOffset->setNodeValue (right, initOffset - leftRootRelative);
  }//end else if
  
  else {
    node chainEnd = chain.back();
    if (superGraph->outdeg (chainEnd) != 0) chain.pop_back();
    deque<node> children = chain;
    layoutUsingRatio (root, chain, radius->getNodeValue (root), false);
    Coord prevOffset (0,0,0);
    while (!children.empty()) {
      node curChild = children.front();
      children.pop_front();
      Coord initOffset = rootRelativeOffset->getNodeValue (curChild);
      rootRelativeOffset->setNodeValue (curChild, initOffset - prevOffset);
      prevOffset = rootRelative->getNodeValue(curChild);
    }//end while
    if (superGraph->outdeg (chainEnd) != 0)
      layoutChain (root, chainEnd, radius->getNodeValue (root));
  }//end else
  while (!chain.empty()) chain.pop_front();
}//end handleChain

void AreaRINGS::AreaRINGSLayout (node curRoot, deque<node> *chainNodes) {
  //if leaf node, it may end a chain.
  if (superGraph->outdeg (curRoot) == 0) return;

  //if degree one, it continues a chain
  else if (superGraph->outdeg (curRoot) == 1) {
    Iterator<node> *nextChain = superGraph->getOutNodes (curRoot);
    node child = nextChain->next();
    delete nextChain;
    chainNodes->push_back (child);
    AreaRINGSLayout (child, chainNodes);
  }//end if
  
  //otherwise, it starts or stops a chain, or is regular AreaRings layout
  else {
    node curChild;
    forEach (curChild, superGraph->getOutNodes (curRoot)) {
      if (superGraph->outdeg (curChild) == 1) 
	AreaRINGSLayout (curChild, &chains[curChild]);
      else AreaRINGSLayout (curChild, NULL);
    }//end forEach
  }//end else
  
  //if degree one, it is part of a chain
  if (superGraph->outdeg (curRoot) == 1) {
    stdext::hash_map<node, deque<node> >::const_iterator rootChain =
      chains.find (curRoot);
    if (rootChain != chains.end()) handleChain (curRoot, chains[curRoot]);
  }//end if
  

  //If there are two nodes, draw a fork
  else if (superGraph->outdeg (curRoot) == 2) {
    node left, right;
    Iterator<node> *children = superGraph->getOutNodes (curRoot);
    left = children->next();
    right = children->next();
    delete children;
    layoutFork (curRoot, left, right, radius->getNodeValue (curRoot));
  }//end if
  
  //otherwise it is standard area aware rings
  else {
    SortNodeIterator children (superGraph->getOutNodes (curRoot), radius);
    deque<node> dequeChildren;
    while (children.hasNext()) dequeChildren.push_back (children.next());
    layoutUsingRatio (curRoot, dequeChildren, radius->getNodeValue (curRoot),
		      true);
  }//end else
}//end AreaRINGSLayout

void AreaRINGS::setGlobalCoords (node curRoot, edge cameFromEdge,
				 Coord curPos) {
  //set the position of the root
  Coord offset = rootRelativeOffset->getNodeValue (curRoot);
  layoutProxy->setNodeValue (curRoot, curPos + offset);
  if (cameFromEdge.isValid()) {
    vector<Coord> rootBend = rootRelative->getEdgeValue (cameFromEdge);
    if (!rootBend.empty()) {
      Coord leafBend = curPos + offset + rootBend[0];
      node other = superGraph->opposite (cameFromEdge, curRoot);
      Coord prevRootBend = layoutProxy->getNodeValue (other) + rootBend[0];
      rootBend.clear();
      rootBend.push_back (prevRootBend);
      rootBend.push_back (leafBend);
      layoutProxy->setEdgeValue (cameFromEdge, rootBend);
    }//end if
  }//end if

  //recurse on the children, if any
  node curChild;
  forEach (curChild, superGraph->getOutNodes (curRoot)) {
    edge rootChildEdge = superGraph->existEdge (curRoot, curChild);
    Coord trans = rootRelative->getNodeValue (curChild);
    setGlobalCoords (curChild, rootChildEdge, curPos + offset + trans);
  }//end forEach
}//end end setGlobalCoords

bool AreaRINGS::run () {
  SizesProxy *sizes = superGraph->getProperty<SizesProxy> ("viewSize");
  rootRelative = 
    superGraph->getLocalProperty<LayoutProxy> ("RINGS root relative");
  rootRelativeOffset = 
    superGraph->getLocalProperty<LayoutProxy> ("RINGS root relative offset");
  
  rootRelative->setAllNodeValue (Coord (0, 0, 0));
  rootRelativeOffset->setAllNodeValue (Coord (0, 0, 0));

  //the squared radius of each node (Area/PI)
  node curNode;
  radius = 
    superGraph->getLocalProperty<MetricProxy> ("RINGS radius");
  forEach (curNode, superGraph->getNodes()) {
    Size curSize = sizes->getNodeValue (curNode);
    radius->setNodeValue (curNode, 
			  sqrt ((curSize.getW()*curSize.getW())/4.0 + 
				(curSize.getH()*curSize.getH())/4.0));
  }//end forEach
  StringProxy *orgEdgeID =
    superGraph->getProperty<StringProxy> ("Original Edge ID");
  edge curEdge;
  forEach (curEdge, superGraph->getEdges()) {
    string edgeString = orgEdgeID->getEdgeValue (curEdge);
    if (edgeString != "") 
      genOriginalEdges (edgeString, originalEdges[curEdge]);
  }//end forEach
  
  //layout the rings and set global coordinates
  node root;
  tlp::getSource (superGraph, root);
  deque<node> *chainNodes = NULL;
  if (superGraph->outdeg (root) == 1) chainNodes = &chains[root];
  AreaRINGSLayout (root, chainNodes);
  edge startEdge (0xffffffff); //id for invalid edge
  setGlobalCoords (root, startEdge, Coord (0, 0, 0));
  superGraph->delLocalProperty ("RINGS root relative");
  superGraph->delLocalProperty ("RINGS radius");
  superGraph->delLocalProperty ("RINGS root relative offset");
  return true;
}//end run

bool AreaRINGS::check (std::string &msg) { 
  if (TreeTest::isTree (superGraph)) {
    msg = "";
    return true;
  }
  else {
    msg = "The graph must be a tree (in Area aware RINGS)";
    return false;
  }
}

void AreaRINGS::reset () {}

LAYOUTPLUGIN(AreaRINGS, "RINGS Area", "Daniel Archambault", "20/06/2005", "Ok", "0", "1");

