/*
 *  
 *  Daniel Archambault 2006
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/

#ifndef _GRID_HIERARCHY_HPP_
#define _GRID_HIERARCHY_HPP_
#include "configs.h"
#include "calcFuncs.h"
#include "grid.hpp"

class GridHierarchy {
  
private:
  unsigned short numLevels;
  Grid_t *hierarchy;
  prec_t maxVoxelSize, minVoxelSize;

  inline unsigned short getLevelFromParticleRadius (prec_t radius) {
    return static_cast<unsigned short> 
      (floor ((log(maxVoxelSize) - log(VOX_SIZE_MULT*radius))/log(2.0)));
  }//end getLevelFromParticle

public:
  GridHierarchy (prec_t minNodeSize, prec_t maxNodeSize, 
		 NodeContainer &nodes) {
    //determine the number of levels.  
    maxVoxelSize = VOX_SIZE_MULT*maxNodeSize;
    minVoxelSize = VOX_SIZE_MULT*minNodeSize;
    numLevels = getLevelFromParticleRadius (minNodeSize);
    ++numLevels;

    //count the number of particles at each level
    unsigned int numInGridCount[numLevels];
    for (unsigned int i = 0; i < numLevels; ++i) numInGridCount[i] = 0;
    for ( NodeContainer::size_type ii=0; ii<nodes.size(); ++ii )
      ++numInGridCount[getLevelFromParticleRadius (nodes[ii].radius())];
    
    //determine the max grid extent by max extent required by a level
    prec_t outerRadius = 0;
    prec_t voxelLength = VOX_SIZE_MULT*maxNodeSize;
    for (unsigned int i = 0; i < numLevels; ++i) {
      prec_t levelOuterRadius = 
	computeOuterRadius2d (voxelLength/VOX_SIZE_MULT, 
			      numInGridCount[i]);
      if (outerRadius < levelOuterRadius) outerRadius = levelOuterRadius;
      voxelLength /= 2.0;
    }//end for i
    
    //make grid hierarchy
    hierarchy = new Grid_t [numLevels];
    voxelLength = VOX_SIZE_MULT*maxNodeSize;
    for (unsigned short i = 0; i < numLevels; ++i) {
      gridPrepAndInit (hierarchy[i], voxelLength, outerRadius);
      voxelLength /= 2.0;
    }//end for i
  }//end constructor

  ~GridHierarchy () {
    delete [] hierarchy;
  }//end destructor

  unsigned short getFinestGridIndex (Node &n) {
    return getLevelFromParticleRadius (n.radius());
  }//end getFinestGrid

  Grid_t &getGrid (Node &n) {
    unsigned short level = n.getHierarchyLevel();
    return hierarchy[level];
  }//end getGrid

  Grid_t &getGrid (unsigned short index) {
    return hierarchy[index];
  }//end getGrid

  void loadNode (Node &n) {
    unsigned short finest = getFinestGridIndex (n);
    n.setHierarchyLevel (finest);
    shift_particle (n, hierarchy[finest]);
  }//end loadNode

  void shiftNode (Node& n) {
    unsigned short level = n.getHierarchyLevel();
    shift_particle (n, hierarchy[level]);
  }//end shiftNode

  unsigned short getNumLevels () {
    return numLevels;
  }//end getNumLevels
};
#endif
