#include <tulip/ForEach.h>
#include "configs.h"
#include "calcFuncs.h"
#include "graph.hpp"
#include "AreaLGL.h"

using namespace std;

Graph_t::vertex_index_map idmap; 
Graph_t::weight_map weightMap;

int runLayout (int argc, char ** argv,  Graph_t &G,
	       NodeContainer &nodes, double maxRad, double minRad, 
	       double &outerRadius, LevelMap &levels, ParentMap &parents,
	       unsigned int totalLevels, bool hasMST);

void initNodes (SuperGraph *graph, node *indexToNode) {
  unsigned int index = 0;
  node curNode;
  idmap.clear();
  forEach (curNode, graph->getNodes()) {
    ostringstream convert;
    convert << curNode.id << flush;
    idmap.createMap (convert.str(), index);
    indexToNode[index] = curNode;
    ++index;
  }//end forEach
}//end initNodes

MetricProxy *getWeights (SuperGraph *graph) {
  bool existWeights = graph->existProperty ("weights");
  if (existWeights)
    return graph->getProperty<MetricProxy> ("weights");
  return NULL;
}//end getWeights

SelectionProxy *getMST (SuperGraph *graph) {
  bool existMST = graph->existProperty ("mst");
  if (existMST)
    return graph->getProperty<SelectionProxy> ("mst");
  return NULL;
}//end getMST

void initEdges (SuperGraph *graph, Graph_t::boost_graph &bg,
		MetricProxy *weights) {
  edge curEdge;
  forEach (curEdge, graph->getEdges()) {
    ostringstream source, target;
    source << graph->source (curEdge).id << flush;
    target << graph->target (curEdge).id << flush; 
    Graph_t::edge_descriptor e;
    bool inserted;
    tie(e,inserted) = add_edge (idmap.findLatter (source.str()), 
				idmap.findLatter (target.str()), bg);
    if (weights != NULL) weightMap[e] = weights->getEdgeValue (curEdge);
  }//end forEach
}//initEdges

void initParticles (Graph_t &g, SuperGraph *graph, node *indexToNode,
		    NodeContainer &nodes, double &maxRad,
		    double &minRad, double &aveRad) {
  SizesProxy *sizes = graph->getProperty<SizesProxy> ("viewSize");
  maxRad = 0;
  minRad = DBL_MAX;
  aveRad = 0;
  // Initialize the ids of the nodes
  for (NodeContainer::size_type ii=0; ii < nodes.size(); ++ii) {
    nodes.ids[ii] = g.idFromIndex(ii);
    nodes[ii].id (g.idFromIndex(ii));
    Size size = sizes->getNodeValue (indexToNode[ii]);
    double rad = sqrt (size.getW()*size.getW()/4.0 +
		       size.getH()*size.getH()/4.0);
    nodes[ii].radius (rad);
    aveRad += rad;
    if (maxRad < rad) maxRad = rad;
    if (minRad > rad) minRad = rad;
  }//end for
  aveRad /= static_cast<double> (graph->numberOfNodes());
}//end initParticles

bool getLGLRoot (SuperGraph *graph, string &lglroot) {
  DataSet &attributes = graph->getAttributes();
  unsigned int lglrootint;
  bool passedLGLRoot = attributes.get ("lglroot", lglrootint);
  if (passedLGLRoot) {
    ostringstream rootstrstr;
    rootstrstr << lglrootint << flush;
    lglroot = rootstrstr.str();
    bool rootFound = idmap.doesMapExist (lglroot);
    if (!rootFound) {
      cerr << "Passed root " << lglrootint << " not found in graph "
	   << "treating as if a root was not passed." << endl;
      return false;
    }//end if
    else cerr << "Using passed root " << lglrootint << "." << endl;
  }//end if
  return passedLGLRoot;
}//end getLGLRoot

unsigned int generateLevelMap (SuperGraph *graph,
			       SelectionProxy *mst,
			       string root,
			       MetricProxy *weights, LevelMap &levels,
			       ParentMap &parents) {
  Graph_t mstGraph;
  Graph_t::vertex_index_map mstIDMap;
  Graph_t::boost_graph mstBG;
  bool inserted;
  Graph_t::edge_descriptor e;
  Graph_t::vertex_descriptor rootDesc = idmap.findLatter (root);
  Graph_t::weight_map mstWeights;

  edge curEdge;
  forEach (curEdge, graph->getEdges()) {
    if (mst->getEdgeValue (curEdge)) {
      ostringstream source, target;
      source << graph->source (curEdge).id << flush;
      target << graph->target (curEdge).id << flush; 
      tie(e,inserted) = add_edge (idmap.findLatter (source.str()), 
				  idmap.findLatter (target.str()), 
				  mstBG);
      if (weights != NULL)
	mstWeights[e] = weights->getEdgeValue (curEdge);
    }//end if
  }//end forEach
  mstGraph.vertexIdMap (idmap);
  mstGraph.boostGraph (mstBG);
  mstGraph.hasWeights (weights != NULL);
  mstGraph.weights (mstWeights);
  setLevelMapFromMST (mstGraph, levels, parents, rootDesc);
  int maxL = 0;
  for (LevelMap::const_iterator i=levels.begin(); i != levels.end(); ++i) {
    maxL = (int) GraphDetail::mmax<long>( *i ,maxL );
  }
  // Return the total number of levels
  return maxL;
}//end generateLevelMap

bool AreaLGL::run () {
  Graph_t::boost_graph bg;
  node indexToNode[superGraph->numberOfNodes()];
  initNodes (superGraph, indexToNode);
  MetricProxy *weights = getWeights (superGraph);
  initEdges (superGraph, bg, weights);
  Graph_t g;
  g.boostGraph(bg);
  g.hasWeights (weights != NULL);
  if (weights != NULL) g.weights (weightMap);
  g.vertexIdMap (idmap);
  
  //init the paricle container for LGL
  NodeContainer nodes(g.vertexCount()); 
  double maxRad, minRad, aveRad;
  initParticles (g, superGraph, indexToNode, nodes,
		 maxRad, minRad, aveRad);

  //get the root node and mst if passed
  string lglroot;
  bool passedLGLRoot = getLGLRoot (superGraph, lglroot);
  SelectionProxy *mst = getMST (superGraph);

  //if there is a passed mst, generate the level map
  LevelMap levels (nodes.size() , 1); 
  ParentMap parents (nodes.size() , 0);
  unsigned int numLevels = 0;
  if ((mst != NULL) && passedLGLRoot)
    numLevels = generateLevelMap (superGraph, mst, lglroot, 
				  weights, levels, parents);
  

  // This determines the space necessary to place all the nodes.
  prec_t outerRadius = -1.0;
  if ( DIMENSION == 2 ) { 
    outerRadius = computeOuterRadius2d (maxRad, nodes.size());
  } else if ( DIMENSION == 3 ) {
    outerRadius = (prec_t)2*VOX_SIZE_MULT*maxRad*sqrt((double)nodes.size());
  } else { cerr << "Only 2 or 3 dimensions\n"; exit(EXIT_FAILURE); }
  
  ostringstream strOuterRadius, strNodeRad, strIdealEdgeLen;
  strOuterRadius << outerRadius << flush;
  strNodeRad << maxRad << flush;
  strIdealEdgeLen << ((VOX_SIZE_MULT*maxRad)/4.0) << flush;

  clock_t start = clock();
  time_t startLongTime;
  time (&startLongTime);
  unsigned int numArgs = 9;
  if (passedLGLRoot) numArgs += 2;
  if (weights != NULL) ++numArgs;
  char *charArguments [numArgs];
  char argumentsArray [numArgs*20];
  for (unsigned int i = 0; i < numArgs; i++)
    charArguments[i] = argumentsArray + i*20;
  sprintf (charArguments[0], "lglayout");
  sprintf (charArguments[1], "-R");
  sprintf (charArguments[2], strOuterRadius.str().c_str());
  sprintf (charArguments[3], "-S");
  sprintf (charArguments[4], strNodeRad.str().c_str());
  sprintf (charArguments[5], "-q");
  sprintf (charArguments[6], strIdealEdgeLen.str().c_str());
  if (passedLGLRoot) {
    sprintf (charArguments[7], "-z");
    sprintf (charArguments[8], lglroot.c_str());
  }//end if
  if (weights != NULL) sprintf (charArguments[numArgs - 1], "-O");

  runLayout (numArgs, charArguments, g, nodes, maxRad, minRad, outerRadius,
	     levels, parents, numLevels, (mst != NULL) && passedLGLRoot);
  clock_t end = clock();
  time_t endLongTime;
  time (&endLongTime);
  double elapsedTime = (double) (end - start) / (double) (CLOCKS_PER_SEC);
  double elapsedLongTime = difftime (endLongTime, startLongTime);
  cout << "Area LGL took " << elapsedTime << " seconds " << endl;
  cout << "In long time " << elapsedLongTime << " seconds " << endl;

  for ( NodeContainer::size_type ii=0; ii<nodes.size(); ++ii ) 
    layoutProxy->setNodeValue (indexToNode[ii], Coord (nodes[ii].X()[0],
	                                               nodes[ii].X()[1]));

  //set the final positions
  /* node botLeft = superGraph->addNode();
  node botRight = superGraph->addNode();
  node topLeft = superGraph->addNode();
  node topRight = superGraph->addNode();
  layoutProxy->setNodeValue (botLeft, Coord (0,0,0));
  layoutProxy->setNodeValue (botRight, Coord (outerRadius, 0, 0));
  layoutProxy->setNodeValue (topLeft, Coord (0, outerRadius, 0));
  layoutProxy->setNodeValue (topRight, Coord (outerRadius, outerRadius, 0));
  superGraph->addEdge (botLeft, botRight);
  superGraph->addEdge (botRight, topRight);
  superGraph->addEdge (topRight, topLeft);
  superGraph->addEdge (topLeft, botLeft);
  for (double i = 0; i < outerRadius; i += VOX_SIZE_MULT*maxRad) {
    node hleft = superGraph->addNode();
    node hright = superGraph->addNode();
    node vleft = superGraph->addNode();
    node vright = superGraph->addNode();
    layoutProxy->setNodeValue (hleft, Coord (0, i , 0));
    layoutProxy->setNodeValue (hright, Coord (outerRadius, i, 0));
    layoutProxy->setNodeValue (vleft, Coord (i, outerRadius, 0));
    layoutProxy->setNodeValue (vright, Coord (i, 0, 0));
    superGraph->addEdge (hleft, hright);
    superGraph->addEdge (vleft, vright);
  }//end for i
  */
  return true;
}//end run

bool AreaLGL::check (std::string &msg) { return true; }

void AreaLGL::reset () {}

LAYOUTPLUGIN(AreaLGL, "LGL Area", "Daniel Archambault", "20/06/2005", "Ok", "0", "1");

