//  
//
//  Copyright (c) 2002 Alex Adai, All Rights Reserved.
//  
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation; either version 2 of
//  the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
//  MA 02111-1307 USA
//

#include <vector>
#include "calcFuncs.h"
#ifdef GRID_HIERARCHY
#include "gridHierarchy.hpp"
#endif

using namespace boost;

//---------------------------------------------------------------

void * calcInteractions ( void * arg_ )
{
  ThreadArgs * args = static_cast<ThreadArgs*>(arg_);
  GridIterator *grid_i = args->gridIterators[0];
  VoxelHandler &vh = *(args->voxelHandler);
  args->nodeHandler->springConstant( args->casualSpringConstant );
  vector<Voxel_t *> touchedVoxels;
  vector<Graph_t::vertex_descriptor>::iterator nodeIt = 
    args->activeNodes->begin();
  vector<Graph_t::vertex_descriptor>::iterator activeEnd =
    args->activeNodes->end();
  NodeContainer &nodes = *(args->nodes);
  for (; nodeIt != activeEnd; ++nodeIt) {
#ifdef GRID_HIERARCHY
    GridHierarchy &gh = *(args->hierarchy);
    short curGridIndex = nodes[*nodeIt].getHierarchyLevel();
    //cout << "node " << nodes[*nodeIt].id() << endl;
    for (; curGridIndex >= 0; --curGridIndex) {
      //cout << "curGridIndex " << curGridIndex << endl;
      Grid_t *curGrid = &gh.getGrid (curGridIndex);
      //cout << curGrid << endl;
      FixedVec<long , DIMENSION> coords;
      Voxel_t *vox1 = curGrid->getVoxelFromPosition (nodes[*nodeIt].X());
      //cout << "vox1 index " << vox1->index() << endl;
      curGrid->coordToVoxCoord (vox1->orig(), coords);
      grid_i = args->gridIterators[curGridIndex];
      grid_i->current (coords);
      args->nodeHandler->eqDistance (curGrid->voxelWidth());
      while (grid_i->incNbhr()) {
	Voxel_t* vox2 = &grid_i->nhbrVox();
	if (!vox2->empty()) {
	  vh.voxelParticleInteractions(nodes[*nodeIt], 
				       vox1->index() == vox2->index(), 
				       *vox2);
	}//end if
      }//end while
    }//end for
#else
    //get this node's voxel
    Grid_t &grid = *(args->grid);
    Voxel_t &vox1 = grid[nodes[*nodeIt].container()];
    if (!vox1.isDone()) {
      //if not touched, compute repulsive forces with all adj voxels
      touchedVoxels.push_back (&vox1);
      vox1.setDone(true);
      FixedVec<long , DIMENSION> coords;
      grid.coordToVoxCoord (vox1.orig(), coords);
      grid_i->current (coords);
      while ( grid_i->incNbhr() ) {
	Voxel_t& vox2 = grid_i->nhbrVox();
	if (!vox2.empty()) {
	  vh.twoVoxelInteractions(vox1 , vox2);
	} 
      }
    }
#endif
  }
  vector<Voxel_t *>::iterator voxIt = touchedVoxels.begin();
  for (; voxIt != touchedVoxels.end(); ++voxIt) (*voxIt)->setDone(false);
  return args;
}

//---------------------------------------------------------------

void * integrateParticles ( void * arg_ )
{
  typedef Graph_t::vertex_iterator Vi;
  typedef Graph_t::out_edge_iterator Ei;
  //cout << "integrateParticles" << endl;
  ThreadArgs& args = *(static_cast<ThreadArgs*>(arg_));  
  NodeContainer& nodes = *(args.nodes);
#ifdef GRID_HIERARCHY
  GridHierarchy &gh = *(args.hierarchy);
#else
  Grid_t& grid = *(args.grid);
#endif
  NodeInteractionHandler& nih = *(args.nodeHandler);
  const Graph_t& layout_graph = *(args.layout_graph);
  Vi v, vend;
  tie( v , vend ) = vertices( layout_graph.boostGraph() );
  vector<Graph_t::vertex_descriptor>::iterator nodeIt = 
    args.activeNodes->begin();
  vector<Graph_t::vertex_descriptor>::iterator activeEnd =
    args.activeNodes->end();
  for (; nodeIt != activeEnd; ++nodeIt) {
    Node& n = nodes[ *nodeIt ];
    nih.enforceFLimit( n );
    nih.integrate( n );
#ifdef GRID_HIERARCHY
    if (gh.getGrid(n).checkInclusion(n.X())) { gh.shiftNode (n); }
#else
    if ( grid.checkInclusion( n.X() ) ) {shift_particle( n , grid );}
#endif  
    // Reset the forces to zero for next 
    // iteration
    n.F(0);
  }
  return arg_;
}

//---------------------------------------------------------------

void * collectEdgeStats( void * arg_ )
{
  //cout << "collectingStats" << endl;
  typedef Graph_t::edge_iterator Ei;
  typedef Graph_t::vertex_descriptor vertex_descriptor;
  ThreadArgs& args = *(static_cast<ThreadArgs*>(arg_));
  const unsigned int currentLevel = args.currentLevel;
  const LevelMap& levels = *(args.levels);
  NodeContainer& nodes = *(args.nodes);
  long whichThread = args.whichThread;
  long threadCount = args.threadCount;
  const Graph_t& layout_graph = *(args.layout_graph);
  int edgeCount = num_edges( layout_graph.boostGraph() );
  int ctr = 0;
  prec_t dx = 0;
  Ei ei , eend;
  tie( ei , eend ) = edges( layout_graph.boostGraph() );
  if ( whichThread != 0 ) { advance( ei , whichThread ); }
  for ( int ectr = whichThread; ectr<edgeCount; ectr+=threadCount )
    {
      if ( ectr != whichThread ) { advance( ei , threadCount ); }
      vertex_descriptor v1 = source( *ei , layout_graph.boostGraph() );
      vertex_descriptor v2 = target( *ei , layout_graph.boostGraph() );
      if ( levels[ v1 ] == currentLevel ||
	   levels[ v2 ] == currentLevel )
	{
	  const Node& n1 = nodes[ v1 ];
	  const Node& n2 = nodes[ v2 ];
	  dx += n1.X().distance(n2.X());
	  dx -= n1.radius() + n2.radius();
	  ++ctr;
	}
    }
  args.stats->add2Stats_dx( dx );
  args.stats->count( ctr );  
  return arg_;
}

//---------------------------------------------------------------

void * onlyEdgeInteractions( void * arg_ )
{
  //cout << "onlyEdgeInteractions" << endl;
  typedef Graph_t::edge_iterator Ei;
  ThreadArgs& args = *(static_cast<ThreadArgs*>(arg_));
  NodeContainer& nodes = *(args.nodes);
  long whichThread = args.whichThread;
  long threadCount = args.threadCount;
  NodeInteractionHandler& nih = *(args.nodeHandler);
  nih.springConstant( args.specialSpringConstant );
  const Graph_t& layout_graph = *(args.layout_graph);
  int edgeCount = num_edges( layout_graph.boostGraph() );
  Ei ei , eend;
  tie( ei , eend ) = edges( layout_graph.boostGraph() );
  if ( whichThread != 0 ) { advance( ei , whichThread ); }
  for ( int ectr = whichThread; ectr<edgeCount; ectr+=threadCount )
    {
      if ( ectr != whichThread ) { std::advance( ei , threadCount ); }
      Node& n1 = nodes[ source( *ei , layout_graph.boostGraph() ) ];
      Node& n2 = nodes[ target( *ei , layout_graph.boostGraph() ) ];
      prec_t d = euclideanDistance( n1.X().begin() , n1.X().end() ,
				    n2.X().begin() );
      if (d > nih.eqDistance()) {
	nih.springRepulsiveInteraction( n1 , n2 );
      }
    }
  return arg_;
}
 
//--------------------------------------------------------------

void beginSimulation( ThreadContainer& threads , 
		      vector<Graph_t::vertex_descriptor> &activeNodes,
		      prec_t cutOffPrecision , TimeKeeper_t& timer , 
		      ThreadArgs * threadArgs , PCChaperone& chaperone ,
		      unsigned int totalLevels , bool givenCoords, prec_t placementDistance,
		      prec_t placementRadius , bool placeLeafsClose , bool silentOutput)
{
  Graph_t& current_layout = *(threadArgs->layout_graph);
  Graph_t& full_graph = *(threadArgs->full_graph);
  LevelMap& levels = *(threadArgs->levels);
  ParentMap& parents = *(threadArgs->parents);
  NodeContainer& nodes = *(threadArgs->nodes);
  Grid_t& grid = *(threadArgs->grid);
  GridHierarchy &gh = *(threadArgs->hierarchy);
  long threadCount = threads.size();
  unsigned int currentLevel = 1;
#ifdef GRID_HIERARCHY
  cout << "grid has " << gh.getNumLevels () << " levels " << endl;
#endif

  while( currentLevel <= totalLevels ) {
    // Place and initialize the next layer of the graph
    if ( ! givenCoords ) {
      addNextLevelFromMap (current_layout, full_graph,
			   levels, currentLevel );    
      initializeCurrentLayer (current_layout, nodes, activeNodes, gh, 
			      levels, 
			      parents, grid, currentLevel, full_graph,
			      placementDistance, placementRadius, 
			      placeLeafsClose);
    } else {
      currentLevel = totalLevels;
    }

    prec_t avgPrevious = 0.0;
    prec_t dx = 10000000.; 
    int iterationCtr = 0; 
  
    do {

      // Repulsive terms
      for ( long ii=0; ii<threadCount; ++ii ) {
	threadArgs[ii].currentLevel = currentLevel;
	threadArgs[ii].nodeHandler->springConstant(threadArgs[ii].casualSpringConstant );
	threadArgs[ii].nodeHandler->eqDistance(threadArgs[ii].nbhdRadius);
	threads[ii].create( calcInteractions , 
			    static_cast<void *>(&threadArgs[ii]) );
      } threads.wait();

      // Attractive terms
      for ( long ii=0; ii<threadCount; ++ii ) {
	threadArgs[ii].nodeHandler->springConstant(threadArgs[ii].specialSpringConstant );
	threadArgs[ii].nodeHandler->eqDistance(threadArgs[ii].eqDistance);
	threads[ii].create( onlyEdgeInteractions , 
			    static_cast<void *>(&threadArgs[ii]) );
      } threads.wait();

      // Integrate for next time step
      for ( long ii=0; ii<threadCount; ++ii ) {
	threads[ii].create( integrateParticles ,
			    static_cast<void *>(&threadArgs[ii]) ); 
      } threads.wait();
      
      // Collect stats for progress
      for ( long ii=0; ii<threadCount; ++ii ) {
	threads[ii].create( collectEdgeStats , 
			    static_cast<void *>(&threadArgs[ii]) );
      } threads.wait();

      prec_t dxNew = collectOutput( &threadArgs[0] , chaperone ); 
      if ( ! silentOutput ) {
	printOutput( timer.iteration() , dxNew , currentLevel , std::cerr );
      }
      chaperone.level( currentLevel );

      prec_t avg = ( dxNew + dx ) * .5;
      if ( abs(dxNew - dx)/dxNew < cutOffPrecision || iterationCtr > 150 ||
	   abs(avgPrevious - avg)/avg < .1 * cutOffPrecision ) {
	++timer;
	break; 
      } 

      avgPrevious = avg;
      dx = dxNew;

      if ( threadArgs[0].stats->collectStatsCheck( timer.iteration() ) ) {
	char layerChar[64];
	sprintf(layerChar,"coords/%dlayout%d",timer.iteration(),currentLevel);
	chaperone.posOutFile( layerChar );
	chaperone.writeOutFiles();
      }

      ++iterationCtr;
      ++timer;
      
      } while ( timer.rangeCheck() ); 
    
    ++currentLevel;
  } // End of iterative layout.  
}

//----------------------------------------------------------

void initializeCurrentLayer( Graph_t& layout_graph , NodeContainer& nodes ,
			     vector<Graph_t::vertex_descriptor> &activeNodes,
			     GridHierarchy &gh,
			     LevelMap& levels , ParentMap& parents , Grid_t& grid ,
			     unsigned int currentLevel , Graph_t& actualG, prec_t placementDistance,
			     prec_t placementRadius , bool placeLeafsClose )
{
  Graph_t::vertex_iterator v, vend; 
  out_graph edges2add;
  FixedVec_p cm;
  // Get the center of mass for the current level
  if ( currentLevel == 1 ) {
    tie( v , vend ) = vertices( layout_graph.boostGraph() );
    for ( ; v!=vend ; ++v ) {
      // Find the root node and that is the cent mass
      if ( levels[*v] == 0 ) { 
	cm = nodes[*v].X(); 
	break; 
      }
    }
  } 
  else {
    cm = calcCenterOfMass( layout_graph , nodes , levels , currentLevel );
  }
  tie( v , vend ) = vertices( layout_graph.boostGraph() );
  // Determine which edges are going to be added to the current layout
  for ( ; v!=vend ; ++v ) {
    if ( levels[*v] == currentLevel ) {
      add_edge( parents[*v] , *v , edges2add );
    }
  }
  layerNPlacement( nodes , activeNodes, gh, grid , edges2add , cm, 
		   currentLevel, parents, actualG, levels, 
		   placementDistance, placementRadius, placeLeafsClose);
}

//----------------------------------------------------------

#define SPRAY_ANGLE (M_PI/9.0)
void layerNPlacement( NodeContainer& nodes , 
		      vector<Graph_t::vertex_descriptor> &activeNodes,
		      GridHierarchy &gh,
		      Grid_t& grid , out_graph& g , 
		      FixedVec_p& cm, unsigned int currentLevel , ParentMap& parents ,
		      Graph_t& actualG, LevelMap& lm , prec_t placementDistance,
		      prec_t placementRadius , bool placeLeafsClose )
{
  typedef Sphere<prec_t> S;
  typedef S::vec_type Vec;
  typedef out_graph::out_edge_iterator oei;
  typedef out_graph::vertex_descriptor vertex_descriptor;
  typedef out_graph::vertex_iterator vertex_iterator;  

  vertex_iterator v , vend;
  tie( v , vend ) = vertices( g );

  for ( ; v!=vend; ++v ) { 
    oei e, eend;
    tie( e , eend ) = out_edges( *v , g );
    int vertices2place=0;
    oei ee = e;
    while( ee++ != eend ) { ++vertices2place; }
    //int vertices2place = distance( e , eend );
    if ( vertices2place == 0 ) { continue; }
    Node& parentNode = nodes[ *v ];
    Node& parentParentNode = nodes[ parents[*v] ];
    Vec spot( parentNode.X().begin() , parentNode.X().end() );
    vector< Vec > x;

    tie( e , eend ) = out_edges( *v , g );
    prec_t maxChildRad = 0;
    prec_t aveChildRad = 0;
    for ( ; e!=eend ; ++e ) {
      vertex_descriptor other = target( *e , g );
      if (maxChildRad < nodes[other].radius())
	maxChildRad = nodes[other].radius();
      aveChildRad += nodes[other].radius();
      activeNodes.push_back (other);
    }
    aveChildRad /= vertices2place;

    S s( spot );
    s.radius (placementRadius + parentNode.radius() + maxChildRad);

    if ( currentLevel == 1 ) {
#ifdef GRID_HIERARCHY
      tie( e , eend ) = out_edges( *v , g );
      for ( ; e!=eend ; ++e ) {
	vertex_descriptor other = target( *e , g );
	s.radius (parentNode.radius() + 
		  nodes[other].radius() + aveChildRad);
	vector< Vec > child = seriesOfPointsOnSphere (s , 1);
	x.push_back (child[0]);
      }
#else
      x =  seriesOfPointsOnSphere( s , vertices2place );
#endif
    } else {

      // Determine if any of these children are parents
      tie( e , eend ) = out_edges( *v , g );
      bool hasChildren = true;
      if ( placeLeafsClose ) {
	for ( ; e!=eend ; ++e ) {
	  vertex_descriptor other = target( *e , g );
	  if ( ! doesVertexHaveAnyChildren( actualG, other , g , parents ) ) {
	    hasChildren = false;
	    break;
	  }
	}
      }
  
      FixedVec_p d;
      generatePlacementVector(d, parentNode.X(), parentParentNode.X(), cm);
      if (DIMENSION == 2) {
	FixedVec_p unitD = d;
	scale (unitD.begin() , unitD.end() , 1.0/unitD.magnitude());
	prec_t angle = acos(unitD[0])*((unitD[1] < 0) ? -1 : 1);
#ifdef GRID_HIERARCHY
	tie( e , eend ) = out_edges( *v , g );
	for ( ; e!=eend ; ++e ) {
	  vertex_descriptor other = target( *e , g );
	  s.radius (parentNode.radius() + 
		    nodes[other].radius() + aveChildRad);
	  vector< Vec > child =
	    seriesOfPointsOnSphereSegment2d (s, 1,
					     angle - SPRAY_ANGLE,
					     angle + SPRAY_ANGLE);
	  x.push_back (child[0]);
	}
#else
	x = seriesOfPointsOnSphereSegment2d (s, vertices2place,
					     angle - SPRAY_ANGLE,
					     angle + SPRAY_ANGLE);
#endif
      }//end if
      else {
	cerr << "Three or higher dimesional points placed on complete sphere" << endl;
	cerr << "Need to fix to get three-dimensions working." << endl;
	x = seriesOfPointsOnSphere( s , vertices2place );
      }
    }

    // Initialize the positions of the next level
    int ctr = 0;    
    tie( e , eend ) = out_edges( *v , g );
    for ( ; e!=eend ; ++e ) {
      vertex_descriptor other = target( *e , g );
      nodes[other].X( x[ctr].begin() , x[ctr].end() );
      ++ctr;
#ifdef GRID_HIERARCHY
      gh.loadNode (nodes[other]);
#else
      shift_particle (nodes[other], grid);
#endif
    }
  }
}  

//----------------------------------------------------------

static std::set<long> areParents;

bool doesVertexHaveAnyChildren( Graph_t& G, Graph_t::vertex_descriptor v , out_graph& g , ParentMap& parents )
{ 
  if ( areParents.empty() ) {
    // Generate list for the first time
    for ( ParentMap::size_type ii=0; ii<parents.size(); ++ii ) {
      if ( parents[ii] >= 0 )
	areParents.insert( parents[ii] );
    }
  }

  typedef out_graph::out_edge_iterator oei;
  
  oei e, eend;
  tie( e , eend ) = out_edges( v , g );

  // Go through all child vertices to see if they 
  // have any of their own.
  for ( ; e!=eend ; ++e ) {
      Graph_t::vertex_descriptor other = target( *e , g );
      if ( areParents.find( other ) != areParents.end() ) {
	return true;
      }
  }

  //cout << G.idFromIndex(v) <<  " has no children.\n";
  return false;
}

//----------------------------------------------------------

void generatePlacementVector( FixedVec_p& d , const FixedVec_p& parentNode,
			      const FixedVec_p& parentParentNode ,
			      const FixedVec_p& cm )
{
  // Find the parent node with respect to c.m.
  d = parentNode;
  d -= cm;
  // Make it a unit vector
  prec_t m =  1.0/magnitude( d.begin() , d.end() );
  scale( d.begin() , d.end() , 1.0/m );
  // Find the position of the parent with respect to its parent
  FixedVec_p d2( parentNode );
  d2 -= parentParentNode;
  // Make it a unit vector
  prec_t mag = magnitude( d2.begin() , d2.end() );
  if ( mag > 0 ) {
    m =  1.0/mag;
    scale( d2.begin() , d2.end() , 1.0/m );
    // Average the two
    d += d2;  
    d.scale( .5 );
  }
}

//----------------------------------------------------------

prec_t placementFormula( prec_t placementDistance , int vertices2place , int dimension, prec_t parentRadius)
{
  if ( placementDistance >= 0 ) // Independent of dimension
    return placementDistance;
  else if ( dimension == 2 )
    return parentRadius*min<prec_t>( .25 * sqrt((double)vertices2place) , 10 );
  else // ( dimension == 3 )
    return parentRadius*min<prec_t>( .25 * pow((double)vertices2place,.34) , 10 );
}

//----------------------------------------------------------

prec_t collectOutput ( ThreadArgs * args , PCChaperone& chaperone )
{
  for ( long ii=1; ii<args[0].threadCount; ++ii ) {
    args[0].stats->integrateWithOther( *(args[ii].stats) );
    args[ii].stats->reset();
  }
  // average the stats and get the average
  args[0].stats->avg();
  prec_t dx = args[0].stats->dist();
  // This clears the stats for the next runs
  args[0].stats->reset();
  // Return the avg dx
  return dx;
}

//----------------------------------------------------------

FixedVec_p calcCenterOfMass( Graph_t& g , NodeContainer& nodes , LevelMap& levels ,
			     unsigned int currentLevel )
{
  FixedVec_p center(0);
  prec_t totalMass = 0;
  Graph_t::vertex_iterator vi, vend;
  for ( tie(vi,vend)=vertices(g.boostGraph()); vi!=vend; ++vi )
    {
      if ( levels[*vi] >= currentLevel ) { continue; }
      Node& n = nodes[ *vi ];
      FixedVec_p location( n.X() );
      location.scale( n.mass() );
      center += location;
      totalMass += n.mass();
    }
  center.scale( 1.0/totalMass );
  return center;
}

//---------------------------------------------------------------

static int iteration__ = 0;

void printOutput( long i , prec_t d , long la , ostream& o )
{
  char * im = "Iteration: ";
  char * dx = " Dx: ";
  char * l  = " Level: ";
  if ( iteration__ > 0 ) { 
    int j=0;
    while ( ++j<=(24+20) ) { o << '\b'; }
  }
  o << im << setw(6) << i;  
  o << dx << setw(10) << d;
  o << l << setw(4) << la << flush;
  ++iteration__;
}

//----------------------------------------------------------
// This determines the size of the grid to generate
// to hold all the particles. It requires that the
// particle positions be set already.
//----------------------------------------------------------

void gridPrepAndInit( Grid_t& nw , 
		      prec_t voxelLength, prec_t outerRadius )
{
  typedef NodeContainer::size_type size_type;
  FixedVec_p mins;
  FixedVec_p maxs;

  for (size_type jj=0; jj<mins.size(); ++jj) {
    mins[jj] = 0;
    maxs[jj] = outerRadius;
  }//end for
  
  nw.min(mins);
  nw.max(maxs);
  nw.voxelWidth(voxelLength);
  nw.initGrid();
}

//----------------------------------------------------------
