
#include <iostream>
#include <deque>
#include <tulip/ForEach.h>
#include <tulip/TlpTools.h>
#include <tulip/StableIterator.h>
#include <tulip/TreeTest.h>
#include <tulip/KruskalMST.h>
#include <HierarchyMap.h>
#include "SmashingPeacock.h"

using namespace std;

#define LEVEL_ZERO_ALPHA (255)
#define LEVEL_ONE_ALPHA (5)
#define LEVEL_ONE_LIGHTEN (75)

//the subgraphs used by the SmashingPeacock algorithm
MetaGraphProxy *subgraphs;

//a string proxy that contains the original edge ids
StringProxy *originalEdge;

//A string proxy which contains pairs of nodes in the biconnected 
//components which form the bridges
StringProxy *bridges;

//counts lglcomponents
unsigned int lglcounter = 0;

//a function to generate the originalEdges 
void genOriginalEdges (SuperGraph *graph, vector <edge> &originalEdges) {
  StringProxy *originalEdgesProxy = 
    graph->getProperty <StringProxy> ("Original Edge ID");
  Iterator <edge> *edges = graph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    if (originalEdgesProxy->getEdgeValue (curEdge) != "") {
      istringstream idparse (originalEdgesProxy->getEdgeValue (curEdge));
      unsigned int edgeid;
      while (!idparse.eof()) {
	idparse >> edgeid;
	originalEdges.push_back (edge (edgeid));
      }//end while
    }//end if
  } delete edges; //end while
}//end genOriginalEdges

void getBiconEnds (string edgestring, node *ends) {
  istringstream idparse (edgestring);
  for (int i  = 0; i < 2; ++i) {
    unsigned int nodeid;
    idparse >> nodeid;
    ends[i] = node (nodeid);
  }//end for
}//end getBiconEnds

node treeCenter (SuperGraph *graph) {
  SuperGraph *centerFind = tlp::newCloneSubGraph (graph);
  deque<node> leaves;
  node curNode;
  forEach (curNode, centerFind->getNodes()) { 
    if (centerFind->deg (curNode) == 1) 
      leaves.push_back (curNode);
  }//forEach
  while (centerFind->numberOfNodes() > 2) {
    node curLeaf = leaves.front();
    leaves.pop_front();
    Iterator<edge> *oneEdge = centerFind->getInOutEdges (curLeaf);
    edge leafEdge = oneEdge->next();
    delete oneEdge;
    node root = centerFind->opposite (leafEdge, curLeaf);
    centerFind->delEdge (leafEdge);
    centerFind->delNode (curLeaf);
    if (centerFind->deg (root) == 1) leaves.push_back (root);
  }//end while
  node centre = centerFind->getOneNode();
  graph->delAllSubGraphs (centerFind);
  return centre;
}//end tree centre

void setInitialAlphas (SuperGraph *graph, unsigned int alpha) {
  ColorsProxy *colours = graph->getProperty<ColorsProxy> ("viewColor");
  edge curEdge;
  forEach (curEdge, graph->getEdges()) {
    Color edgeColour = colours->getEdgeValue (curEdge);
    edgeColour.setA (alpha);
    colours->setEdgeValue (curEdge, edgeColour);
  }//end forEach
  node curNode;
  forEach (curNode, graph->getNodes()) {
    Color nodeColour = colours->getNodeValue (curNode);
    nodeColour.setA (alpha);
    colours->setNodeValue (curNode, nodeColour);
  }//end forEach
}//end setInitialAlphas

void setLevelOneAlphas (SuperGraph *graph, SuperGraph *finalLayoutGraph) {
  ColorsProxy *finalColours = 
    finalLayoutGraph->getProperty<ColorsProxy> ("viewColor"); 
  vector<edge> originalEdges;
  genOriginalEdges (graph, originalEdges);
  vector<edge>::iterator originalEdgesIt = originalEdges.begin();
  for (; originalEdgesIt != originalEdges.end(); ++originalEdgesIt) {
    Color orgEdgeColour = finalColours->getEdgeValue (*originalEdgesIt);
    orgEdgeColour.setA (LEVEL_ONE_ALPHA);
    if (orgEdgeColour.getR() + LEVEL_ONE_LIGHTEN < 255)
      orgEdgeColour.setR (orgEdgeColour.getR() + LEVEL_ONE_LIGHTEN);
    else orgEdgeColour.setR (255);
    if (orgEdgeColour.getG() + LEVEL_ONE_LIGHTEN < 255)
      orgEdgeColour.setG (orgEdgeColour.getG() + LEVEL_ONE_LIGHTEN);
    else orgEdgeColour.setG (255);
    if (orgEdgeColour.getB() + LEVEL_ONE_LIGHTEN < 255)
      orgEdgeColour.setB (orgEdgeColour.getB() + LEVEL_ONE_LIGHTEN);
    else orgEdgeColour.setB (255);
    finalColours->setEdgeValue (*originalEdgesIt, orgEdgeColour);
  }//end for
}//end setLevelOneAlphas

void SmashingPeacock::setMST () {
  mst = superGraph->getProperty<SelectionProxy> ("mst");
  KruskalMST::mst (superGraph, mst);
  SuperGraph *mstGraph = tlp::newSubGraph (superGraph);
  node curNode;
  forEach (curNode, superGraph->getNodes()) mstGraph->addNode (curNode);
  edge curEdge;
  forEach (curEdge, superGraph->getEdges())
    if (mst->getEdgeValue (curEdge)) mstGraph->addEdge (curEdge);
  cout << mstGraph->numberOfNodes() << " " << mstGraph->numberOfEdges() << endl;
  mstRoot = treeCenter (mstGraph);
  superGraph->delAllSubGraphs (mstGraph);
}//end getMST

#define NO_ROOT_ID (0xffffffff)
void SmashingPeacock::prepPeacock (SuperGraph *hierarchy, node treeRoot, 
				   node cameFrom, node biconCompRoot) {
  SuperGraph *treeRootSubGraph = subgraphs->getNodeValue (treeRoot);
  if (treeRootSubGraph != NULL) {
    treeRootSubGraph->setAttribute<unsigned int> ("lglroot", biconCompRoot.id);
    if (isType (treeRootSubGraph, UNKNOWN_IDENT)) {
      char lglnumber [10];
      char rootnumber[10];
      sprintf (lglnumber, "%d", lglCounter);
      sprintf (rootnumber, "%d", biconCompRoot.id);
      treeRootSubGraph->setAttribute<string> ("name", string (LGL_IDENT)
					      + lglnumber + " root "
					      + rootnumber);
      ++lglCounter;
    }//end if isType
  }//end if
  edge curEdge;
  node lglRoot;
  forEach (curEdge, hierarchy->getInOutEdges(treeRoot)) {
    node opposite = hierarchy->opposite (curEdge, treeRoot);
    if (opposite != cameFrom) {
      node ends[2];
      getBiconEnds (bridges->getEdgeValue (curEdge), ends);
      if (treeRootSubGraph != NULL) {
	if (treeRootSubGraph->isElement (ends[0])) lglRoot = ends[1];
	else lglRoot = ends[0];
      }//end if
      else {
	if (treeRoot == ends[0]) lglRoot = ends[1];
	else lglRoot = ends[0];
      }//end else
      prepPeacock (hierarchy, opposite, treeRoot, lglRoot);
    }//end if
  }//end forEach
}//end layoutPeacock

bool SmashingPeacock::run () {
  //first compute the mst from the graph
  setMST ();
  setInitialAlphas (superGraph, LEVEL_ZERO_ALPHA);

  //snip all the trees and then all the biconnected components exactly once
  //to layout the quasi-tree
  string errMsg;
  DataSet clusterData;
  SuperGraph *hierarchy, *finalLayoutGraph;

  //set the metric algorithms we choose to run
  string operatorOnce = string (BICONNECTED_IDENT) + ";";
  clusterData.set ("TopoLayout Operator Once", operatorOnce);
  string operatorRecursive = "";
  clusterData.set ("TopoLayout Operator Recursive", operatorRecursive);
  clusterData.set (DATASET_KEEPCOLOURS, true);
  bool success = 
    tlp::clusterizeGraph (superGraph, errMsg, &clusterData, 
			  "Topological Feature Clustering");
  if (!success) {
    cerr << "Topological Feature Clustering returned the following error:\n"
	 << errMsg << endl;
    cerr << "(is the plugin properly installed?)" << endl;
    return false;
  }//end if
  
  //get the final layout and hierarchy graphs
  clusterData.get <SuperGraph *> ("finalLayoutGraph", finalLayoutGraph);
  clusterData.get <SuperGraph *> ("hierarchy", hierarchy);
  subgraphs = hierarchy->getProperty<MetaGraphProxy> (FEATURE_NODE_NAME);
  bridges = hierarchy->getProperty<StringProxy> (BRIDGE_PROXY);
  setLevelOneAlphas (hierarchy, finalLayoutGraph);

  //The top of the hierarchy is a set of biconnected components.
  //We find the component containing the root of the mst
  node biconTreeRoot, curNode;
  forEach (curNode, hierarchy->getNodes()) {
    SuperGraph *curSubGraph = subgraphs->getNodeValue (curNode);
    if (curSubGraph == NULL) {
      if (curNode == mstRoot) {
	biconTreeRoot = curNode;
	break;
      }//end if
    }//end if
    else {
      if (curSubGraph->isElement (mstRoot)) {
	biconTreeRoot = curNode;
	break;
      }//end if
    }//end else
  }//end forEach
  TreeTest::makeRootedTree (hierarchy, biconTreeRoot);
  prepPeacock (hierarchy, biconTreeRoot, biconTreeRoot, mstRoot);
  
  DataSet topoLayoutDrawData;
  string drawOperator = string (BICONNECTED_IDENT) + ";" + 
    string ("RINGS Area") + ";" + string ("RINGS Area") + ";" +
    string (LGL_IDENT) + ";" + string ("LGL Area") + ";" +
    string ("LGL Area") + ";";
  topoLayoutDrawData.set (DATASET_DRAW_OPERATOR, drawOperator);
  topoLayoutDrawData.set<bool> (DATASET_CLUSTER_BUSTER, false);
  topoLayoutDrawData.set<bool> (DATASET_FAST_OVERLAPS, false);
  topoLayoutDrawData.set<bool> (DATASET_TORQUES, false);
  LayoutProxy hierLayout (superGraph);
  success = 
    superGraph->computeProperty ("TopoLayout", &hierLayout, errMsg,
				 NULL, &topoLayoutDrawData);
  if (!success) {
    cerr << "Error in calling TopoLayout:" << endl << errMsg << endl;
    return false;
  }//end if
  forEach (curNode, superGraph->getNodes())
    layoutProxy->setNodeValue (curNode, hierLayout.getNodeValue (curNode));
  return true;
}//end run

bool SmashingPeacock::check (std::string &msg) { return true; }

void SmashingPeacock::reset () {}

LAYOUTPLUGIN(SmashingPeacock, "Smashing Peacock", "Daniel Archambault", "20/06/2005", "Ok", "0", "1");
